package SNMP_lib;

use strict;
use warnings;
use SNMP;
use Socket;
use Data::Dumper;
use Xormon;
use Scalar::Util qw(looks_like_number);


# set directory where MIBs are located
# this works only when SNMP_lib is called from the MS directory e.g. from "<XormonNG wrkdir>/server-nest/microservices/sanbrcd"
my $mibs_dir = "../../../mibs";

our $snmp_host;

# SNMP default parameters
our $snmp_community_string = 'public';
our $snmp_version          = '2c';
our $snmp_timeout          = '50000000'; # default '1000000', micro-seconds before retry
our $snmp_port             = "161";

#SNMP v3 parameters
our $snmp_security_user;
our $snmp_security_level;
our $snmp_auth_protocol;
our $snmp_priv_protocol;
our $snmp_auth_pass;
our $snmp_priv_pass;

#our $security_engine_id;
#our $context_engine_id;
#our $context;

#my $auth_master_key;
#my $priv_master_key;
#my $auth_localized_key;
#my $priv_localized_key;

sub initialize_snmp {
  my $cfg = shift;

  # possible configurations:
  ##################################################################################
  # hw_type: sanbrcd
  #
  # 1. snmpVersion: 1 or 2c
  #    cfg keys:   snmpHost          (ip or hostname, mandatory)
  #                snmpCommunity     (community string, default=public, mandatory)
  #                snmpPort          (default=161, must be less than 65536, mandatory)
  #                virtualFabric     (optional)
  #
  #    snmpwalks:  snmpwalk -v 1  -c 'public' 192.168.1.1     1.3.6.1.2.1.1.5
  #                snmpwalk -v 2c -c 'public' 192.168.1.1     1.3.6.1.2.1.1.5
  #                snmpwalk -v 2c -c 'public' 192.168.1.1:162 1.3.6.1.2.1.1.5
  #
  # 2. snmpVersion: 3
  #    cfg keys:   snmpHost          (ip or hostname, mandatory)
  #                snmpSecUser       (security user, mandatory)
  #                snmpPort          (default=161, must be less than 65536, mandatory)
  #                virtualFabric     (optional)
  #                virtualFabricId   (possible range 1-128 if defined, optional)
  #                snmpSecLevel      (security level, noAuthNoPriv|authNoPriv|authPriv, mandatory)
  #
  #    authNoPriv: snmpAuthProtocol  (MD5|SHA|SHA512, mandatory)
  #                snmpAuthPassword  (<password>, mandatory)
  #
  #    authPriv:   snmpAuthProtocol  (MD5|SHA|SHA512, mandatory)
  #                snmpAuthPassword  (<password>, mandatory)
  #                snmpPrivProtocol  (DES|AES|AES128|AES256)
  #                snmpPrivPassword  (<password>)
  #
  #    snmpwalks:  snmpwalk -v 3 -u 'snmpuser1'                                                                                     192.168.1.1 1.3.6.1.2.1.1.5
  #                snmpwalk -v 3 -u 'snmpuser1' -n VF:<1-128>                                                                       192.168.1.1 1.3.6.1.2.1.1.5
  #                snmpwalk -v 3 -u 'snmpuser1' -l authNoPriv -a <MD5|SHA|SHA512> -A <snmpAuthPassword>                                    192.168.1.1 1.3.6.1.2.1.1.5
  #                snmpwalk -v 3 -u 'snmpuser1' -l authPriv   -a <MD5|SHA|SHA512> -A <snmpAuthPassword> -x <DES|AES|AES128|AES256> -X <snmpPrivPassword> 192.168.1.1 1.3.6.1.2.1.1.5
  #
  ##################################################################################

  #
  # mandatory global
  #
  unless ( exists $cfg->{'snmpHost'} )    { Xormon::error( "Variable \$cfg->{'snmpHost'} doesn't exist! $!"    . __FILE__ . ":" . __LINE__ ) && exit 1; }
  unless ( exists $cfg->{'snmpVersion'} ) { Xormon::error( "Variable \$cfg->{'snmpVersion'} doesn't exist! $!" . __FILE__ . ":" . __LINE__ ) && exit 1; }

  $snmp_host    = $cfg->{'snmpHost'};
  $snmp_version = $cfg->{'snmpVersion'};

  #
  # optional global
  #
  if ( exists $cfg->{'snmpPort'} ) {
    if ( Xormon::isdigit($cfg->{'snmpPort'}) && $cfg->{'snmpPort'} < 65536 ) {
      $snmp_port = $cfg->{'snmpPort'};
    }
    else {
      Xormon::error( "Unsupported variable \$cfg->{'snmpPort'}=$cfg->{'snmpPort'}, automatically used default port $snmp_port! $!"    . __FILE__ . ":" . __LINE__ );
    }
  }

  if ( $cfg->{'snmpVersion'} eq "1" || $cfg->{'snmpVersion'} eq "2c" ) {
    #
    # snmpVersion 1, 2c
    #
    # snmpCommunity (community string, default=public, mandatory)
    if ( exists $cfg->{'snmpCommunity'} && $cfg->{'snmpCommunity'} ne '' ) {
      $snmp_community_string = $cfg->{'snmpCommunity'};
    }
    else {
      Xormon::error( "Variable \$cfg->{'snmpCommunity'} doesn't exist! $!"    . __FILE__ . ":" . __LINE__ ) && exit 1;
    }
    # virtualFabric -> only for MS purposes, not needed in snmp requests (optional)
  }
  elsif ( $cfg->{'snmpVersion'} eq "3" ) {
    #
    # snmpVersion 3
    #
    # snmpSecUser (security user, mandatory)
    if ( exists $cfg->{'snmpSecUser'} && $cfg->{'snmpSecUser'} ne '' ) {
      $snmp_security_user = $cfg->{'snmpSecUser'};
    }
    else {
      Xormon::error( "Variable \$cfg->{'snmpSecUser'} doesn't exist! $!"    . __FILE__ . ":" . __LINE__ ) && exit 1;
    }
    # snmpSecLevel (security level, noAuthNoPriv|authNoPriv|authPriv)
    if ( exists $cfg->{'snmpSecLevel'} ) {
      if ( $cfg->{'snmpSecLevel'} =~ m/^noAuthNoPriv$|^authNoPriv$|^authPriv$/ ) {
        $snmp_security_level = $cfg->{'snmpSecLevel'};
        # authNoPriv
        if ( $cfg->{'snmpSecLevel'} eq "authNoPriv" ) {
          # snmpAuthProtocol (MD5|SHA|SHA512)
          if ( exists $cfg->{'snmpAuthProtocol'} ) {
            if ( $cfg->{'snmpAuthProtocol'} =~ m/^MD5$|^SHA$|^SHA512$/ ) {
              $snmp_auth_protocol = $cfg->{'snmpAuthProtocol'};
            }
            else {
              Xormon::error( "Unsupported variable \$cfg->{'snmpAuthProtocol'}=$cfg->{'snmpAuthProtocol'}! $!" . __FILE__ . ":" . __LINE__ ) && exit 1;
            }
          }
          else {
            Xormon::error( "Variable \$cfg->{'snmpAuthProtocol'} doesn't exist! $!"    . __FILE__ . ":" . __LINE__ ) && exit 1;
          }
          # snmpAuthPassword (<password>)
          if ( exists $cfg->{'snmpAuthPassword'} ) {
            #$snmp_auth_pass = Xormon::unobscure_password($cfg->{'snmpAuthPassword'});
            $snmp_auth_pass = $cfg->{'snmpAuthPassword'};
          }
          else {
            Xormon::error( "Variable \$cfg->{'snmpAuthPassword'} doesn't exist! $!"    . __FILE__ . ":" . __LINE__ ) && exit 1;
          }
        }
        # authPriv
        if ( $cfg->{'snmpSecLevel'} eq "authPriv" ) {
          # snmpAuthProtocol (MD5|SHA|SHA512)
          if ( exists $cfg->{'snmpAuthProtocol'} ) {
            if ( $cfg->{'snmpAuthProtocol'} =~ m/^MD5$|^SHA$|^SHA512$/ ) {
              $snmp_auth_protocol = $cfg->{'snmpAuthProtocol'};
            }
            else {
              Xormon::error( "Unsupported variable \$cfg->{'snmpAuthProtocol'}=$cfg->{'snmpAuthProtocol'}! $!" . __FILE__ . ":" . __LINE__ ) && exit 1;
            }
          }
          else {
            Xormon::error( "Variable \$cfg->{'snmpAuthProtocol'} doesn't exist! $!"    . __FILE__ . ":" . __LINE__ ) && exit 1;
          }
          # snmpAuthPassword (<password>)
          if ( exists $cfg->{'snmpAuthPassword'} ) {
            #$snmp_auth_pass = Xormon::unobscure_password($cfg->{'snmpAuthPassword'});
            $snmp_auth_pass = $cfg->{'snmpAuthPassword'};
          }
          else {
            Xormon::error( "Variable \$cfg->{'snmpAuthPassword'} doesn't exist! $!"    . __FILE__ . ":" . __LINE__ ) && exit 1;
          }
          # snmpPrivProtocol (DES|AES|AES128|AES256)
          if ( exists $cfg->{'snmpPrivProtocol'} ) {
            if ( $cfg->{'snmpPrivProtocol'} =~ m/^DES$|^AES$|^AES128$|^AES256$/ ) {
              $snmp_priv_protocol = $cfg->{'snmpPrivProtocol'};
            }
            else {
              Xormon::error( "Unsupported variable \$cfg->{'snmpPrivProtocol'}=$cfg->{'snmpPrivProtocol'}! $!" . __FILE__ . ":" . __LINE__ ) && exit 1;
            }
          }
          else {
            Xormon::error( "Variable \$cfg->{'snmpPrivProtocol'} doesn't exist! $!"    . __FILE__ . ":" . __LINE__ ) && exit 1;
          }
          # snmpPrivPassword (<password>)
          if ( exists $cfg->{'snmpPrivPassword'} ) {
            #$snmp_priv_pass = Xormon::unobscure_password($cfg->{'snmpPrivPassword'});
            $snmp_priv_pass = $cfg->{'snmpPrivPassword'};
          }
          else {
            Xormon::error( "Variable \$cfg->{'snmpPrivPassword'} doesn't exist! $!"    . __FILE__ . ":" . __LINE__ ) && exit 1;
          }
        }
      }
      else {
        Xormon::error( "Unsupported variable \$cfg->{'snmpSecLevel'}=$cfg->{'snmpSecLevel'}! $!" . __FILE__ . ":" . __LINE__ ) && exit 1;
      }
    }
    else {
      Xormon::error( "Variable \$cfg->{'snmpSecLevel'} doesn't exist! $!"    . __FILE__ . ":" . __LINE__ ) && exit 1;
    }
    # virtualFabric -> only for MS purposes, not needed in snmp requests (optional)
    # virtualFabricId (possible range 1-128 if defined, optional) #TODO
  }
  else {
    Xormon::error( "Unsupported variable \$cfg->{'snmpVersion'}=$cfg->{'snmpVersion'}! $!" . __FILE__ . ":" . __LINE__ ) && exit 1;
  }

  return 1;
}

sub initialize_mibs {

  # loads MIBs needed for given type of data collection
  my $type = shift;

  &SNMP::addMibDirs($mibs_dir);
  $SNMP::verbose = 0;


  if ( $type eq "san" ) {
    &SNMP::loadModules( 'SW-MIB', 'Brocade-REG-MIB', 'Brocade-TC', 'FCMGMT-MIB', 'FA-EXT-MIB', 'FIBRE-CHANNEL-FE-MIB', 'CISCO-FC-FE-MIB', 'CISCO-DM-MIB', 'CISCO-LICENSE-MGR-MIB', 'HA-MIB', 'CISCO-SYSTEM-EXT-MIB' );
  }
  if ( $type eq "lan" ) {
    &SNMP::loadModules( 'SW-MIB',                             'Brocade-REG-MIB',              'Brocade-TC',    'FCMGMT-MIB',    'CISCO-FC-FE-MIB',
                        'CISCO-DM-MIB',                       'Q-BRIDGE-MIB',                 'CISCO-VTP-MIB', 'CISCO-CDP-MIB', 'IEEE8023-LAG-MIB',
                        'CISCO-ETHERNET-FABRIC-EXTENDER-MIB', 'CISCO-STACKWISE-MIB',          'CISCO-VIRTUAL-SWITCH-MIB',
                        'CISCO-PAGP-MIB',                     'CISCO-ENTITY-FRU-CONTROL-MIB', 'CISCO-VLAN-IFTABLE-RELATIONSHIP-MIB',
                        'CISCO-MEMORY-POOL-MIB',              'CISCO-PROCESS-MIB',            'CISCO-SYSTEM-EXT-MIB',
                        'CISCOSB-MIB',                        'CISCOSB-rndMng',               'HUAWEI-MIB', 'HUAWEI-ENTITY-EXTENT-MIB'
    );
  }
  if ( $type eq "SYNOLOGY" ) {
    &SNMP::loadModules( 'SYNOLOGY-FLASHCACHE-MIB', 'SYNOLOGY-SERVICES-MIB', 'SYNOLOGY-SPACEIO-MIB', 'SYNOLOGY-UPS-MIB',
    'SYNOLOGY-DISK-MIB', 'SYNOLOGY-ISCSILUN-MIB', 'SYNOLOGY-SHA-MIB', 'SYNOLOGY-STORAGEIO-MIB', 'UCD-SNMP-MIB',
    'SYNOLOGY-EBOX-MIB', 'SYNOLOGY-RAID-MIB', 'SYNOLOGY-SMART-MIB', 'SYNOLOGY-SYSTEM-MIB', 'IF-MIB', 'HOST-RESOURCES-MIB', 'SYNOLOGY-GPUINFO-MIB' );
  }
  if ($type eq "IFT" ){
    &SNMP::loadModules( 'IF-MIB', 'IFT-SNMP-MIB' );
  }
  if ($type eq "HCP") {
      &SNMP::loadModules('HCP-MIB');
  }

  &SNMP::initMib();

  return 1;
}

sub new_session {

  # creates new session session with given snmp params
  # if some parameter is not defined, then either value from configuration file or default value is used
  my $params          = shift;
  my $community_index = shift;                               # indexing for MIBs which have multiple instances (like for each vlan)
  my %snmpparms       = defined $params ? %{$params} : ();
  my $sess;

  #$snmpparms{DestHost}       = inet_ntoa( inet_aton( $snmpparms{DestHost} ) );

  $snmpparms{DestHost}       = inet_ntoa( inet_aton( $snmp_host ) );
  $snmpparms{Version}        = defined $snmpparms{Version}    ? $snmpparms{Version}    : $snmp_version;
  $snmpparms{Timeout}        = defined $snmpparms{Timeout}    ? $snmpparms{Timeout}    : $snmp_timeout;
  $snmpparms{RemotePort}     = defined $snmpparms{RemotePort} ? $snmpparms{RemotePort} : $snmp_port;
  $snmpparms{UseSprintValue} = '1';

  if ( $snmp_version eq '1' || $snmp_version eq '2c' ) {
    if ( defined $community_index && $community_index ne '' ) {
      $snmpparms{Community} = "$snmp_community_string\@$community_index";
    }
    else {
      $snmpparms{Community} = $snmp_community_string;
    }
  }
  elsif ( $snmp_version eq '3' ) {
    $snmpparms{SecName} = $snmp_security_user;
    #my $session_context = $context;

    #if ( defined $ENV{LAN_TYPE} && $ENV{LAN_TYPE} eq 'CISCO' && $community_index ) {

    #  # in case of v3 community string is used in context parameter
    #  $session_context = "vlan-$community_index";
    #}

    if ( defined $snmp_security_level) { $snmpparms{SecLevel}  = $snmp_security_level; }
    if ( defined $snmp_auth_pass )     { $snmpparms{AuthPass}  = $snmp_auth_pass; }
    if ( defined $snmp_priv_pass )     { $snmpparms{PrivPass}  = $snmp_priv_pass; }
    if ( defined $snmp_auth_protocol ) { $snmpparms{AuthProto} = $snmp_auth_protocol; }
    if ( defined $snmp_priv_protocol ) { $snmpparms{PrivProto} = $snmp_priv_protocol; }
    #if ( defined $sec_engine_id )      { $snmpparms{SecEngineId}      = $sec_engine_id; }
    #if ( defined $context_engine_id )  { $snmpparms{ContextEngineId}  = $context_engine_id; }
    #if ( defined $session_context )    { $snmpparms{Context}          = $session_context; }
    #if ( defined $auth_master_key )    { $snmpparms{AuthMasterKey}    = $auth_master_key; }
    #if ( defined $priv_master_key )    { $snmpparms{PrivMasterKey}    = $priv_master_key; }
    #if ( defined $auth_localized_key ) { $snmpparms{AuthLocalizedKey} = $auth_localized_key; }
    #if ( defined $priv_localized_key ) { $snmpparms{PrivLocalizedKey} = $priv_localized_key; }
  }

  $sess = new SNMP::Session(%snmpparms);

  return $sess;
}

sub collect_san {
  my $data      = shift;
  my $sess      = shift;
  my $section   = shift;
  my $mib       = shift;
  my $mib_alias = shift;
  my $conn_test = shift;
  my $vb        = new SNMP::Varbind( [$mib] );

  if ( !defined $mib_alias || $mib_alias eq '' ) { $mib_alias = $mib; }

  if ( defined $section && $section eq 'configuration' ) {
    my $var = $sess->getnext($vb);

    if ( defined $var ) {
      $var =~ s/"//g;
      $var =~ s/^\s+//g;
      $var =~ s/\s+$//g;
      unless ( $var eq '' ) {
        $data->{configuration}{$mib_alias} = $var;
      }
    }
  }
  else {
    for ( my $var = $sess->getnext($vb); ( $vb->tag eq $mib ) and not( $sess->{ErrorNum} ); $var = $sess->getnext($vb) ) {
      my $index = $vb->iid;
      if ( $mib ne "zoneEnforcedZoneName" && $mib ne "zoneEnforcedZoneMemberID" && $mib ne "zoneSetName" && $mib ne "fcIfNxPortName" && $mib ne "fcFxPortNxPortName" && $mib ne "clmNoOfLicenseCurrentUsages" ) {
        $index =~ s/^.+\.//g;
      }
      $var =~ s/"//g;
      $var =~ s/^\s+//g;
      $var =~ s/\s+$//g;

      if ( $var eq '' ) { next; }

      # get indexed data into data hash
      if ( defined $section ) {
        $data->{$section}{$index}{$mib_alias} = $var;
      }
      else {
        $data->{$mib_alias}{$index} = $var;
      }
    }
  }



  if ( $sess->{ErrorNum} ) {
    if ( defined $conn_test && $conn_test eq "1" ) {
      return (0,"Got $sess->{ErrorStr} querying $sess->{DestHost} with $sess->{Community} for $mib.");
    }
    else {
      Xormon::error( "Got $sess->{ErrorStr} querying $sess->{DestHost} with $sess->{Community} for $mib. $!" . __FILE__ . ":" . __LINE__ ) && exit;
    }
  }

  return 1;
}

sub collect_lan {
  my $data               = shift;
  my $sess               = shift;
  my $section            = shift;
  my $mib                = shift;
  my $mib_alias          = shift;
  my $conn_test          = shift;
  my $iid_index          = shift;
  my $transformation_sub = shift; # is applied to each value before storing in data hash
  my $vb                 = new SNMP::Varbind( [$mib] );
  my %output;

  if ( !defined $mib_alias || $mib_alias eq '' ) { $mib_alias = $mib; }

  if ( defined $section && $section eq 'configuration' ) {
    my $var = $sess->getnext($vb);

    if ( defined $var ) {
      $var =~ s/"//g;
      $var =~ s/^\s+//g;
      $var =~ s/\s+$//g;
      unless ( $var eq '' ) {
        $data->{configuration}{$mib_alias} = $var;
      }
    }
  }
  else {
    for ( my $var = $sess->getnext($vb); ( $vb->tag eq $mib ) and not( $sess->{ErrorNum} ); $var = $sess->getnext($vb) ) {
      my $index = $vb->iid;

      if ( defined $iid_index ) {
        # used for wrongly indexed entries with leading or trailing indexes
        $index = ( split /\./, $vb->iid )[$iid_index];
      }
      $var =~ s/"//g;
      $var =~ s/^\s+//g;
      $var =~ s/\s+$//g;
      $var = defined $transformation_sub ? $transformation_sub->($var) : $var;

      if ( $var eq '' ) { next; }

      # get indexed data into data hash
      if ( defined $section ) {
        $data->{$section}{$index}{$mib_alias} = $var;
      }
      else {
        $data->{$mib_alias}{$index} = $var;
      }

      # get data into hash and return it as reference
      $output{$index} = $var;
    }

  }
  if ( $sess->{ErrorNum} ) {
    if ( defined $conn_test && $conn_test eq "1" ) {
      return (0,"Got $sess->{ErrorStr} querying $sess->{DestHost} with $sess->{Community} for $mib.");
    }
    else {
      Xormon::error( "Got $sess->{ErrorStr} querying $sess->{DestHost} with $sess->{Community} for $mib. $!" . __FILE__ . ":" . __LINE__ ) && exit;
    }
  }

  if ( defined $conn_test && $conn_test eq "1" ) {
    return 1;
  }
  else {
    return \%output;
  }
}

sub collect_synology {
    my ( $data_tmp, $sess, $section, $mib, $metric ) = @_;

    # Dereference the data_tmp hash reference
    my %data = %{$data_tmp};
    my @debug_data = ();
    my %volumes;

    # Ensure $mib is defined and not empty
    unless (defined $mib && $mib ne '') {
        return (0, "MIB is not defined or is empty");
    }

    # Initialize $vb with $mib
    my $vb = new SNMP::Varbind([$mib]);

    if ($section eq 'configuration') {
          my $var = $sess->getnext($vb);

          if (defined $var) {
              $var =~ s/[^\w\s.-]//g;
              $var =~ s/^\s+|\s+$//g;
              unless ($var eq '') {
                  $data{$section}{$mib} = $var;
              }
          }

          if ($sess->{ErrorNum}) {
              if ($sess->{ErrorStr} =~ /Unknown Object Identifier/) {
                return (0, "Unknown Object Identifier querying $sess->{DestHost} with $sess->{Community} for $mib.");
              } elsif ($sess->{ErrorStr} =~ /Timeout/) {
                return (0, "Timeout occurred while querying $sess->{DestHost} with $sess->{Community} for $mib.");
              } else {
                Xormon::error("Got $sess->{ErrorStr} querying $sess->{DestHost} with $sess->{Community} for $mib. $!" . __FILE__ . ":" . __LINE__);
                return (0,"Got $sess->{ErrorStr} querying $sess->{DestHost} with $sess->{Community} for $mib.");
              }
          }

        return (1, $data{$section}{$mib} || '');

      } else {
        # Special case for GPU metrics
        if (defined $metric && $metric =~ m/^gpu/) {
            my $var = $sess->get('enterprises.6574.108.1.0');

            # If the metric is in MB and numeric, convert from MB to B
            if ( $metric =~ m/MB/ && Xormon::isdigit($var) ) {
                $var = Xormon::round( $var * 1024 * 1024, 0 );
            }

            # If the value is not numeric, set it to an empty string
            if ( !Xormon::isdigit($var) ) {
                $var = "";
            }

            # Assign the value to the data hash for the section
            $data{$section}{0}{$metric} = $var;
            $data{$section}{0}{id} = 0;

            # Return the updated data hash
            return \%data;
        }
        else {
            $vb = new SNMP::Varbind([$mib]) if not defined $vb;
        }

        my $i = -1;

        # Iterate over SNMP values
        for ( my $var = $sess->getnext($vb); (defined $var && $vb->tag eq $mib) && not( $sess->{ErrorNum} ); $var = $sess->getnext($vb) ) {
            my $index    = $vb->iid;
            my $drive_id = "";
            $i++;

            # Modify the index format based on the section
            if ( defined $section && $section =~ /CPU-CORE Level|Port Level|Temperature sensor|Power supply|Chassis Disk|Battery|Fan speed/ ) {
                $index =~ s/\./-/g;
            }
            elsif ( defined $section && $section eq "Drive Level Statistics" ) {
                $drive_id = $index;
                $drive_id =~ s/^.+\.//g;
                if ( Xormon::isdigit($drive_id) ) {
                    $drive_id--;
                }
            }
            else {
                $index =~ s/^.+\.//g if defined $index;
            }

            # Clean up the value (remove extra spaces, etc.)
            if (defined $var) {
                $var =~ s/"//g;
                $var =~ s/^\s+//g;
                $var =~ s/\s+$//g;
            }
            next if !defined $var || $var eq '';

            # Special formatting for Volume and Pool levels
            if ( defined $section && defined $metric && $section =~ /Volume Level|Pool Level/ && $metric =~ /name|Free|Capacity|capacity|health_status/ ) {
                my $pom = "$index-$index";
                $index = $pom;
            }

            # Skip metrics that do not match certain conditions
            if ( ( defined $section && defined $metric && $section eq "Port Level Statistics" || $section eq "Port Level Configuration" ) && ( $var !~ m/eth/ && $metric eq "name" ) ) {
                next;
            }
            if ( ( defined $section && defined $metric && $section eq "Port Level Statistics" || $section eq "Port Level Configuration" ) && ( $metric ne "name" && !defined $data{$section}{$index} ) ) {
                next;
            }
            if ( defined $section && $section eq "Memory Level Statistics" && $metric eq "description" && $var =~ m/\// ) {
                next;
            }

            # Adjust memory metrics
            if ( defined $section && $section eq "Memory Level Statistics" && $metric ne "description" ) {
                if ( defined $mib && $mib eq "memAvailReal" ) {
                    $var =~ s/kB//g;
                    $var =~ s/^\s+//g;
                    $var =~ s/\s+$//g;
                    $var = Xormon::round( $var * 1024, 0 ); # to B
                }
            }

            # Assign type for Port Level metrics
            if ( defined $section && $section =~ /Port Level/ ) {
                $data{$section}{$index}{type} = "ETH";
                if ( $section eq "Port Level Configuration" ) {
                    $data{$section}{$index}{Protocol} = "ETH";
                }
            }

            # Convert Latency values to milliseconds
            if ( defined $mib && $mib =~ /Latency/ ) {
                $var = Xormon::round( $var / 1000 );
            }

            # Convert disk and RAID statuses to human-readable form
            if ( defined $mib && $mib =~ /diskStatus/ ) {
                if ( Xormon::isdigit($var) ) {
                    $var = convert_disk_status($var);
                }
            }
            if ( defined $mib && $mib =~ /raidStatus/ ) {
                if ( Xormon::isdigit($var) ) {
                    $var = convert_raid_status($var);
                }
            }

            # Convert system status values to human-readable form
            if ( defined $mib && $mib =~ /systemStatus|powerStatus|cpuFanStatus|systemFanStatus/ ) {
                if ( Xormon::isdigit($var) ) {
                    $var = convert_system_status($var);
                }
            }

            # Assign the final value to the data structure
            $data{$section}{$index}{$metric} = $var;
            $data{$section}{$index}{id} = $index;
            if ( defined $section && $section eq "Drive Level Statistics" ) {
                $data{$section}{$index}{id} = $drive_id;
            }
            if ( defined $section && $section eq "Port Level Configuration" && $metric eq "name" ) {
                $data{$section}{$index}{id} = $index;
            }
            if ( defined $section && $section eq "Port Level Configuration" && $metric eq "alias" ) {
                $data{$section}{$index}{name} = $var;
            }

            if ( defined $mib && $mib eq "raidName" && ($section eq "Volume Level Configuration" || $section eq "Volume Level Statistics") ) {
                if ( defined $var && $var =~ m/Volume/ ) {
                    $volumes{$var} = $var;
                }
                if ( defined $var && $var =~ m/Pool/ ) {
                    my $pool_id = $var;
                    $pool_id =~ s/Storage Pool //g;
                    $pool_id =~ s/^\s+|\s+$//g;
                    next unless Xormon::isdigit($pool_id);
                    foreach my $volume (keys %{$data{$section}}) {
                        my $name = $data{$section}{$volume}{name};
                        if (defined $name && defined $volumes{$name}) {
                            $data{$section}{$volume}{pool_name} = $var;
                            $data{$section}{$volume}{pool_id}   = $pool_id;
                        }
                    }
                    %volumes = ();
                }
            }
        }
      }

    # Return the updated data
    return \%data;
}


# Helper function to convert disk status codes to human-readable form
sub convert_disk_status {
    my ($var) = @_;
    return "normal"                if $var == 1;
    return "Initialized"            if $var == 2;
    return "NotInitialized"         if $var == 3;
    return "SystemPartitionFailed"  if $var == 4;
    return "Crashed"                if $var == 5;
    return "unknown";
}

# Helper function to convert RAID status codes to human-readable form
sub convert_raid_status {
    my ($var) = @_;
    return "normal"    if $var == 1;
    return "degrade"   if $var == 11;
    return "crashed"   if $var == 12;
    return "unknown";
}

# Helper function to convert system status codes to human-readable form
sub convert_system_status {
    my ($var) = @_;
    return "normal"    if $var == 1;
    return "failed"    if $var == 2;
    return "unknown";
}

sub collect_selected {

  # similiar to collect sub but it is used in cases where list of wanted indexes is available
  # so it doesn't need to walk whole OID
  # usually used for querying entPhysicalTable where every OID can have hundreds of entries
  my $data               = shift;           # hash reference - output parameter
  my $sess               = shift;           # session object
  my @selected_indexes   = @{ shift @_ };
  my $section            = shift;           # name of key under which new values will be stored in data hash
  my $mib                = shift;           # name for OID
  my $transformation_sub = shift;           # is applied to each value before storing in data hash
  my $iid_index          = shift;           # used for wrongly indexed entries with leading or trailing indexes
  my %output;
  my $var;
  my $vb;

  foreach my $selected_index (@selected_indexes) {
    $vb  = new SNMP::Varbind( [ $mib, $selected_index ] );
    $var = $sess->get($vb);

    if (  ( $vb->tag eq $mib )
      and not( $sess->{ErrorNum} )
      and ( $var ne 'No Such Instance currently exists at this OID' ) )
    {
      my $index = defined $iid_index ? ( split /\./, $vb->iid )[$iid_index] : $vb->iid;
      $var =~ s/"//g;
      $var = defined $transformation_sub ? $transformation_sub->($var) : $var;

      if ( $var eq '' ) { next; }    # skip empty values

      # get data into hash and return it as reference
      $output{$index} = $var;

      # get indexed data into data hash
      if ( defined $section ) {
        $data->{$section}{$index}{$mib} = $var;
      }
    }
  }

  if ( $sess->{ErrorNum} ) {
    Xormon::error( "Got $sess->{ErrorStr} querying $sess->{DestHost} with $sess->{Community} for $mib. $!" . __FILE__ . ":" . __LINE__ ) && exit;
    return 0;
  }

  return \%output;
}

sub check_mib_compatibility {
  my $hw_type = shift;

  my %compatibility = ();

  #
  # LAN
  #
  if ( $hw_type eq "lancisco" || $hw_type eq "lanothers" ) {
    my $sess_act = new_session( { UseLongNames => 1, Timeout => 1000000 } );

    $compatibility{"Q-BRIDGE-MIB"} = test_mib( $sess_act, 'qBridgeMIB' );
    $compatibility{"LLDP-MIB"}     = test_mib( $sess_act, 'lldpMIB' );

    if ( $hw_type eq 'lancisco' ) {
      $compatibility{"CISCO-CDP-MIB"}                       = test_mib( $sess_act, 'ciscoCdpMIB' );
      $compatibility{"CISCO-VTP-MIB"}                       = test_mib( $sess_act, 'ciscoVtpMIB' );
      $compatibility{"CISCO-PAGP-MIB"}                      = test_mib( $sess_act, 'ciscoPagpMIB' );
      $compatibility{"CISCO-VIRTUAL-SWITCH-MIB"}            = test_mib( $sess_act, 'ciscoVirtualSwitchMIB' );
      $compatibility{"CISCO-STACKWISE-MIB"}                 = test_mib( $sess_act, 'ciscoStackWiseMIB' );
      $compatibility{"CISCO-ETHERNET-FABRIC-EXTENDER-MIB"}  = test_mib( $sess_act, 'ciscoEthernetFabricExtenderMIB' );
      $compatibility{"CISCO-PROCESS-MIB"}                   = test_mib( $sess_act, 'ciscoProcessMIB' );
      $compatibility{"CISCO-MEMORY-POOL-MIB"}               = test_mib( $sess_act, 'ciscoMemoryPoolMIB' );
      $compatibility{"CISCO-SYSTEM-EXT-MIB"}                = test_mib( $sess_act, 'ciscoSystemExtMIB' );
      $compatibility{"CISCO-ENTITY-FRU-CONTROL-MIB"}        = test_mib( $sess_act, 'ciscoEntityFRUControlMIB' );
      $compatibility{"CISCO-VLAN-IFTABLE-RELATIONSHIP-MIB"} = test_mib( $sess_act, 'ciscoVlanIfTableRelationshipMIB' );
      $compatibility{"CISCOSB-rndMng"}                      = test_mib( $sess_act, 'rndMng' );

      # HUAWEI CPU/MEM
      $compatibility{"HUAWEI-ENTITY-EXTENT-MIB"} = test_mib( $sess_act, 'hwEntityExtentMIB' );

      # KZ: some cisco 2950SX switches with older OS version don't support OID dot3adAggPortSelectedAggID
      $compatibility{"IEEE8023-LAG-MIB"} = test_mib( $sess_act, 'lagMIB' );

      # test connection with community string index on vlan 1
      # if it works then switch supports community string indexing and VLAN configruation
      # has to be collected diffrently
      $sess_act = new_session( { UseLongNames => 1, Timeout => 1000000 }, 1 );
      $compatibility{"COMMUNITY-STRING-INDEX"} = test_mib( $sess_act, 'dot1dBridge' );
    }
  }

  return \%compatibility;
}

sub test_mib {

  # predicate testing if session/switch supports given MIB
  my $sess   = shift;
  my $mib    = shift;
  my $vb     = new SNMP::Varbind( [$mib] );
  my $result = $sess->getnext($vb);

  if ( $vb->tag !~ /\.$mib\./ || $sess->{ErrorNum} ) {
    return 0;
  }

  return 1;
}

#INFORTRENT
sub collect_eql {
  my ( $data_tmp, $sess, $section, $mib, $index ) = @_;
  my %data = %{$data_tmp};
  my $i    = 0;
  my $vb;

  if ( $section eq "pool" and defined $index and $index ne "" ) {

    #$vb = new SNMP::Varbind( [$mib] );
    my $var = $sess->get("$mib");
    if ( defined $var and $var ne "" ) {
      $data{$section}{$mib}{$index} = $var;
    }
    else {
      Xormon::error( "Section : $section, MIB : $mib, Metric : $mib, Index : $index, Var : $var" . __FILE__ . ":" . __LINE__ );
      $data{$section}->{$mib}{$index} = "error\n";
    }
    return \%data;
  }    #end if for pools
  else {
    $vb = new SNMP::Varbind( [$mib] );

    for ( my $var = $sess->getnext($vb); ( $vb->tag eq $mib ) and not( $sess->{ErrorNum} ); $var = $sess->getnext($vb) ) {
      my $index = $vb->iid;
      if ( $sess->{ErrorNum} ) {
        Xormon::error( "Section : $section, MIB : $mib, Metric : $mib, Index : $index, Var : $var" . __FILE__ . ":" . __LINE__ );
        $data{$section}{$mib}[$i] = "error";
      }
      else {
        #$data->{$section}->{$mib}[$i] = $var;
        $data{$section}{$mib}[$i] = $var;
        $data{$section}{'index'}[$i] = $vb->iid;
      }
      $i++;
    }

    return \%data;
  }
}

sub collect_hcp {
  my $data      = shift;
  my $sess      = shift;
  my $mib       = shift;
  my $section   = shift || 'configuration';
  my $conn_test = shift || 0;
  my $i  = 0;

  my $vb = new SNMP::Varbind( [$mib] );

  if ( $section eq 'configuration' ) {
    my $var = $sess->getnext($vb);

    if ( defined $var ) {
      $var =~ s/"//g;
      $var =~ s/^\s+//g;
      $var =~ s/\s+$//g;
      unless ( $var eq '' ) {
        $data->{configuration}{$mib} = $var;
      }
    }
  } else {
    for ( my $var = $sess->getnext($vb); ( $vb->tag eq $mib ) and not( $sess->{ErrorNum} ); $var = $sess->getnext($vb) ) {
      # my $index = $vb->iid;
      # next if $var eq '';
      if ( $var eq '' ){ $data->{$section}{$mib}[$i] = undef; }
      else { $data->{$section}{$mib}[$i] = $var; }

      $i++;
    }
  }

  if ( $sess->{ErrorNum} ) {
    if ($conn_test) {
      return (0,"Got $sess->{ErrorStr} querying $sess->{DestHost} with $sess->{Community} for $mib.");
    }

    Xormon::error( "Got $sess->{ErrorStr} querying $sess->{DestHost} with $sess->{Community} for $mib. $!" . __FILE__ . ":" . __LINE__ ) && exit;
  }

  return 1;
}

sub build_snmpwalk_cmd {
  my $params   = shift;
  my $snmpwalk = "snmpwalk";

  if ( exists $params->{Version} ) {
    $snmpwalk .= " -v $params->{Version}";

    if ( $params->{Version} eq "1" || $params->{Version} eq "2c" ) {
      #
      # SNMPv1 and SNMPv2c
      #
      if ( exists $params->{Community} ) {
        #
        # community string
        #
        $snmpwalk .= " -c '$params->{Community}'";
      }
    }
    elsif ( $params->{Version} eq "3" ) {
      #
      # SNMPv3
      #
      if ( exists $params->{SecName} ) {
        #
        # security user
        #
        $snmpwalk .= " -u '$params->{SecName}'";
      }
      if ( exists $params->{Context} && $params->{Context} ne '' ) {
        #
        # VF:ID
        #
        $snmpwalk .= " -n '$params->{Context}'";
      }

      if ( exists $params->{SecLevel} ) {
        #
        # security level: noAuthNoPriv, authNoPriv, authPriv
        #
        if ( $params->{SecLevel} eq "1" ) {
          #
          # noAuthNoPriv
          #
          $snmpwalk .= " -l noAuthNoPriv";
        }
        elsif ( $params->{SecLevel} eq "2" ) {
          #
          # authNoPriv
          #
          $snmpwalk .= " -l AuthNoPriv";

          if ( exists $params->{AuthProto} ) {
            $snmpwalk .= " -a $params->{AuthProto}";
          }
          if ( exists $params->{AuthPass} ) {
            $snmpwalk .= " -A <AuthPass>";
          }
        }
        elsif ( $params->{SecLevel} eq "3" ) {
          #
          # authPriv
          #
          $snmpwalk .= " -l AuthPriv";

          if ( exists $params->{AuthProto} ) {
            $snmpwalk .= " -a $params->{AuthProto}";
          }
          if ( exists $params->{AuthPass} ) {
            $snmpwalk .= " -A <AuthPass>";
          }
          if ( exists $params->{PrivProto} ) {
            $snmpwalk .= " -x $params->{PrivProto}";
          }
          if ( exists $params->{PrivPass} ) {
            $snmpwalk .= " -X <PrivPass>";
          }
        }
      }
    }
  }

  # IP or hostname
  if ( exists $params->{DestHost} ) {
    $snmpwalk .= " '$params->{DestHost}'";
  }

  # sysName OID
  $snmpwalk .= " 1.3.6.1.2.1.1.5";

  return $snmpwalk;
}

1;
