"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.RequestBuilder = exports.RequestMethod = void 0;
const common_1 = require("@nestjs/common");
const hostcfgData_types_1 = require("../common/hostcfgs/hostcfgData.types");
const http = require("http");
const https = require("https");
var RequestMethod;
(function (RequestMethod) {
    RequestMethod["POST"] = "POST";
    RequestMethod["GET"] = "GET";
    RequestMethod["PUT"] = "PUT";
    RequestMethod["DELETE"] = "DELETE";
    RequestMethod["PATCH"] = "PATCH";
})(RequestMethod || (exports.RequestMethod = RequestMethod = {}));
;
class RequestBuilder {
    constructor(proto, url, method) {
        this.proto = proto;
        this.url = url;
        this.method = method;
        this.stream = null;
        this.responseEncoding = "utf8";
        this.protocol = proto === hostcfgData_types_1.PROTO.HTTPS ? https : http;
        this.options = {
            method: method
        };
    }
    addAgent() {
        const agent = this.proto === hostcfgData_types_1.PROTO.HTTPS ? new https.Agent({ rejectUnauthorized: false }) : new http.Agent();
        this.options.agent = agent;
        return this;
    }
    setRequestTimeout(timeoutInMs) {
        this.options.timeout = timeoutInMs;
        return this;
    }
    setStream(stream) {
        this.stream = stream;
        return this;
    }
    setHeaders(headers) {
        this.options.headers = headers;
        return this;
    }
    addBody(body) {
        this.body = body;
        return this;
    }
    setRequestEncoding(encoding) {
        this.responseEncoding = encoding;
        return this;
    }
    static defineBaseRequestEvents(req, reject) {
        req.on("timeout", () => {
            reject({
                statusCode: common_1.HttpStatus.REQUEST_TIMEOUT,
                error: "Request timed out."
            });
        });
        req.on("error", (error) => {
            reject({
                error: error
            });
        });
        req.on("abort", () => {
            reject({
                error: "Request aborted."
            });
        });
    }
    static async simpleRequest(options, useAgent = true) {
        const protocol = options.protocol === hostcfgData_types_1.PROTO.HTTPS ? https : http;
        const rOptions = {
            headers: {
                ...options.headers,
                "content-length": +options.body.length
            },
            method: options.method,
            agent: useAgent ? options.protocol === hostcfgData_types_1.PROTO.HTTPS ? new https.Agent({ rejectUnauthorized: false }) : new http.Agent() : undefined
        };
        return new Promise((resolve, reject) => {
            const req = protocol.request(options.url, rOptions, (res) => {
                res.setEncoding(options.responseEncoding);
                let responseBody = "";
                res.on("data", (chunk) => {
                    responseBody += chunk;
                });
                res.on("error", (err) => {
                    reject({
                        statusCode: res.statusCode,
                        body: responseBody,
                        headers: res.headers,
                        error: err
                    });
                });
                res.on("end", () => {
                    resolve({
                        statusCode: res.statusCode,
                        body: responseBody,
                        headers: res.headers
                    });
                });
            });
            RequestBuilder.defineBaseRequestEvents(req, reject);
            if (options.body) {
                req.write(options.body);
            }
            else {
                req.write("");
            }
            req.end();
        });
    }
    processRequest() {
        return new Promise((resolve, reject) => {
            if (this.stream) {
                const req = this.protocol[this.method === RequestMethod.GET ? "get" : "request"](this.url, this.options, (res) => {
                    if (res.statusCode !== 200) {
                        this.stream.close();
                        reject({
                            statusCode: res.statusCode,
                            headers: res.headers,
                            error: "Something happened."
                        });
                    }
                    res.on("error", (err) => {
                        reject({
                            statusCode: res.statusCode,
                            headers: res.headers,
                            error: err
                        });
                    });
                    res.pipe(this.stream);
                });
                this.stream.on('finish', () => {
                    this.stream.close();
                    resolve({
                        statusCode: common_1.HttpStatus.OK
                    });
                });
                RequestBuilder.defineBaseRequestEvents(req, reject);
            }
            else {
                const req = this.protocol.request(this.url, this.options, (res) => {
                    res.setEncoding(this.responseEncoding);
                    let responseBody = "";
                    res.on("data", (chunk) => {
                        responseBody += chunk;
                    });
                    res.on("error", (err) => {
                        reject({
                            statusCode: res.statusCode,
                            body: responseBody,
                            headers: res.headers,
                            error: err
                        });
                    });
                    res.on("end", () => {
                        resolve({
                            statusCode: res.statusCode,
                            headers: res.headers,
                            body: responseBody
                        });
                    });
                });
                RequestBuilder.defineBaseRequestEvents(req, reject);
                if (this.body) {
                    req.write(JSON.stringify(this.body));
                }
                else {
                    req.write("");
                }
                req.end();
            }
        });
    }
}
exports.RequestBuilder = RequestBuilder;
;
//# sourceMappingURL=request.builder.js.map