"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Upgrade = void 0;
const publicUrl_1 = require("../../common/publicUrl");
const dataSource_1 = require("../../config/dataSource");
const version_entity_1 = require("../../entity/version.entity");
const path_1 = require("path");
const App_1 = require("./App");
const fs_1 = require("fs");
const WinstonLogger_1 = require("../log/WinstonLogger");
class Upgrade {
    static async install(file, allowDowngrade = false, rollback = false, forced = false) {
        const logger = (0, WinstonLogger_1.LoggerFactory)('upgrade');
        const util = require('util');
        const exec = util.promisify(require('child_process').exec);
        const preUpgradeScriptPath = this.UPGRADE_TMP_PATH + '/server-nest/dist/pre-upgrade.js';
        const postUpgradeScriptPath = this.UPGRADE_TMP_PATH + '/server-nest/dist/post-upgrade.js';
        if (rollback) {
            logger.log("info", "---------- ROLLBACK ----------");
        }
        else {
            logger.log("info", "---------- UPGRADE ----------");
        }
        try {
            if (!this.fs.existsSync(this.TMP_PATH)) {
                await this.fs.mkdirSync(this.TMP_PATH);
            }
            const tarFile = file.startsWith("/")
                ? file
                : this.TMP_PATH + '/' + file;
            await this.cleanUpgradeDir();
            if (tarFile.endsWith('.tar.gz')) {
                await this.fse.copySync(tarFile, this.UPGRADE_TMP_PATH + '/pre-installed.tar.gz', { overwrite: true });
                await this.tar.extract({
                    file: tarFile,
                    cwd: this.UPGRADE_TMP_PATH,
                    sync: true,
                    gzip: true
                });
            }
            else if (tarFile.endsWith('.tar')) {
                await this.fse.copySync(tarFile, this.UPGRADE_TMP_PATH + '/pre-installed.tar', { overwrite: true });
                await this.tar.extract({
                    file: tarFile,
                    cwd: this.UPGRADE_TMP_PATH,
                    sync: true
                });
            }
            else {
                await this.fs.copySync(tarFile, this.UPGRADE_TMP_PATH + '/' + 'pre-installed-' + path_1.default.basename(tarFile), { overwrite: true });
                console.error(`[Upgrade] Unsupported file extension: ${tarFile}`);
                logger.log("error", `Unsupported file extension: ${tarFile}`);
                return {
                    success: false,
                    message: `The file ${tarFile} has an unsupported extension. Please upload a file with a supported extension.`
                };
            }
            if (!(0, fs_1.existsSync)(this.UPGRADE_TMP_PATH + '/server-nest/package.json')) {
                console.error("[Upgrade] package.json not found in the archive. It's possible that an incorrect archive has been uploaded.");
                logger.log("error", "package.json not found in the archive");
                return {
                    success: false,
                    message: "The archive does not contain a package.json file. Please ensure you're uploading the correct archive."
                };
            }
            if (this.fs.existsSync(preUpgradeScriptPath)) {
                try {
                    console.log("Running pre-upgrade script...");
                    logger.log("info", "Running pre-upgrade script...");
                    const { stdout } = await exec(`node ${preUpgradeScriptPath}`);
                    if (stdout.trim()) {
                        console.log(`Pre-upgrade script output: ${stdout}`);
                        logger.log("info", `Pre-upgrade script output: ${stdout}`);
                    }
                }
                catch (error) {
                    console.error(`Error running pre-upgrade script: ${error}`);
                    logger.log("error", `Error running pre-upgrade script: ${error}`);
                    return {
                        success: false,
                        message: "An error occurred while running the pre-upgrade script. Please check the logs for more details."
                    };
                }
            }
            const packageJson = require(this.ROOT_PATH + '/server-nest/package.json');
            const version = packageJson.version;
            const upgradePackageJson = require(this.UPGRADE_TMP_PATH + '/server-nest/package.json');
            const upgradeVersion = upgradePackageJson.version;
            let developVersion = false;
            if (this.fs.existsSync(this.UPGRADE_TMP_PATH + '/server-nest/.develop')) {
                developVersion = true;
            }
            console.log("[Upgrade] current version: " + version + ", version to upgrade: " + upgradeVersion);
            logger.log("info", "Current version: " + version + ", version to upgrade: " + upgradeVersion);
            let downgrade = false;
            if ((!(0, publicUrl_1.isNewerVersion)(version, upgradeVersion)) && (!(developVersion && version === upgradeVersion))) {
                if (allowDowngrade || rollback) {
                    downgrade = true;
                }
                else {
                    this.fs.rmSync(this.UPGRADE_TMP_PATH, { recursive: true, force: true });
                    this.fs.rmSync(this.TMP_PATH + '/' + file + '.tar.gz', { recursive: true, force: true });
                    logger.log("info", "The same or newer version is already installed");
                    return {
                        success: false,
                        message: "the same or newer version is already installed"
                    };
                }
            }
            if (downgrade && !rollback) {
                console.log("Downgrade to a lower version...");
                logger.log("info", "Downgrade to a lower version...");
                const downgradedVersionSplit = upgradeVersion.split(".");
                const downgradedVersionFolder = downgradedVersionSplit[0] + "." + downgradedVersionSplit[1];
                const migrationFolders = (0, fs_1.readdirSync)(this.MIGRATION_PATH, { withFileTypes: true })
                    .filter(item => item.isDirectory() && +item.name >= +downgradedVersionFolder)
                    .map(item => item.name)
                    .sort((a, b) => { return +b - +a; });
                for (const migrationFolder of migrationFolders) {
                    const migrationFiles = (0, fs_1.readdirSync)(this.MIGRATION_PATH + "/" + migrationFolder, { withFileTypes: true })
                        .filter(item => !item.isDirectory() &&
                        (item.name.includes(".migration.js") && (!item.name.includes(".js.map"))) &&
                        (!this.fs.existsSync(this.UPGRADE_TMP_PATH + '/server-nest/dist/migration/' + migrationFolder + "/" + item.name)))
                        .map(item => item.name)
                        .sort((a, b) => b.localeCompare(a, undefined, { numeric: true, sensitivity: 'base' }));
                    try {
                        for (const migrationFile of migrationFiles) {
                            console.log("[Revert] " + migrationFile);
                            logger.log("info", "[Revert] " + migrationFile);
                            await (0, dataSource_1.getDataSource)().undoLastMigration({ transaction: 'each' });
                        }
                    }
                    catch (error) {
                        console.error("[Migration Error]: " + error);
                        console.log("Initiating rollback and reapplying migrations...");
                        logger.log("error", "[Migration Error]: " + error);
                        logger.log("info", "Initiating rollback and reapplying migrations...");
                        await (0, dataSource_1.getDataSource)().runMigrations();
                        return {
                            success: false,
                            message: 'An error occurred during the database migration process. The system is now rolling back and reapplying migrations.'
                        };
                    }
                }
            }
            const toUpgrade = [
                'server-nest/dist',
                'server-nest/client',
                'server-nest/templates',
                'server-nest/os-agent',
                'server-nest/node_modules',
                'server-nest/microservices',
                'server-nest/ssl',
                'server-nest/production.config.js',
                'server-nest/key.pub',
                'server-nest/package.json',
                'server-nest/package-lock.json',
                'perl-modules',
                'mibs',
                'xormon.sh',
                'COPYRIGHT',
                'LICENSING'
            ];
            for (const file of toUpgrade) {
                if (this.fs.existsSync(this.UPGRADE_TMP_PATH + "/" + file)) {
                    const title = downgrade ? "[Downgrade] " : "[Upgrade] ";
                    console.log(title + this.ROOT_PATH + "/" + file);
                    logger.log("info", title + this.ROOT_PATH + "/" + file);
                    if (file !== 'perl-modules') {
                        await this.fs.rmSync(this.ROOT_PATH + "/" + file, { recursive: true, force: true });
                    }
                    await this.fse.copySync(this.UPGRADE_TMP_PATH + "/" + file, this.ROOT_PATH + "/" + file, { overwrite: true });
                }
            }
            if ((0, dataSource_1.getDataSource)() !== undefined) {
                try {
                    if ((0, publicUrl_1.isNewerVersion)('0.5.5', upgradeVersion)) {
                        await (0, dataSource_1.getDataSource)().getRepository(version_entity_1.Version).insert({
                            version: upgradeVersion
                        });
                    }
                }
                catch (error) {
                    console.warn("[Upgrade] Error while inserting new version into the version table: " + error);
                    logger.log("warn", "Error while inserting new version into the version table: " + error);
                }
            }
            try {
                const publicUrl = await (0, publicUrl_1.getPublicUrl)();
                if (publicUrl !== undefined && publicUrl !== null) {
                    await (0, publicUrl_1.setPublicUrlToUI)(publicUrl);
                }
                await (0, publicUrl_1.setPublicUrlToSourceMaps)(publicUrl);
            }
            catch (error) {
                console.error("[PublicUrlSet] Error while setting public URL: " + error);
                logger.log("error", "Error while setting public URL: " + error);
            }
            if (this.fs.existsSync(this.UPGRADE_TMP_PATH + "/gui.sh")) {
                const exec = require('child_process').exec;
                await new Promise((resolve, reject) => {
                    exec(this.UPGRADE_TMP_PATH + "/gui.sh", (error, stdout, stderr) => {
                        console.log('[Upgrade] Executing post-upgrade GUI script');
                        console.log('[Upgrade] Script output: ' + stdout);
                        if (stderr) {
                            console.log('[Upgrade] Script error output: ' + stderr);
                        }
                        if (error !== null) {
                            console.error('[Upgrade] Error during script execution: ' + error);
                            return reject(error);
                        }
                        resolve(null);
                    });
                });
            }
            logger.log("info", "Running migrations...");
            await this.runMigrations();
            logger.log("info", "Migrations were successful");
            if (!rollback) {
                if (this.fs.existsSync(this.FILES_PATH + '/installed-version.tar.gz')) {
                    await this.fse.copySync(this.FILES_PATH + '/installed-version.tar.gz', this.FILES_PATH + '/previous-version.tar.gz', { overwrite: true });
                }
                await this.fse.copySync(tarFile, this.FILES_PATH + '/installed-version.tar.gz', { overwrite: true });
                await this.fs.rmSync(this.TMP_PATH + '/' + file + '.tar.gz', { recursive: true, force: true });
            }
            if (this.fs.existsSync(postUpgradeScriptPath)) {
                try {
                    console.log("Running post-upgrade script...");
                    logger.log("info", "Running post-upgrade script...");
                    const { stdout } = await exec(`node ${postUpgradeScriptPath}`);
                    if (stdout.trim()) {
                        console.log(`Post-upgrade script output: ${stdout}`);
                        logger.log("info", `Post-upgrade script output: ${stdout}`);
                    }
                }
                catch (error) {
                    console.error(`Error running post-upgrade script: ${error}`);
                    logger.log("error", `Error running post-upgrade script: ${error}`);
                    return {
                        success: false,
                        message: "An error occurred while running the post-upgrade script. Please check the logs for more details."
                    };
                }
            }
            await this.fs.rmSync(this.UPGRADE_TMP_PATH, { recursive: true, force: true });
            try {
                console.log("Cleaning up logs.");
                logger.log("info", "Cleaning up logs.");
                await this.cleanupLogs();
            }
            catch (error) {
                console.error("Error while cleaning up logs: ", error);
                logger.log("error", "Error while cleaning up logs: " + error);
            }
            setTimeout(async () => {
                if (rollback) {
                    console.log(`Rolled back to version ${upgradeVersion}. The application is now resetting.`);
                    logger.log("info", `Rolled back to version ${upgradeVersion}. The application is now resetting.`);
                }
                else {
                    console.log(`Successfully ${downgrade ? 'downgraded' : 'upgraded'} to version ${upgradeVersion}. The application is now resetting.`);
                    logger.log("info", `Successfully ${downgrade ? 'downgraded' : 'upgraded'} to version ${upgradeVersion}. The application is now resetting.`);
                }
                const { execSync, spawn } = require("child_process");
                try {
                    spawn('./xormon.sh', ['upgrade_start'], {
                        cwd: App_1.App.ROOT_PATH,
                        detached: true,
                        stdio: 'inherit'
                    }, (err, stdout, stderr) => {
                        if (err) {
                            console.error('Error occurred while starting the upgrade: ', err);
                            return;
                        }
                        console.log('Upgrade start script output: ', stdout);
                    });
                }
                catch (error) {
                    console.error('Error occurred while executing the upgrade start script: ', error);
                }
            }, 5000);
            if (rollback && forced) {
                return {
                    success: false,
                    message: "The upgrade process failed. Please check the logs for more details. The application is now resetting."
                };
            }
            return {
                success: true,
                message: `Successfully ${rollback ? 'rolled back' : downgrade ? 'downgraded' : 'upgraded'} to version ${upgradeVersion}. The application is now resetting to apply changes.`
            };
        }
        catch (error) {
            console.error(error);
            logger.error(error);
            if (error.code === 'TAR_BAD_ARCHIVE' || error.code === 'TAR_ABORT') {
                await this.fse.copySync(error.file, error.file + '.corrupted', { overwrite: true });
                return {
                    success: false,
                    message: "The uploaded tar file is invalid or corrupted. Please verify the file and try again."
                };
            }
            if (!rollback) {
                const installedVersion = this.FILES_PATH + '/installed-version.tar.gz';
                if (this.fs.existsSync(installedVersion)) {
                    console.log("Initiating rollback to previous version...");
                    logger.log("info", "Initiating rollback to previous version...");
                    return this.install(installedVersion, true, true, true);
                }
            }
            else {
                console.error("Rollback failed. Please generate logs and contact our support for further assistance.");
                console.log("For log generation, navigate to the server-nest directory and run the following command: node dist/dump");
                logger.log("error", "Rollback failed. Please check the logs for more details.");
            }
            return {
                success: false,
                message: "An error occurred. Please check the logs for more details. Error: " + error
            };
        }
    }
    static async cleanUpgradeDir() {
        if (!this.fs.existsSync(this.UPGRADE_TMP_PATH)) {
            await this.fs.mkdirSync(this.UPGRADE_TMP_PATH);
        }
        else {
            await this.fs.rmSync(this.UPGRADE_TMP_PATH, { recursive: true, force: true });
            await this.fs.mkdirSync(this.UPGRADE_TMP_PATH);
        }
    }
    static async runMigrations() {
        return new Promise((resolve, reject) => {
            const { execSync, spawn } = require("child_process");
            spawn('npm', ['run', 'typeorm:migration:run'], {
                cwd: App_1.App.ROOT_PATH + "/server-nest",
                detached: true,
                stdio: 'inherit'
            })
                .on('data', function (data) {
                console.log('Migration output: ', data);
            })
                .on('exit', function (code) {
                if (code == 0) {
                    console.log('Migrations completed successfully.');
                    resolve(true);
                }
                else {
                    const message = "Error occurred during migrations. Exit code: " + code;
                    console.error(message);
                    reject(message);
                }
            }).on('error', function (error) {
                const message = `Error occurred during migrations: ${error}`;
                console.error(message);
                reject(message);
            });
        });
    }
    static async cleanupLogs() {
        try {
            const files = (0, fs_1.readdirSync)(this.LOGS_PATH);
            files.forEach(file => {
                const filePath = this.LOGS_PATH + '/' + file;
                if (!this.logsToKeep.includes(file)) {
                    this.fs.rmSync(filePath, { recursive: true, force: true });
                }
            });
        }
        catch (error) {
            console.error('Error while cleaning up logs: ', error);
        }
    }
}
exports.Upgrade = Upgrade;
Upgrade.ROOT_PATH = (0, path_1.resolve)('../');
Upgrade.FILES_PATH = (0, path_1.resolve)('./files');
Upgrade.TMP_PATH = (0, path_1.resolve)('./files/tmp');
Upgrade.UPGRADE_TMP_PATH = (0, path_1.resolve)('./files/tmp/upgrade');
Upgrade.MIGRATION_PATH = (0, path_1.resolve)('./dist/migration');
Upgrade.LOGS_PATH = (0, path_1.resolve)('./logs');
Upgrade.fs = require('fs');
Upgrade.fse = require('fs-extra');
Upgrade.tar = require('tar');
Upgrade.logsToKeep = [
    'alerting.log',
    'upgrade.log',
    'upgrade-error.log',
    'metrics-table.log',
];
//# sourceMappingURL=Upgrade.js.map