"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ConfigurationTable = void 0;
const pageResponse_constants_1 = require("../../../common/constants/pageResponse.constants");
const dataSource_1 = require("../../../config/dataSource");
const itemProperty_entity_1 = require("../../../entity/itemProperty.entity");
const objectItem_entity_1 = require("../../../entity/objectItem.entity");
const MetricsWrapper_1 = require("../../metrics/MetricsWrapper");
const typeorm_1 = require("typeorm");
const Table_1 = require("./Table");
const URL_1 = require("./URL");
const subsystem_entity_1 = require("../../../entity/subsystem.entity");
const ACLCheck_1 = require("../../acl/ACLCheck");
const User_1 = require("../../users/User");
class ConfigurationTable {
    constructor(itemIds, properties, parentSubsystems) {
        this.fixedCols = 0;
        this.itemIds = [];
        this.items = [];
        this.properties = [];
        this.sumProperties = [];
        this.parentSubsystems = [];
        this.parentSubsystemsDefinition = [];
        this.sortIndex = 0;
        this.sortAsc = true;
        this.sortString = false;
        this.parentIdAccess = [];
        this.itemIds = itemIds;
        if (properties) {
            this.properties = properties;
        }
        if (parentSubsystems) {
            this.parentSubsystems = parentSubsystems;
        }
    }
    setFixedCols(fixedCols) {
        this.fixedCols = fixedCols;
    }
    setItemLabel(label) {
        this.itemLabel = label;
    }
    sortByColumn(index, sort, isString = false) {
        this.sortIndex = index;
        this.sortAsc = sort === 'asc' ? true : false;
        this.sortString = isString;
    }
    setAcl(request) {
        this.request = request;
        return this;
    }
    setSumProperties(properties) {
        this.sumProperties = properties;
    }
    setGroup(group) {
        this.group = group;
        return this;
    }
    async loadItems() {
        if (this.itemIds.length === 0) {
            console.error("[ConfigurationTable]: no items defined");
            return;
        }
        this.items = await (0, dataSource_1.getDataSource)().getRepository(objectItem_entity_1.ObjectItem).find({
            where: {
                item_id: (0, typeorm_1.In)(this.itemIds),
                parents: this.parentSubsystems.length >= 1 ? {
                    _parent: {
                        subsystem: (0, typeorm_1.In)(this.parentSubsystems)
                    }
                } : undefined
            }, relations: this.parentSubsystems.length >= 1 ? {
                parents: {
                    _parent: {
                        server_group_items: {
                            _server_group: true
                        }
                    }
                },
                server_group_items: {
                    _server_group: true
                }
            } : {
                server_group_items: {
                    _server_group: true
                }
            }
        });
        if (this.request) {
            const ids = await ACLCheck_1.ACLCheck.filterItemsWithRequest(this.request, this.items.map(item => item.item_id));
            this.items = this.items.filter(item => ids.includes(item.item_id));
            this.parentIdAccess = await ACLCheck_1.ACLCheck.filterItemsWithRequest(this.request, this.items.map(i => i.parents ? i.parents.map(p => p.parent) : []).flat());
        }
        if (this.items[0]) {
            this.hwType = this.items[0].hw_type;
            this.subsystem = this.items[0].subsystem;
            if (!this.itemLabel) {
                this.itemLabel = this.items[0].subsystem;
            }
        }
        if (this.parentSubsystems.length >= 1 && this.hwType) {
            this.parentSubsystemsDefinition = await (0, dataSource_1.getDataSource)().getRepository(subsystem_entity_1.Subsystem).findBy({ subsystem: (0, typeorm_1.In)(this.parentSubsystems), hw_type: this.hwType });
        }
    }
    async generate() {
        await this.loadItems();
        if (this.items.length === 0) {
            return {
                type: pageResponse_constants_1.BOX_TYPE.TextBox,
                text: "No items found"
            };
        }
        const configurations = await MetricsWrapper_1.MetricsWrapper.getConfigurationFindClass(this.hwType, this.subsystem);
        if (Object.keys(configurations).length === 0 || !configurations) {
            return {
                type: pageResponse_constants_1.BOX_TYPE.TextBox,
                text: "No configuration data found"
            };
        }
        if (this.properties.length === 0) {
            this.properties = Object.keys(configurations);
        }
        const props = await (0, dataSource_1.getDataSource)().getRepository(itemProperty_entity_1.ItemProperty).findBy({
            item_id: (0, typeorm_1.In)(this.itemIds),
            property_name: (0, typeorm_1.In)(this.properties)
        });
        let table = new Table_1.Table();
        table.addHeader({ value: this.itemLabel });
        for (const parentSubsystem of this.parentSubsystems) {
            const subsystemDefinition = this.parentSubsystemsDefinition.find(subsys => subsys.subsystem === parentSubsystem);
            table.addHeader({ value: subsystemDefinition ? subsystemDefinition.label : parentSubsystem });
        }
        const isString = (string) => {
            return isNaN(Number(string));
        };
        let isNumberMapping = {};
        for (const propertyKey of this.properties) {
            let isNumber = false;
            if (configurations[propertyKey] && configurations[propertyKey].type) {
                isNumber = configurations[propertyKey].type === 'number';
            }
            else {
                const checkFirstValue = props.find(prop => prop.property_name === propertyKey && prop.property_value != null && prop.property_value != '' && prop.property_value != ' ');
                isNumber = checkFirstValue && !isString(checkFirstValue.property_value);
            }
            isNumberMapping[propertyKey] = isNumber;
            if (!configurations[propertyKey]) {
                table.addHeader({
                    value: propertyKey,
                    number: isNumber ? isNumber : undefined
                });
            }
            else {
                table.addHeader({
                    value: configurations[propertyKey].label ? configurations[propertyKey].label : propertyKey,
                    number: isNumber ? isNumber : undefined,
                    unit: configurations[propertyKey].unit ? configurations[propertyKey].unit : undefined,
                    prefix: configurations[propertyKey].defaultPrefix ? configurations[propertyKey].defaultPrefix : undefined,
                    decimal: configurations[propertyKey].decimal !== undefined ? configurations[propertyKey].decimal : undefined
                });
            }
        }
        let propKeyValue = {};
        for (const prop of props) {
            if (propKeyValue[prop.item_id]) {
                propKeyValue[prop.item_id][prop.property_name] = prop.property_value;
            }
            else {
                propKeyValue[prop.item_id] = {
                    [prop.property_name]: prop.property_value
                };
            }
        }
        let user = null;
        if (this.request?.session) {
            user = await User_1.User.getUserByRequest(this.request);
        }
        let rows = [];
        for (const item of this.items) {
            let row = [];
            row.push({
                order: 0,
                value: item.label,
                url: new URL_1.URL().generateFromObject(item, null, this.group, user),
            });
            let i = 1;
            for (const parentSubsystem of this.parentSubsystems) {
                const parent = item.parents.find(item => item._parent.subsystem === parentSubsystem);
                row.push({
                    order: i,
                    value: parent ? parent._parent.label : ' ',
                    url: parent && this.parentIdAccess.includes(parent.parent) || !this.request ? new URL_1.URL().generateFromObject(parent._parent, null, this.group, user) : undefined
                });
                i++;
            }
            for (const propertyKey of this.properties) {
                const value = propKeyValue[item.item_id] && propKeyValue[item.item_id][propertyKey] ? propKeyValue[item.item_id][propertyKey] : null;
                row.push({
                    order: i,
                    value: isNumberMapping[propertyKey] && value !== null ? +value : value,
                });
                i++;
            }
            rows.push(row);
        }
        table.addRows(rows);
        if (this.sumProperties) {
            let sumPropertiesData = {};
            for (const sumProperty of this.sumProperties) {
                const propertyIndex = this.properties.findIndex(property => property === sumProperty);
                if (propertyIndex !== -1) {
                    const index = this.parentSubsystems.length + (propertyIndex + 1);
                    let sum = 0;
                    for (const row of rows) {
                        const dataColumn = row.find(column => column.order === index);
                        if (dataColumn) {
                            sum += +dataColumn.value;
                        }
                    }
                    sumPropertiesData[index] = sum;
                }
            }
            if (Object.keys(sumPropertiesData).length > 1) {
                const sumRow = [];
                sumRow.push({
                    order: 0,
                    value: 'TOTAL',
                    styleClass: pageResponse_constants_1.TABLE_STYLE_CLASS.bold
                });
                let i = 1;
                for (const parentSubsystem of this.parentSubsystems) {
                    sumRow.push({
                        order: i,
                        value: null
                    });
                    i++;
                }
                for (const propertyKey of this.properties) {
                    sumRow.push({
                        order: i,
                        value: sumPropertiesData[i] ? sumPropertiesData[i] : null,
                        styleClass: pageResponse_constants_1.TABLE_STYLE_CLASS.bold
                    });
                    i++;
                }
                table.addRow(sumRow);
            }
        }
        if (this.fixedCols) {
            table.setFixedCols(this.fixedCols);
        }
        table.addExport(this.itemIds, this.properties);
        table.sortByIndex(this.sortIndex, this.sortAsc, this.sortString);
        return table.generateBox();
    }
}
exports.ConfigurationTable = ConfigurationTable;
//# sourceMappingURL=ConfigurationTable.js.map