"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.HealthStatus = void 0;
const healthStatus_constants_1 = require("../../../common/constants/healthStatus.constants");
const pageResponse_constants_1 = require("../../../common/constants/pageResponse.constants");
const subsystemMetricsResponse_constants_1 = require("../../../common/constants/subsystemMetricsResponse.constants");
const dataSource_1 = require("../../../config/dataSource");
const healthStatus_entity_1 = require("../../../entity/healthStatus.entity");
const hostcfg_entity_1 = require("../../../entity/hostcfg.entity");
const objectItem_entity_1 = require("../../../entity/objectItem.entity");
const Definitions_1 = require("../../metrics/Definitions");
const typeorm_1 = require("typeorm");
const Table_1 = require("./Table");
const URL_1 = require("./URL");
const User_1 = require("../../users/User");
const ACLCheck_1 = require("../../acl/ACLCheck");
const pageResponse_dto_1 = require("../../../common/pageResponse.dto");
class HealthStatus {
    constructor() {
        this.objectItemsRepository = (0, dataSource_1.getDataSource)().getRepository(objectItem_entity_1.ObjectItem);
        this.hostcfgRepository = (0, dataSource_1.getDataSource)().getRepository(hostcfg_entity_1.HostCfg);
        this.healthstatusRepository = (0, dataSource_1.getDataSource)().getRepository(healthStatus_entity_1.HealthStatus);
        this.LAST_DATA_UPDATE_LIMIT = HealthStatus.LAST_DATA_UPDATE_LIMIT;
        this.deviceAliasColumn = false;
        this.linkToItem = false;
        this.clickableAlias = true;
    }
    async getStorageGlobal() {
        let devices = await this.hsQueryFrame({
            subsystem: 'device',
            hw_type_entity: {
                class: 'storage'
            }
        });
        return await this.generateBasicTable(devices, 'Storage');
    }
    async getStatusByIds(deviceIds, itemColumnLabel = 'Label') {
        const device = await this.hsQueryFrame({
            item_id: (0, typeorm_1.In)(deviceIds)
        });
        return await this.generateBasicTable(device, itemColumnLabel);
    }
    async getStatusByParameters(subsystem, _class, itemName, hw_type) {
        let devices = await this.hsQueryFrame({
            subsystem: subsystem.toLowerCase(),
            hw_type_entity: {
                class: _class.toLowerCase()
            },
            hw_type: hw_type ? hw_type.toLowerCase() : undefined
        });
        return await this.generateBasicTable(devices, itemName);
    }
    async hsQueryFrame(whereClause) {
        return await this.objectItemsRepository.find({
            where: {
                ...whereClause,
                health_status: {
                    type: "health"
                }
            }, relations: {
                health_status: true,
                hw_type_entity: true,
                hostcfg_relations: {
                    _hostcfg: true
                }
            }
        });
    }
    async generateBasicTable(devices, itemName, user, path) {
        if (this.request) {
            const devicesIds = await ACLCheck_1.ACLCheck.filterItemsWithRequest(this.request, devices.map(item => item.item_id));
            devices = devices.filter(device => devicesIds.includes(device.item_id));
        }
        let rows = [];
        let header = [
            { value: 'Status', styleClass: pageResponse_constants_1.TABLE_STYLE_CLASS.healthStatus },
            { value: itemName },
            { value: 'Last update', number: true, unit: subsystemMetricsResponse_constants_1.UNIT.timestamp },
            { value: 'Status reason', trim: true, trimLength: 80 }
        ];
        if (this.deviceAliasColumn) {
            header = [
                { value: 'Status', styleClass: pageResponse_constants_1.TABLE_STYLE_CLASS.healthStatus },
                { value: "Alias" },
                { value: itemName },
                { value: 'Last update', number: true, unit: subsystemMetricsResponse_constants_1.UNIT.timestamp },
                { value: 'Status reason', trim: true, trimLength: 80 }
            ];
        }
        let hsTable = new Table_1.Table();
        for (const columnHeader of header) {
            hsTable.addHeader(columnHeader);
        }
        const url = new URL_1.URL();
        const score = HealthStatus.getScoreForSorting();
        for (const device of devices) {
            let status = healthStatus_constants_1.STATUS.UNKNOWN;
            let last_update;
            if (!('health_status' in device) || device.health_status === null) {
                continue;
            }
            else {
                if (score[device.health_status.status] > score[status]) {
                    status = this.transformEntityStatus(device.health_status.status);
                    last_update = device.health_status.updatedAt !== null && device.health_status.updatedAt !== undefined ? device.health_status.updatedAt : null;
                }
            }
            let statusReason = device.health_status ? device.health_status.data : '';
            if (device.hostcfg_relations && device.hostcfg_relations.find(hostcfg => hostcfg._hostcfg.disabled === true)) {
                status = healthStatus_constants_1.STATUS.UNKNOWN;
                statusReason = 'Device is disabled';
            }
            else if (device.hostcfg_relations && device.hostcfg_relations.find(hostcfg => hostcfg._hostcfg.ignore_health_status === true)) {
                status = healthStatus_constants_1.STATUS.IGNORED;
                const hostcfg = device.hostcfg_relations.find(hostcfg => hostcfg._hostcfg.ignore_health_status === true);
                statusReason = hostcfg._hostcfg.ignore_health_status_reason ? `Ignored: ${hostcfg._hostcfg.ignore_health_status_reason}` : 'Status is ignored';
            }
            else if (status !== healthStatus_constants_1.STATUS.CRITICAL && (new Date().getTime() - new Date(last_update).getTime() >= this.LAST_DATA_UPDATE_LIMIT)) {
                status = healthStatus_constants_1.STATUS.UNKNOWN;
                statusReason = 'No current data';
            }
            let defaultRows = [
                { order: 0, value: status, bgStatus: status },
                { order: 1, value: device.label, url: url.generateFromObject(device, 'health-status', path, user, null) },
                { order: 2, value: last_update ? new Date(last_update).getTime() / 1000 : null },
                { order: 3, value: statusReason }
            ];
            if (this.linkToItem) {
                defaultRows = [
                    { order: 0, value: status, bgStatus: status },
                    { order: 1, value: device.label, url: url.generateFromObject(device, null, path, user, null) },
                    { order: 2, value: last_update ? new Date(last_update).getTime() / 1000 : null },
                    { order: 3, value: statusReason }
                ];
            }
            if (this.deviceAliasColumn) {
                let id = "";
                let label = "";
                if (device.hostcfg_relations && device.hostcfg_relations[0]._hostcfg) {
                    let hstcfg_id = device.hostcfg_relations[0]._hostcfg.hostcfg_id;
                    label = device.hostcfg_relations[0]._hostcfg.label;
                    id = hstcfg_id;
                }
                defaultRows = [
                    { order: 0, value: status, bgStatus: status },
                    { order: 1, value: label, url: url.generateFromObject(device, 'health-status', path, user, null) },
                    { order: 2, value: device.label, url: url.generateFromObject(device, 'health-status', path, user, null) },
                    { order: 3, value: last_update ? new Date(last_update).getTime() / 1000 : null },
                    { order: 4, value: statusReason }
                ];
                if (this.linkToItem) {
                    defaultRows = [
                        { order: 0, value: status, bgStatus: status },
                        { order: 1, value: label, url: this.clickableAlias ? url.generateFromObject(device, 'health-status', path, user, null) : "" },
                        { order: 2, value: device.label, url: url.generateFromObject(device, null, path, user, null) },
                        { order: 3, value: last_update ? new Date(last_update).getTime() / 1000 : null },
                        { order: 4, value: statusReason }
                    ];
                }
            }
            rows.push(defaultRows);
        }
        rows.sort(function (a, b) {
            let aliasScore = b[1].value.localeCompare(a[1].value, undefined, { numeric: true, sensitivity: 'base' });
            let aScore = score[a[0].value];
            let bScore = score[b[0].value];
            if (aScore > (bScore + aliasScore)) {
                return -1;
            }
            else if (aScore == (bScore + aliasScore)) {
                return aliasScore ? -1 : 1;
            }
            else {
                return 1;
            }
        });
        rows.sort((a, b) => ((score[a[0].value] < score[b[0].value]) ? 1 : -1));
        hsTable.addRows(rows);
        return hsTable.generateBox();
    }
    async getDeviceSubsystems(device_id, subsystems) {
        const items = await this.objectItemsRepository.find({
            where: {
                parents: {
                    parent: device_id
                },
                subsystem: subsystems ? (0, typeorm_1.In)(subsystems) : (0, typeorm_1.In)(['node', 'drive', 'port', 'pool', 'volume', 'vdc'])
            }, relations: {
                parents: true,
                health_status: true
            }
        });
        let hsTable = new Table_1.Table()
            .addHeader({ value: 'Status', styleClass: pageResponse_constants_1.TABLE_STYLE_CLASS.healthStatus })
            .addHeader({ value: 'Subsystem' })
            .addHeader({ value: 'Label' })
            .addHeader({ value: 'Last update', number: true, unit: subsystemMetricsResponse_constants_1.UNIT.timestamp })
            .addHeader({ value: 'Status reason', trim: true, trimLength: 80 });
        for (const item of items) {
            if (!('health_status' in item) || item.health_status === null) {
                continue;
            }
            hsTable.addRow([
                { order: 0, value: item.health_status ? item.health_status.status : healthStatus_constants_1.STATUS.UNKNOWN, bgStatus: item.health_status ? item.health_status.status : healthStatus_constants_1.STATUS.UNKNOWN },
                { order: 1, value: item.subsystem },
                { order: 2, value: item.label },
                { order: 3, value: item.health_status.updatedAt !== null ? new Date(item.health_status.updatedAt).getTime() / 1000 : 'unknown' },
                { order: 4, value: item.health_status ? item.health_status.data : '' }
            ]);
        }
        return hsTable.generateBox();
    }
    async getHistory(_class, subsystem, hwType, itemLabel) {
        let where = {
            hw_type_entity: {
                class: _class
            },
            subsystem: subsystem
        };
        if (hwType) {
            where['hw_type'] = hwType;
        }
        let items = await this.objectItemsRepository.find({
            where: where,
            relations: {
                hw_type_entity: true,
                health_status_history: true
            },
            take: 50
        });
        if (this.request) {
            items = await ACLCheck_1.ACLCheck.filterObjectItemsWithRequest(this.request, items);
        }
        let historyTable = new Table_1.Table()
            .addHeader({ value: 'Status', styleClass: pageResponse_constants_1.TABLE_STYLE_CLASS.healthStatus })
            .addHeader({ value: itemLabel ? itemLabel : 'Item' })
            .addHeader({ value: 'Type' })
            .addHeader({ value: 'Created', number: true, unit: subsystemMetricsResponse_constants_1.UNIT.timestamp })
            .addHeader({ value: 'Ended', number: true, unit: subsystemMetricsResponse_constants_1.UNIT.timestamp })
            .addHeader({ value: 'Duration (hours)', number: true })
            .addHeader({ value: 'Status reason', trim: true, trimLength: 80 });
        for (const item of items) {
            if (item.health_status_history) {
                for (const history of item.health_status_history) {
                    const created = history.createdAt.getTime() / 1000;
                    const ended = history.endedAt ? history.endedAt.getTime() / 1000 : undefined;
                    historyTable.addRow([
                        { order: 0, value: history.status, bgStatus: history.status },
                        { order: 1, value: item.label, url: URL_1.URL.generate(item.hw_type, 'health-status', item.item_id) },
                        { order: 2, value: history.type },
                        { order: 3, value: created },
                        { order: 4, value: ended },
                        { order: 5, value: ended ? Math.floor((ended - created) / 60 / 60) : Math.floor(((new Date().getTime() / 1000) - created) / 60 / 60) },
                        { order: 6, value: history.reason ? history.reason : '' }
                    ]);
                }
            }
        }
        historyTable.sortByIndex(3);
        return historyTable.generateBox();
    }
    getMirrors() {
        return new pageResponse_dto_1.MirrorBox();
    }
    async getGlobalStatus(_class = undefined, _hwType = undefined, statusSubsystems = undefined) {
        let statusDefinitions;
        if (!statusSubsystems) {
            statusDefinitions = (0, Definitions_1.getStatusDefinitions)(_class, _hwType);
            statusSubsystems = Object.values(statusDefinitions).map(definition => definition.explicit);
        }
        const hsTableDump = await this.healthstatusRepository.find({
            where: {
                _item: {
                    subsystem: (0, typeorm_1.In)(statusSubsystems)
                }
            },
            relations: {
                _item: {
                    hw_type_entity: {
                        _class: true
                    },
                    hostcfg_relations: {
                        _hostcfg: true
                    }
                }
            },
        });
        const ids = (hsTableDump.map(hs => hs.item_id));
        const uniqueIds = ids.filter((value, index) => ids.indexOf(value) === index);
        const user = await User_1.User.getUserByRequest(this.request, true);
        const groupIds = user ? user.groups.map(group => group.group_id) : [];
        const cleanItemIds = groupIds !== undefined ? await ACLCheck_1.ACLCheck.filterItems(groupIds, uniqueIds) : [];
        let statuses = [];
        for (const item_id of cleanItemIds) {
            const deviceStatuses = hsTableDump.filter(value => value.item_id === item_id);
            const device = deviceStatuses[0]._item;
            if (statusDefinitions && statusDefinitions[device.hw_type].explicit !== device.subsystem) {
                continue;
            }
            const hostcfg = device.hostcfg_relations[0]._hostcfg;
            let deviceStatus = {
                id: item_id,
                health: healthStatus_constants_1.STATUS.UNDEFINED,
                connection: healthStatus_constants_1.STATUS.UNDEFINED,
                data: healthStatus_constants_1.STATUS.UNDEFINED,
                healthUpdatedAt: null,
                connectionUpdatedAt: null,
                dataUpdatedAt: null,
                name: device.label,
                class: device.hw_type_entity._class.label ? device.hw_type_entity._class.label : "unknown",
                hwTypeLabel: device.hw_type_entity.label ? device.hw_type_entity.label : device.hw_type,
                hwType: device.hw_type,
                parent: hostcfg.label,
                reason: "",
                link: hostcfg.data['managementUrl'] ? hostcfg.data['managementUrl'] : undefined
            };
            for (const typeStatus of deviceStatuses) {
                deviceStatus[typeStatus.type] = typeStatus.status ? typeStatus.status : healthStatus_constants_1.STATUS.UNKNOWN;
                deviceStatus[`${typeStatus.type}UpdatedAt`] = typeStatus.updatedAt !== null ? typeStatus.updatedAt : null;
                if (deviceStatus[typeStatus.type] !== healthStatus_constants_1.STATUS.CRITICAL && this.isLastUpdateOld(typeStatus.updatedAt)) {
                    deviceStatus[typeStatus.type] = healthStatus_constants_1.STATUS.UNKNOWN;
                }
                if (deviceStatus[typeStatus.type] !== healthStatus_constants_1.STATUS.OK) {
                    deviceStatus.reason += typeStatus.data ? typeStatus.data + "\n" : "\n";
                }
                if (typeStatus.type === "health" && statusDefinitions && statusDefinitions[deviceStatus.hwType].calculated) {
                    const childrenHealth = await this.objectItemsRepository.find({
                        where: {
                            parents: {
                                _parent: {
                                    item_id: deviceStatus.id
                                }
                            },
                            subsystem: (0, typeorm_1.In)(statusDefinitions[deviceStatus.hwType].calculated),
                        },
                        relations: {
                            parents: {
                                _parent: true
                            },
                            health_status: true
                        }
                    });
                    let children = childrenHealth.filter(childrenHealth => childrenHealth.health_status.status != 'ok');
                    for (const child of children) {
                        if ('health_status' in child && child.health_status.type === 'health') {
                            let currentHS = child.health_status;
                            if (currentHS.status === healthStatus_constants_1.STATUS.CRITICAL || currentHS.status === healthStatus_constants_1.STATUS.WARNING) {
                                if (deviceStatus.health !== healthStatus_constants_1.STATUS.CRITICAL) {
                                    deviceStatus.health = currentHS.status;
                                    deviceStatus.healthUpdatedAt = currentHS.updatedAt;
                                }
                                deviceStatus.reason += `${currentHS.data}\n`;
                            }
                        }
                    }
                    if (deviceStatus.health !== healthStatus_constants_1.STATUS.CRITICAL && this.isLastUpdateOld(deviceStatus.healthUpdatedAt)) {
                        deviceStatus.health = healthStatus_constants_1.STATUS.UNKNOWN;
                    }
                }
            }
            let url;
            if (deviceStatus.class === 'Storage' || deviceStatus.class === 'SAN') {
                url = URL_1.URL.generate(deviceStatus.hwType, 'health-status', deviceStatus.id);
            }
            else if (deviceStatus.class === 'Cloud' || deviceStatus.class === 'Container' || deviceStatus.class === 'Server') {
                if (deviceStatus.hwType === 'power') {
                    url = URL_1.URL.generate(deviceStatus.hwType, 'health-status');
                }
                else {
                    url = URL_1.URL.generate(deviceStatus.hwType, 'heatmap');
                }
            }
            else {
                url = URL_1.URL.generate(deviceStatus.hwType, 'health-status');
            }
            deviceStatus.url = url;
            statuses.push(deviceStatus);
        }
        return statuses;
    }
    getGlobalStatusTable(type = 'all', deviceStatuses) {
        let rows = [];
        let header;
        if (type !== 'all') {
            header = [
                { value: type, styleClass: pageResponse_constants_1.TABLE_STYLE_CLASS.healthStatus },
                { value: 'Last update', number: true, unit: subsystemMetricsResponse_constants_1.UNIT.timestamp },
                { value: "Label" },
                { value: "Parent" },
                { value: "Class" },
                { value: "HW Type" },
                { value: "Mgmt" },
                { value: 'Status reason', trim: true, trimLength: 80 }
            ];
        }
        else {
            header = [
                { value: 'Health', styleClass: pageResponse_constants_1.TABLE_STYLE_CLASS.healthStatus },
                { value: 'Connection', styleClass: pageResponse_constants_1.TABLE_STYLE_CLASS.healthStatus },
                { value: 'Data', styleClass: pageResponse_constants_1.TABLE_STYLE_CLASS.healthStatus },
                { value: "Label" },
                { value: "Parent" },
                { value: "Class" },
                { value: "HW Type" },
                { value: "Mgmt" },
                { value: 'Status reason', trim: true, trimLength: 80 }
            ];
        }
        let hsTable = new Table_1.Table();
        for (const columnHeader of header) {
            hsTable.addHeader(columnHeader);
        }
        const score = HealthStatus.getScoreForSorting();
        score.ok = 2;
        score.unknown = 0;
        score.ignored = 0;
        score.undefined = 0;
        for (const deviceStatus of deviceStatuses) {
            let defaultRow;
            if (type !== 'all') {
                const lcType = type.toLocaleLowerCase();
                const displayReason = (deviceStatus[lcType] === healthStatus_constants_1.STATUS.OK ? "" : deviceStatus.reason);
                const time = deviceStatus['last' + type] ? new Date(deviceStatus['last' + type]).getTime() / 1000 : null;
                defaultRow = [
                    { order: 0, value: [deviceStatus[lcType], ` - ${time}`], bgStatus: deviceStatus[lcType] },
                    { order: 1, value: deviceStatus[`${lcType}UpdatedAt`] ? new Date(deviceStatus[`${lcType}UpdatedAt`]).getTime() / 1000 : null },
                    { order: 2, value: deviceStatus.name, url: deviceStatus.url },
                    { order: 3, value: deviceStatus.parent, url: deviceStatus.url },
                    { order: 4, value: deviceStatus.class },
                    { order: 5, value: deviceStatus.hwTypeLabel },
                    { order: 6, value: deviceStatus.link ? 'link' : '', url: deviceStatus.link ? deviceStatus.link : ' ' },
                    { order: 7, value: displayReason }
                ];
            }
            else {
                defaultRow = [
                    { order: 0, value: [deviceStatus.health, ` - ${deviceStatus.healthUpdatedAt}`], bgStatus: deviceStatus.health },
                    { order: 1, value: [deviceStatus.connection, ` - ${deviceStatus.connectionUpdatedAt}`], bgStatus: deviceStatus.connection },
                    { order: 2, value: [deviceStatus.data, ` - ${deviceStatus.dataUpdatedAt}`], bgStatus: deviceStatus.data },
                    { order: 3, value: deviceStatus.name, url: deviceStatus.url },
                    { order: 4, value: deviceStatus.parent, url: deviceStatus.url },
                    { order: 5, value: deviceStatus.class },
                    { order: 6, value: deviceStatus.hwTypeLabel },
                    { order: 7, value: deviceStatus.link ? 'link' : '', url: deviceStatus.link ? deviceStatus.link : ' ' },
                    { order: 8, value: deviceStatus.reason }
                ];
            }
            rows.push(defaultRow);
        }
        if (type !== 'all') {
            rows.sort(function (a, b) {
                let aliasScore = b[2].value[0].localeCompare(a[2].value[0], undefined, { numeric: true, sensitivity: 'base' });
                let aScore = score[a[0].value[0]];
                let bScore = score[b[0].value[0]];
                if (aScore > (bScore + aliasScore)) {
                    return -1;
                }
                else if (aScore == (bScore + aliasScore)) {
                    return aliasScore ? -1 : 1;
                }
                else {
                    return 1;
                }
            });
            rows.sort((a, b) => ((score[a[0].value[0]] < score[b[0].value[0]]) ? 1 : -1));
        }
        else {
            rows.sort(function (a, b) {
                let aliasScore = a[3].value[0].localeCompare(b[3].value[0], undefined, { numeric: true, sensitivity: 'base' });
                let aScore = (score[a[0].value[0]] + score[a[1].value[0]] + score[a[2].value[0]]);
                let bScore = (score[b[0].value[0]] + score[b[1].value[0]] + score[b[2].value[0]]);
                return aScore > (bScore + aliasScore) ? -1 : 1;
            });
        }
        hsTable.addRows(rows);
        return hsTable.generateBox();
    }
    async getCustomStatusTable(_class, _hwType, _subsystem) {
        const deviceStatuses = await this.getGlobalStatus(_class, _hwType, [_subsystem]);
        let devices = await this.hsQueryFrame({
            subsystem: _subsystem,
            hw_type_entity: {
                class: _class.toLowerCase()
            },
            hw_type: _hwType
        });
        let idServerState = {};
        devices.forEach(function (item) {
            idServerState[item.item_id] = item.health_status;
        });
        let hsTable = new Table_1.Table();
        let rows = [];
        let header = [
            { value: 'Health', styleClass: pageResponse_constants_1.TABLE_STYLE_CLASS.healthStatus },
            { value: 'Connection', styleClass: pageResponse_constants_1.TABLE_STYLE_CLASS.healthStatus },
            { value: 'Data', styleClass: pageResponse_constants_1.TABLE_STYLE_CLASS.healthStatus },
            { value: 'Server' },
            { value: 'State' },
            { value: 'Last update', number: true, unit: subsystemMetricsResponse_constants_1.UNIT.timestamp },
            { value: 'Status reason', trim: true, trimLength: 80 }
        ];
        for (const columnHeader of header) {
            hsTable.addHeader(columnHeader);
        }
        for (const deviceStatus of deviceStatuses) {
            let serverState = "";
            if (deviceStatus.id in idServerState) {
                serverState = idServerState[deviceStatus.id].data;
            }
            let lastUpdate = deviceStatus.dataUpdatedAt ? deviceStatus.dataUpdatedAt : deviceStatus.healthUpdatedAt;
            let lastUpdateDate = lastUpdate ? new Date(lastUpdate).getTime() / 1000 : null;
            let statusRow = {};
            if (/Long term monitor is not enabled/.test(deviceStatus.reason)) {
                statusRow = { order: 6, value: [`${deviceStatus.reason}`, "Follow documentation to enable it."], url: [null, 'https://xormon.com/IBM-Power-monitoring-installation.php'] };
            }
            else {
                statusRow = { order: 6, value: deviceStatus.reason };
            }
            let defaultRow = [
                { order: 0, value: [deviceStatus.health, ` - ${deviceStatus.healthUpdatedAt}`], bgStatus: deviceStatus.health },
                { order: 1, value: [deviceStatus.connection, ` - ${deviceStatus.connectionUpdatedAt}`], bgStatus: deviceStatus.connection },
                { order: 2, value: [deviceStatus.data, ` - ${deviceStatus.dataUpdatedAt}`], bgStatus: deviceStatus.data },
                { order: 3, value: deviceStatus.name, url: URL_1.URL.generate(_hwType, _subsystem, deviceStatus.id) },
                { order: 4, value: serverState },
                { order: 5, value: lastUpdateDate },
                statusRow
            ];
            rows.push(defaultRow);
        }
        const score = HealthStatus.getScoreForSorting();
        score.ok = 2;
        score.unknown = 0;
        score.ignored = 0;
        score.undefined = 0;
        rows.sort(function (a, b) {
            let aliasScore = a[3].value[0].localeCompare(b[3].value[0], undefined, { numeric: true, sensitivity: 'base' });
            let aScore = (score[a[0].value[0]] + score[a[1].value[0]] + score[a[2].value[0]]);
            let bScore = (score[b[0].value[0]] + score[b[1].value[0]] + score[b[2].value[0]]);
            return aScore > (bScore + aliasScore) ? -1 : 1;
        });
        hsTable.addRows(rows);
        return hsTable.generateBox();
    }
    async getGlobalStatusPage() {
        let page = {
            title: "Global status",
            tabs: [],
        };
        const help = {
            text: 'How it works?',
            url: 'https://xormon.com/healthstatus.php'
        };
        let statuses = await this.getGlobalStatus();
        page.tabs.push({
            title: 'Summary',
            content: [this.getGlobalStatusTable('all', statuses)],
            help: help
        });
        page.tabs.push({
            title: 'Health',
            content: [this.getGlobalStatusTable("Health", statuses)],
            help: help
        });
        page.tabs.push({
            title: 'Connection',
            content: [this.getGlobalStatusTable("Connection", statuses)],
            help: help
        });
        page.tabs.push({
            title: 'Data',
            content: [this.getGlobalStatusTable("Data", statuses)],
            help: help
        });
        return page;
    }
    isLastUpdateOld(timestamp) {
        if (timestamp) {
            return (new Date().getTime() - new Date(timestamp).getTime() > this.LAST_DATA_UPDATE_LIMIT);
        }
        else {
            return 1;
        }
    }
    transformEntityStatus(status) {
        if (status === healthStatus_entity_1.Status.ERROR) {
            return healthStatus_constants_1.STATUS.CRITICAL;
        }
        else {
            return healthStatus_constants_1.STATUS[status.toUpperCase()];
        }
    }
    static getScoreForSorting() {
        return {
            [healthStatus_constants_1.STATUS.UNDEFINED]: 1,
            [healthStatus_constants_1.STATUS.IGNORED]: 2,
            [healthStatus_constants_1.STATUS.UNKNOWN]: 3,
            [healthStatus_constants_1.STATUS.OK]: 4,
            [healthStatus_constants_1.STATUS.WARNING]: 5,
            [healthStatus_constants_1.STATUS.CRITICAL]: 6,
        };
    }
    setAcl(request) {
        this.request = request;
        return this;
    }
    useDeviceAliasColumn(deviceAliasColumn) {
        this.deviceAliasColumn = deviceAliasColumn;
        return this;
    }
    useLinkToItem(linkToItem) {
        this.linkToItem = linkToItem;
        return this;
    }
    useClickableAlias(clickableAlias) {
        this.clickableAlias = clickableAlias;
        return this;
    }
}
exports.HealthStatus = HealthStatus;
HealthStatus.LAST_DATA_UPDATE_LIMIT = 60 * 60 * 1000 * 3;
//# sourceMappingURL=HealthStatus.js.map