"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TableGenerator = void 0;
const dataSource_1 = require("../../config/dataSource");
const typeorm_1 = require("typeorm");
const winston = require("winston");
const MetricsWrapper_1 = require("./MetricsWrapper");
const Format_1 = require("../log/Format");
const hwType_entity_1 = require("../../entity/hwType.entity");
class TableGenerator {
    static getMaxMetricsByClass() {
        return {
            'storage': {
                'device': [
                    'data_rate', 'read', 'write', 'io_rate', 'read_io', 'write_io', 'respt_t', 'resp_t_r', 'resp_t_w'
                ],
                'pool': [
                    'data_rate', 'read', 'write', 'io_rate', 'read_io', 'write_io', 'respt_t', 'resp_t_r', 'resp_t_w'
                ],
                'port': [
                    'data_rate', 'read', 'write'
                ],
                'fcoeport': [
                    'data_rate', 'read', 'write'
                ],
                'fcport': [
                    'data_rate', 'read', 'write'
                ],
                'iscsiport': [
                    'data_rate', 'read', 'write'
                ],
                'sasport': [
                    'data_rate', 'read', 'write'
                ],
                'ethport': [
                    'data_rate', 'read', 'write'
                ],
                'logport': [
                    'data_rate', 'read', 'write'
                ],
                'beport': [
                    'data_rate', 'read', 'write'
                ],
                'feport': [
                    'data_rate', 'read', 'write'
                ],
                'rdfport': [
                    'data_rate', 'read', 'write'
                ],
                'cpu': [
                    'cpu_system', 'cpu_busy'
                ],
                'node': [
                    'cpu_system', 'cpu_busy'
                ],
                'fedirector': [
                    'cpu_system', 'cpu_busy'
                ],
                'bedirector': [
                    'cpu_system', 'cpu_busy'
                ],
                'rdfdirector': [
                    'cpu_system', 'cpu_busy'
                ],
            },
            'san': {
                'device': [
                    'data_in', 'data_out', 'cpu_system'
                ],
                'sanport': [
                    'data_in', 'data_out'
                ],
            }
        };
    }
    static async getMaxMetrics() {
        const maxMetricsByClass = this.getMaxMetricsByClass();
        let maxMetricsByHwType = {
            'power': {
                'vm': [
                    'cpu_physical_cores',
                    'cpu_physical_percent',
                    'cpu_utilization_cores',
                    'cpu_utilization_percent'
                ],
                'server': [
                    'cpu_physical_cores',
                    'cpu_physical_percent',
                    'cpu_utilization_cores',
                    'cpu_utilization_percent',
                    'totalProcUnits'
                ],
                'pool': [
                    'totalUtilizedProcUnitsPercent',
                    'utilizedProcUnits',
                ]
            },
            'nutanix': {
                'vm': [
                    'cpu_utilization_cores',
                    'cpu_utilization_percent'
                ],
                'host': [
                    'cpu_utilization_cores',
                    'cpu_utilization_percent'
                ]
            },
            'proxmox': {
                'vm': [
                    'cpu_utilization_cores',
                    'cpu_utilization_percent'
                ],
                'node': [
                    'cpu_utilization_cores',
                    'cpu_utilization_percent'
                ]
            },
            'oraclevm': {
                'vm': [
                    'cpu_utilization_cores',
                    'cpu_utilization_percent'
                ],
                'server': [
                    'cpu_utilization_cores',
                    'cpu_utilization_percent'
                ]
            },
            'ovirt': {
                'vm': [
                    'cpu_utilization_cores',
                    'cpu_utilization_percent'
                ],
                'host': [
                    'cpu_utilization_cores',
                    'cpu_utilization_percent'
                ]
            },
            'vmware': {
                'vm': [
                    'cpu_utilization_cores',
                    'cpu_utilization_percent',
                    'cpu_usage'
                ],
                'esxi': [
                    'cpu_utilization_percent',
                    'cpu_usage'
                ]
            },
            'windows': {
                'vm': [
                    'cpu_utilization_cores'
                ],
                'server': [
                    'cpu_usage',
                    'cpu_usage_perc'
                ]
            },
            'timescaledb': {
                'db': [
                    'deadlocks'
                ]
            }
        };
        if ((0, dataSource_1.getDataSource)() == undefined) {
            await (0, dataSource_1.createDataSource)();
        }
        const hwTypes = await (0, dataSource_1.getDataSource)().getRepository(hwType_entity_1.HwType).findBy({
            class: (0, typeorm_1.In)(Object.keys(maxMetricsByClass))
        });
        for (const _class in maxMetricsByClass) {
            const classHwTypes = hwTypes.filter(hwType => hwType.class === _class);
            for (const classHwType of classHwTypes) {
                for (const subsystem in maxMetricsByClass[_class]) {
                    if (!maxMetricsByHwType[classHwType.hw_type]) {
                        maxMetricsByHwType[classHwType.hw_type] = {
                            [subsystem]: maxMetricsByClass[_class][subsystem]
                        };
                    }
                    else {
                        maxMetricsByHwType[classHwType.hw_type][subsystem] = maxMetricsByClass[_class][subsystem];
                    }
                }
            }
        }
        return maxMetricsByHwType;
    }
    static async createPerformanceTableFromEntity(entity) {
        if (entity.schema === "metrics") {
            const tableNameSplit = entity.tableName.split("_");
            const hwType = tableNameSplit[1];
            const subsystem = tableNameSplit[3] !== undefined
                ? tableNameSplit[2] + "_" + tableNameSplit[3]
                : tableNameSplit[2];
            const min5minutesInterval = await MetricsWrapper_1.MetricsWrapper.getMin5MinutesInterval();
            const maxMetrics = await this.getMaxMetrics();
            const hasMax = maxMetrics[hwType] && maxMetrics[hwType][subsystem];
            try {
                const entityManager = (0, dataSource_1.getDataSource)();
                const exists = await entityManager.query("SELECT to_regclass('metrics." + entity.tableName + "')");
                if (exists[0]['to_regclass'] === null) {
                    await entityManager.transaction(async (transactionalEntityManager) => {
                        let sql = 'CREATE TABLE IF NOT EXISTS metrics.' +
                            entity.tableName +
                            ' ( time TIMESTAMP WITH TIME ZONE NOT NULL, item_id varchar NOT NULL';
                        for (const column in entity.columns) {
                            if (entity.columns[column].propertyName !== 'time' && entity.columns[column].propertyName !== 'item_id') {
                                sql += ', "' + entity.columns[column].propertyName + '" ' + entity.columns[column].type + ' NULL';
                            }
                        }
                        sql += ', PRIMARY KEY (time, item_id));';
                        await transactionalEntityManager.query(sql);
                        this.winstonLogger.log('info', '[' + entity.tableName + '] Created table');
                        await transactionalEntityManager.query("SELECT create_hypertable('metrics." +
                            entity.tableName +
                            "','time', create_default_indexes => FALSE, chunk_time_interval => " + this.rawDataInterval.chunk_interval + ");");
                        this.winstonLogger.log('info', '[' + entity.tableName + '] Created hypertable');
                        await transactionalEntityManager.query("SELECT add_retention_policy('metrics." + entity.tableName + "', " + this.rawDataInterval.retention_policy + ");");
                        this.winstonLogger.log('info', '[' + entity.tableName + '] Created retention policy');
                        await transactionalEntityManager.query('ALTER TABLE metrics.' + entity.tableName + " SET ( timescaledb.compress, timescaledb.compress_segmentby = 'item_id' );");
                        if (this.rawDataInterval.compress_after) {
                            await transactionalEntityManager.query("SELECT add_compression_policy('metrics." + entity.tableName + "', " + this.rawDataInterval.compress_after + ");");
                            this.winstonLogger.log('info', '[' + entity.tableName + '] Created compression');
                        }
                        for (const [index, interval] of this.intervals.entries()) {
                            if (index === 0 && min5minutesInterval.includes(hwType)) {
                                continue;
                            }
                            sql = `
                                CREATE MATERIALIZED VIEW metrics.${entity.tableName}_${interval.label}
                                WITH (timescaledb.continuous) AS
                                SELECT time_bucket(INTERVAL '${interval.interval}', "${index === 0 || (index === 1 && min5minutesInterval.includes(hwType)) ? 'time' : 'bucket'}") AS bucket, item_id
                            `;
                            for (const column in entity.columns) {
                                if ((entity.columns[column].propertyName !== 'time' && entity.columns[column].propertyName !== 'bucket') && entity.columns[column].propertyName !== 'item_id') {
                                    const func = this.sumMetrics[hwType] && this.sumMetrics[hwType][subsystem] && this.sumMetrics[hwType][subsystem].includes(entity.columns[column].propertyName)
                                        ? 'SUM'
                                        : 'AVG';
                                    sql += ', ' + func + '("' + entity.columns[column].propertyName + '") as "' + entity.columns[column].propertyName + '"';
                                    if (hasMax && maxMetrics[hwType][subsystem].includes(entity.columns[column].propertyName)) {
                                        const maxFunc = index === 0
                                            ? 'AVG'
                                            : 'MAX';
                                        const sourceColumn = index === 0 || index === 1 && min5minutesInterval.includes(hwType)
                                            ? entity.columns[column].propertyName
                                            : entity.columns[column].propertyName + "_max";
                                        sql += ', ' + maxFunc + '("' + sourceColumn + '") as "' + entity.columns[column].propertyName + '_max"';
                                    }
                                }
                            }
                            let minValuesCondition = interval.minValues;
                            if (this.withoutMinValues[hwType] && this.withoutMinValues[hwType][subsystem]) {
                                minValuesCondition = undefined;
                            }
                            if (index === 0 || (index === 1 && min5minutesInterval.includes(hwType))) {
                                sql += ' FROM metrics.' + entity.tableName + ' GROUP BY 1, item_id WITH NO DATA; ';
                            }
                            else if (minValuesCondition !== undefined) {
                                sql += ' FROM metrics.' + entity.tableName + '_' + this.intervals[index - 1].label + ' GROUP BY 1, item_id HAVING count(item_id) >= ' + minValuesCondition + ' WITH NO DATA; ';
                            }
                            else {
                                sql += ' FROM metrics.' + entity.tableName + '_' + this.intervals[index - 1].label + ' GROUP BY 1, item_id WITH NO DATA;';
                            }
                            await transactionalEntityManager.query(sql);
                            this.winstonLogger.log('info', '[' + entity.tableName + '] sql: ' + sql);
                            this.winstonLogger.log('info', '[' + entity.tableName + '] Created continuous aggregate');
                            sql = `select materialization_hypertable_schema, materialization_hypertable_name from timescaledb_information.continuous_aggregates where view_schema = 'metrics' AND view_name = '${entity.tableName}_${interval.label}'`;
                            const chunks = await transactionalEntityManager.query(sql);
                            if (chunks[0]['materialization_hypertable_schema'] && chunks[0]['materialization_hypertable_name']) {
                                sql = `SELECT set_chunk_time_interval('${chunks[0]['materialization_hypertable_schema']}.${chunks[0]['materialization_hypertable_name']}', ${interval.chunk_interval});`;
                                await transactionalEntityManager.query(sql);
                                this.winstonLogger.log('info', '[' + entity.tableName + '] Edited chunk interval');
                            }
                            sql = `
                                SELECT add_continuous_aggregate_policy('metrics.${entity.tableName}_${interval.label}',
                                start_offset => INTERVAL '${interval.start_offset}',
                                end_offset => ${interval.end_offset},
                                schedule_interval => INTERVAL '${interval.schedule_interval}',
                                initial_start => '${this.getInitialTime(index)}');
                            `;
                            await transactionalEntityManager.query(sql);
                            sql = `SELECT add_retention_policy('metrics.${entity.tableName}_${interval.label}', INTERVAL '${interval.retention_policy}');`;
                            await transactionalEntityManager.query(sql);
                            this.winstonLogger.log('info', '[' + entity.tableName + '] Created continuous aggregate retention policy');
                            if (interval.compress_after) {
                                sql = `ALTER MATERIALIZED VIEW metrics.${entity.tableName}_${interval.label} set (timescaledb.compress = true);`;
                                await transactionalEntityManager.query(sql);
                                sql = `SELECT add_compression_policy('metrics.${entity.tableName}_${interval.label}', compress_after=>'${interval.compress_after}'::interval);`;
                                await transactionalEntityManager.query(sql);
                                this.winstonLogger.log('info', '[' + entity.tableName + '] Created compression');
                            }
                            if (interval.materialized) {
                                sql = `ALTER MATERIALIZED VIEW metrics.${entity.tableName}_${interval.label} set (timescaledb.materialized_only = true);`;
                                await transactionalEntityManager.query(sql);
                                this.winstonLogger.log('info', '[' + entity.tableName + '] Created materialized option');
                            }
                        }
                    });
                }
            }
            catch (e) {
                console.error("Metrics table error [" + entity.tableName + "]: " + e);
            }
        }
    }
    static getInitialTime(index) {
        const now = Date.now();
        const getOffset = () => {
            return (10 + Math.floor(Math.random() * 1000) / 100) * 60 * 1000;
        };
        switch (index) {
            case 0:
                return new Date(now + (5 * 60 * 1000)).toISOString();
            case 1:
                return new Date(now + (5 * 60 * 1000)).toISOString();
            case 2:
                return new Date(now - (now % (60 * 60 * 1000)) + (60 * 60 * 1000) + getOffset()).toISOString();
            case 3:
                return new Date(now - (now % (6 * 60 * 60 * 1000)) + (6 * 60 * 60 * 1000) + getOffset()).toISOString();
            case 4:
                return new Date(now - (now % (24 * 60 * 60 * 1000)) + (24 * 60 * 60 * 1000) + getOffset()).toISOString();
            default:
                return new Date(now).toISOString();
        }
    }
}
exports.TableGenerator = TableGenerator;
TableGenerator.rawDataInterval = {
    retention_policy: "INTERVAL '2 days'",
    chunk_interval: "INTERVAL '12 hours'",
};
TableGenerator.withoutMinValues = {
    'application': {
        'task': true
    },
    'linux': {
        'job': true
    },
    'as400': {
        'job': true
    },
    'windows': {
        'job': true,
        'vm': true,
        'server': true,
        'cluster': true,
        'fixed_disk': true
    },
    'kubernetes': {
        'node': true,
        'namespace': true,
        'pod': true,
        'container': true
    },
    'openshift': {
        'node': true,
        'project': true,
        'pod': true,
        'container': true
    },
    'timescaledb': {
        'db': true,
    },
    'aws': {
        'ec2': true,
        'ebs': true,
        'lambda': true
    }
};
TableGenerator.intervals = [
    {
        label: '1m',
        interval: '1 minute',
        start_offset: '1 hour',
        end_offset: 'NULL',
        schedule_interval: '5 minutes',
        retention_policy: '1 months',
        chunk_interval: "INTERVAL '12 hours'",
        compress_after: '12 hours',
        materialized: true,
    },
    {
        label: '5m',
        interval: '5 minutes',
        start_offset: '1 hour',
        end_offset: 'NULL',
        schedule_interval: '5 minutes',
        retention_policy: '3 months',
        chunk_interval: "INTERVAL '36 hours'",
        compress_after: '36 hours',
        materialized: true,
        minValues: 4
    },
    {
        label: '1h',
        interval: '1 hour',
        start_offset: '3 hours',
        end_offset: 'NULL',
        schedule_interval: '1 hour',
        retention_policy: '6 months',
        chunk_interval: "INTERVAL '7 days'",
        compress_after: '14 days',
        materialized: true,
        minValues: 10
    },
    {
        label: '6h',
        interval: '6 hours',
        start_offset: '12 hours',
        end_offset: 'NULL',
        schedule_interval: '6 hours',
        retention_policy: '1 year',
        chunk_interval: "INTERVAL '1 month'",
        compress_after: '2 months',
        materialized: true,
        minValues: 4
    },
    {
        label: '1d',
        interval: '1 day',
        start_offset: '2 days',
        end_offset: 'NULL',
        schedule_interval: '1 day',
        retention_policy: '3 years',
        chunk_interval: "INTERVAL '6 months'",
        compress_after: '1 year',
        materialized: true,
        minValues: 3
    },
];
TableGenerator.winstonLogger = winston.createLogger({
    format: winston.format.combine(winston.format.timestamp(), Format_1.logFormat),
    transports: [new winston.transports.File({ filename: 'logs/metrics-table.log' })],
});
TableGenerator.sumMetrics = {
    'vmware': {
        'esxi_vm': [
            'start_or_vmotion'
        ]
    },
};
//# sourceMappingURL=TableGenerator.js.map