"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.GraphData = void 0;
;
const MetricsWrapper_1 = require("../metrics/MetricsWrapper");
const Vega_1 = require("./Vega");
class GraphData {
    constructor(metricName, itemIds, start, end, timezone) {
        this.stacked = false;
        this.includeLegend = false;
        this.format = 'png';
        this.timezone = "Etc/Utc";
        this.graphTitle = undefined;
        this.objectItems = undefined;
        this.objectItemsById = new Map();
        this.labelsById = new Map();
        this.rootParents = new Map();
        this.timeDataPoints = [];
        this.metricName = metricName;
        this.itemIds = itemIds;
        this.start = start;
        this.end = end;
        this.timezone = timezone;
    }
    haveData() {
        return (this.timeDataPoints.length) ? true : false;
    }
    haveObjectItemsLoaded() {
        return (this.objectItems) ? true : false;
    }
    haveNonEmptyObjectItems() {
        return Array.isArray(this.objectItems) && (this.objectItems.length > 0);
    }
    getObjectItems() {
        return this.objectItems;
    }
    getFirstObjectItem() {
        return Array.isArray(this.objectItems) ? this.objectItems[0] : undefined;
    }
    getMetricSpecification() {
        return this.metricSpecification;
    }
    getLabels() {
        return this.labels;
    }
    getSortedLabels(singleItem = undefined) {
        if (singleItem) {
            return [{ uuid: singleItem.uuid, label: singleItem.label }];
        }
        this.timeserieItems.sort(function (a, b) {
            return b.metadata.avg - a.metadata.avg;
        });
        return this.timeserieItems.map(item => ({ uuid: item.metadata.uuid, label: item.metadata.name + ` [${this.getRootParentLabel(item.metadata.uuid)}]` }));
    }
    getTimeDataPoints(singleItem = undefined) {
        if (singleItem) {
            return this.timeDataPoints.filter(tdPoint => tdPoint.item === singleItem.uuid);
        }
        return this.timeDataPoints;
    }
    getTimeserieItems() {
        return this.timeserieItems;
    }
    getSpecificTimeserieItem(itemToFind) {
        return this.timeserieItems.find(item => item.metadata.uuid === itemToFind);
    }
    getMax() {
        return this.max;
    }
    setObjectItems(objectItems) {
        this.objectItems = objectItems;
        this.labels = [];
        for (const objectItem of objectItems) {
            this.objectItemsById.set(objectItem.item_id, objectItem);
            this.labelsById.set(objectItem.item_id, objectItem.label);
            this.labels.push(objectItem.label);
        }
    }
    setRootParents(itemsWithParents) {
        for (const item of itemsWithParents) {
            if (item.root_parent && 'item_id' in item.root_parent && 'label' in item.root_parent) {
                this.rootParents.set(item.item_id, { uuid: item.root_parent.item_id,
                    label: item.root_parent.label });
            }
        }
    }
    getRootParentLabel(itemId) {
        let parentInfo = this.rootParents.get(itemId);
        return parentInfo && 'label' in parentInfo ? parentInfo.label : "";
    }
    async fillAndProcessData(rawData, interval) {
        this.rawData = rawData;
        this.interval = interval;
        this.calculateStartAndEndForItems();
        this.makeTimeserieItems();
        const firstObjectItem = this.getFirstObjectItem();
        let ms = {};
        if (firstObjectItem && 'hw_type_entity' in firstObjectItem && firstObjectItem['hw_type_entity'].class) {
            ms = MetricsWrapper_1.MetricsWrapper.getMetric(firstObjectItem['hw_type_entity'].class, firstObjectItem.hw_type, firstObjectItem.subsystem, this.metricName);
        }
        else if (firstObjectItem) {
            ms = await MetricsWrapper_1.MetricsWrapper.getMetricFindClass(firstObjectItem.hw_type, firstObjectItem.subsystem, this.metricName);
        }
        this.setMetricSpecification(ms);
        this.makeTimeDataPoints();
        this.calculateMaxAndMin();
    }
    setMetricSpecification(metricSpecification) {
        if (metricSpecification
            && ('metric' in metricSpecification)
            && (typeof (metricSpecification.metric) === "string")) {
            this.metricSpecification = metricSpecification;
        }
    }
    calculateStartAndEndForItems() {
        const data = this.rawData;
        this.startsAndEnds = new Map();
        for (const uuid in data) {
            for (const metric in data[uuid]) {
                let current = this.startsAndEnds.get(metric);
                if (!current) {
                    let start = Object.keys(data[uuid][metric])[0];
                    let end = Object.keys(data[uuid][metric])[Object.keys(data[uuid][metric]).length - 1];
                    this.startsAndEnds.set(metric, { uuid: uuid, start: new Date(start).valueOf(), end: new Date(end).valueOf() });
                }
            }
        }
    }
    makeTimeserieItems() {
        this.timeserieItems = [];
        let newItemIds = [];
        for (const itemId of this.itemIds) {
            if (!this.rawData[itemId]) {
                continue;
            }
            for (const metric in this.rawData[itemId]) {
                if (!this.rawData[itemId][metric]) {
                    continue;
                }
                newItemIds.push(itemId);
                const label = this.labelsById.get(itemId) ?? itemId;
                const startAndEnd = this.startsAndEnds.get(metric);
                let start = null;
                let end = null;
                if (startAndEnd) {
                    start = startAndEnd.start / 1000;
                    end = startAndEnd.end / 1000;
                }
                const timeserie = [];
                for (const dataPoint in this.rawData[itemId][metric]) {
                    timeserie.push(this.rawData[itemId][metric][dataPoint]);
                }
                const countNumbers = timeserie.filter((e) => { return e !== null; }).length;
                let sum, avg, max;
                sum = timeserie.reduce((a, b) => a + b, 0);
                avg = +Number(sum / countNumbers).toFixed(2) || 0;
                max = Math.max(...timeserie);
                this.timeserieItems.push({
                    metadata: {
                        uuid: itemId,
                        name: label,
                        start: start,
                        end: end,
                        interval: this.interval,
                        avg: avg,
                        max: max,
                    },
                    timeseries: timeserie,
                });
            }
        }
        this.itemIds = newItemIds;
        this.setObjectItems((this.getObjectItems().filter(item => newItemIds.includes(item.item_id))));
        this.timeserieItems.sort((a, b) => (a.metadata.avg < b.metadata.avg ? 1 : -1));
    }
    makeTimeDataPoints() {
        if (!this.timeserieItems.length) {
            return;
        }
        let date = new Date();
        let baseline = new Date((date.toLocaleString("en-US", { timeZone: "Etc/Utc" }))).valueOf();
        let userTime = new Date((date.toLocaleString("en-US", { timeZone: this.timezone }))).valueOf();
        let offset = baseline - userTime;
        for (const item of this.timeserieItems) {
            let timePoint = item.metadata.start;
            for (const timeserieIndex in item.timeseries) {
                this.timeDataPoints.push(new Vega_1.TimeDataPoint(new Date((timePoint * 1000) + offset).toISOString(), item.timeseries[timeserieIndex], item.metadata.uuid));
                timePoint += this.interval * 60;
            }
        }
    }
    calculateMaxAndMin() {
        if (this.stacked) {
            this.findSumMaxAndMin();
        }
        else {
            this.findItemsMaxAndMin();
        }
    }
    findSumMaxAndMin() {
        let max = -Infinity;
        const sums = new Map();
        this.timeDataPoints.forEach(timeDataPoint => {
            let time = timeDataPoint.time;
            let sumSoFar = sums.get(time) ?? 0;
            sumSoFar += timeDataPoint.value;
            sums.set(time, sumSoFar);
            max = Math.max(max, sumSoFar);
        });
        this.max = max;
    }
    findItemsMaxAndMin() {
        let max = -Infinity;
        this.timeserieItems.forEach(item => {
            max = Math.max(max, item.metadata.max);
        });
        this.max = max;
    }
}
exports.GraphData = GraphData;
//# sourceMappingURL=GraphData.js.map