"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.MetricTools = exports.UNIT_PREFIX_BASE = void 0;
const pageResponse_constants_1 = require("../../common/constants/pageResponse.constants");
const subsystemMetricsResponse_constants_1 = require("../../common/constants/subsystemMetricsResponse.constants");
exports.UNIT_PREFIX_BASE = {
    base10: 'base10',
    base2: 'base2',
    noprefix: 'noprefix',
};
class MetricTools {
    static formatNumber(value, metricSpecification, addUnit = false) {
        if (!(metricSpecification && 'unit' in metricSpecification && metricSpecification.unit) || value === "-") {
            return value;
        }
        value = Number.parseFloat(value);
        if (metricSpecification.defaultPrefix) {
            value /= MetricTools.prefixToValue(metricSpecification.defaultPrefix);
        }
        const precision = metricSpecification.defaultPrecision;
        if (precision !== undefined) {
            value *= Math.pow(10, precision);
        }
        value = (value > 0) ? Math.ceil(value) : Math.floor(value);
        if (precision !== undefined) {
            value /= Math.pow(10, precision);
        }
        let result = value.toFixed(precision);
        if (addUnit) {
            const unitString = MetricTools.makeUnitString(metricSpecification.unit, metricSpecification.defaultPrefix);
            if (unitString) {
                result = `${result} ${unitString}`;
            }
        }
        return result;
    }
    static makeUnitString(unit, prefix = null) {
        let unitString = subsystemMetricsResponse_constants_1.SHORTCUT.hasOwnProperty(unit) ? subsystemMetricsResponse_constants_1.SHORTCUT[unit] : String(unit);
        if (prefix) {
            unitString = `${prefix}${unitString}`;
        }
        return unitString;
    }
    static makeLabel(metricSpecification, useDefaultPrefix) {
        if (!metricSpecification) {
            return 'Missing Specification';
        }
        let label = metricSpecification.label;
        let unit = metricSpecification && 'unit' in metricSpecification && metricSpecification.unit ? metricSpecification.unit : undefined;
        if (unit) {
            const prefix = useDefaultPrefix ? metricSpecification.defaultPrefix : null;
            let unitString = MetricTools.makeUnitString(unit, prefix);
            if (unitString) {
                label = `${label} [${unitString}]`;
            }
        }
        return label;
    }
    static prefixToValue(prefix) {
        if (!prefix) {
            return 1;
        }
        const decimalBase = 1000;
        const binaryBase = 1024;
        switch (prefix) {
            case pageResponse_constants_1.UNIT_PREFIX.kilo:
                return decimalBase;
            case pageResponse_constants_1.UNIT_PREFIX.mega:
                return Math.pow(decimalBase, 2);
            case pageResponse_constants_1.UNIT_PREFIX.giga:
                return Math.pow(decimalBase, 3);
            case pageResponse_constants_1.UNIT_PREFIX.tera:
                return Math.pow(decimalBase, 4);
            case pageResponse_constants_1.UNIT_PREFIX.peta:
                return Math.pow(decimalBase, 5);
            case pageResponse_constants_1.UNIT_PREFIX.exa:
                return Math.pow(decimalBase, 6);
            case pageResponse_constants_1.UNIT_PREFIX.zetta:
                return Math.pow(decimalBase, 7);
            case pageResponse_constants_1.UNIT_PREFIX.yotta:
                return Math.pow(decimalBase, 8);
            case pageResponse_constants_1.UNIT_PREFIX.kibi:
                return binaryBase;
            case pageResponse_constants_1.UNIT_PREFIX.mebi:
                return Math.pow(binaryBase, 2);
            case pageResponse_constants_1.UNIT_PREFIX.gibi:
                return Math.pow(binaryBase, 3);
            case pageResponse_constants_1.UNIT_PREFIX.tebi:
                return Math.pow(binaryBase, 4);
            case pageResponse_constants_1.UNIT_PREFIX.pebi:
                return Math.pow(binaryBase, 5);
            case pageResponse_constants_1.UNIT_PREFIX.exbi:
                return Math.pow(binaryBase, 6);
            case pageResponse_constants_1.UNIT_PREFIX.zebi:
                return Math.pow(binaryBase, 7);
            case pageResponse_constants_1.UNIT_PREFIX.yobi:
                return Math.pow(binaryBase, 8);
            default:
                console.warn(`MetricTools:prefixToValue Warning: No value defined for prefix ${prefix}`);
                return 1;
        }
    }
    static makeBinaryPrefix(value) {
        const binaryBase = 1024;
        switch (value) {
            case 1:
                return '';
            case binaryBase:
                return pageResponse_constants_1.UNIT_PREFIX.kibi;
            case Math.pow(binaryBase, 2):
                return pageResponse_constants_1.UNIT_PREFIX.mebi;
            case Math.pow(binaryBase, 3):
                return pageResponse_constants_1.UNIT_PREFIX.gibi;
            case Math.pow(binaryBase, 4):
                return pageResponse_constants_1.UNIT_PREFIX.tebi;
            case Math.pow(binaryBase, 5):
                return pageResponse_constants_1.UNIT_PREFIX.pebi;
            case Math.pow(binaryBase, 6):
                return pageResponse_constants_1.UNIT_PREFIX.exbi;
            case Math.pow(binaryBase, 7):
                return pageResponse_constants_1.UNIT_PREFIX.zebi;
            case Math.pow(binaryBase, 8):
                return pageResponse_constants_1.UNIT_PREFIX.yobi;
            default:
                console.warn(`MetricTools:makeBinaryPrefix Warning: No prefix defined for value ${value}`);
                return '';
        }
    }
    static getUnitPrefixBase(unit, usedIn) {
        switch (usedIn) {
            case 'graphYAxis':
                switch (unit) {
                    case subsystemMetricsResponse_constants_1.UNIT.hertz:
                    case subsystemMetricsResponse_constants_1.UNIT.bytes_per_second:
                    case subsystemMetricsResponse_constants_1.UNIT.count:
                    case subsystemMetricsResponse_constants_1.UNIT.quantity:
                    case subsystemMetricsResponse_constants_1.UNIT.packets:
                    case subsystemMetricsResponse_constants_1.UNIT.packets_per_second:
                    case subsystemMetricsResponse_constants_1.UNIT.errors_per_second:
                    case subsystemMetricsResponse_constants_1.UNIT.pcs_errors:
                    case subsystemMetricsResponse_constants_1.UNIT.crc_errors:
                    case subsystemMetricsResponse_constants_1.UNIT.crc_errors_per_second:
                        return exports.UNIT_PREFIX_BASE.base10;
                    case subsystemMetricsResponse_constants_1.UNIT.bytes:
                        return exports.UNIT_PREFIX_BASE.base2;
                    case subsystemMetricsResponse_constants_1.UNIT.IOPS:
                    case subsystemMetricsResponse_constants_1.UNIT.cpu_core:
                    case subsystemMetricsResponse_constants_1.UNIT.percent:
                    case subsystemMetricsResponse_constants_1.UNIT.millisecond:
                    default:
                        return exports.UNIT_PREFIX_BASE.noprefix;
                }
            default:
                console.warn(`MetricTools:getUnitPrefixBase Warning: Usage ${usedIn} not implemented yet`);
                return exports.UNIT_PREFIX_BASE.noprefix;
        }
    }
}
exports.MetricTools = MetricTools;
//# sourceMappingURL=MetricTools.js.map