"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const App_1 = require("./service/configuration/App");
const Format_1 = require("./service/log/Format");
const readline = require("readline");
const uuid_1 = require("uuid");
const winston = require("winston");
const Crypto_1 = require("./service/configuration/Crypto");
const publicUrl_js_1 = require("./common/publicUrl.js");
const winstonLogger = winston.createLogger({
    format: winston.format.combine(winston.format.timestamp(), Format_1.logFormat),
    transports: [
        new winston.transports.File({ filename: App_1.App.LOGS_PATH + '/verify-error.log', level: 'error' }),
        new winston.transports.File({ filename: App_1.App.LOGS_PATH + '/verify.log' }),
    ],
});
require('dotenv').config();
const fs = require('fs');
const env = '.env';
;
const TSDB_MIN_VERSION = '2.9.1';
const rl = readline.createInterface({
    input: process.stdin,
    output: process.stdout,
});
rl.on('SIGINT', () => {
    cleanUp(rl, 1);
});
process.on('SIGINT', () => {
    cleanUp(rl, 1);
});
function cleanUp(rl, ret) {
    console.log('\nExiting verify.ts...');
    winstonLogger.log("info", 'Exiting verify.ts...');
    rl.close(ret);
    process.exit(ret);
}
async function checkLogsPath() {
    const ROOT_PATH = App_1.App.ROOT_PATH;
    const LOGS_PATH = App_1.App.LOGS_PATH;
    await fs.writeFileSync(ROOT_PATH + "/server-nest/path.json", JSON.stringify({
        LOGS_PATH: LOGS_PATH
    }));
}
async function question(rl, elm, defaultVal) {
    return new Promise((resolve, reject) => {
        rl.question(elm, (answer) => {
            if (answer === "") {
                resolve(defaultVal);
            }
            resolve(answer);
        });
    });
}
async function getSection(rl, infoToGet, importance) {
    if (importance === 'required') {
        console.log('\nPlease, input the ' + importance + ' information. Default info will be entered if left empty\n');
    }
    else if (importance === 'optional') {
        console.log('Feel free to either input the ' + importance + ' information or leave blank\n');
    }
    for (const element in infoToGet) {
        let defaultVal = infoToGet[element].toString();
        let answer;
        let counter = 0;
        do {
            counter++;
            if (counter >= 10) {
                console.log("You cannot be serious about this, enter a password or press ctrl+c to end the setup.");
            }
            answer = await question(rl, element + '[' + defaultVal + ']= ', defaultVal);
        } while (counter < 12 && answer.toString() === "" && importance === "required");
        infoToGet[element] = answer.trim();
    }
    return infoToGet;
}
async function createEnv(info) {
    let appInfo = {
        APP_PORT: process.env.DOCKER && process.env.APP_PORT ? process.env.APP_PORT : (info[1].APP_PORT ? info[1].APP_PORT : 8443),
        APP_DEBUG: process.env.DOCKER && process.env.APP_DEBUG ? process.env.APP_DEBUG : false,
        APP_FORCE_MICROSERVICES: process.env.DOCKER && process.env.APP_FORCE_MICROSERVICES ? process.env.APP_FORCE_MICROSERVICES : true,
        APP_TIMEZONE: process.env.DOCKER && process.env.APP_TIMEZONE ? process.env.APP_TIMEZONE : 'Europe/Prague',
    };
    if (process.env.APP_PORT) {
        delete appInfo.APP_PORT;
    }
    let jwt = {
        JWT_SECRET: process.env.DOCKER && process.env.JWT_SECRET ? process.env.JWT_SECRET : (0, uuid_1.v4)(),
    };
    let uuids = {
        UUID_NAMESPACE: process.env.DOCKER && process.env.UUID_NAMESPACE ? process.env.UUID_NAMESPACE : (0, uuid_1.v4)(),
        UUID_BACKEND: process.env.DOCKER && process.env.UUID_BACKEND ? process.env.UUID_BACKEND : (0, uuid_1.v4)(),
        APP_SECRET: process.env.DOCKER && process.env.APP_SECRET ? process.env.APP_SECRET : (0, uuid_1.v4)(),
    };
    let database = {
        DB_USERNAME: info[0].DB_USERNAME,
        DB_HOST: info[0].DB_HOST,
        DB_PASSWORD: await Crypto_1.Crypto.encrypt(info[0].DB_PASSWORD, uuids.APP_SECRET),
        DB_PORT: info[0].DB_PORT,
        DB_DATABASE: info[0].DB_DATABASE,
    };
    let wholeArr = [].concat([appInfo, jwt, uuids, database]);
    let text = '';
    for (let section of wholeArr) {
        for (const element in section) {
            section[element];
            text += element + '=' + section[element] + '\n';
        }
        text += '\n';
    }
    await appendToFile(env, text);
}
async function migration() {
    const { spawn, exec } = require('child_process');
    console.log('Running migrations...');
    winstonLogger.log("info", 'Running migrations...');
    const child = exec('npm run typeorm:migration:run');
    process.stdin.pipe(child.stdin);
    for await (const data of child.stdout) {
        console.log(`${data}`);
        winstonLogger.log("info", `${data}`);
    }
}
async function checkEnvProperties(env, customProperties) {
    let additionalProperties = {
        NODE_ENV: "production",
        APP_SECRET: (0, uuid_1.v4)(),
    };
    if (customProperties) {
        additionalProperties = customProperties;
    }
    let text = "";
    for (let property in additionalProperties) {
        if (!(property in process.env)) {
            text += '\n' + property + '=' + additionalProperties[property] + '\n';
        }
    }
    await appendToFile(env, text);
}
async function appendToFile(file, text) {
    fs.writeFileSync(file, text, { flag: 'a+' }, (err) => {
        if (err)
            console.log(err);
        else {
            this.logger.log('Created file.');
            winstonLogger.log("info", 'Created file.');
        }
    });
}
async function writeToFile(file, text) {
    fs.writeFileSync(file, text, { flag: 'w' }, (err) => {
        if (err) {
            console.log(err);
            cleanUp(rl, 1);
        }
        else {
            this.logger.log('Created file.');
            winstonLogger.log("info", 'Created file.');
        }
    });
}
async function connectToDatabase(USER, HOST, PASSWORD, PORT, DATABASE, suppress) {
    try {
        const { Client } = require('pg');
        const client = new Client({
            user: USER,
            host: HOST,
            password: PASSWORD,
            port: PORT,
            database: DATABASE,
        });
        await client.connect();
        if (!suppress) {
            console.log(`Connection to database ${DATABASE} was successful!`);
        }
        return {
            message: 'OK',
            return_value: 1,
            dbConnection: client
        };
    }
    catch (error) {
        if (!suppress) {
            winstonLogger.log("error", '\nERROR: Database connection denied. Used credentials:\n' +
                '\nDB_HOST=' + HOST +
                '\nDB_PORT=' + PORT +
                '\nDB_USERNAME=' + USER +
                '\nDB_DATABASE=' + DATABASE +
                '\n');
            console.log('\nERROR: Database connection denied. Used credentials:\n' +
                '\nDB_HOST=' + HOST +
                '\nDB_PORT=' + PORT +
                '\nDB_USERNAME=' + USER +
                '\nDB_DATABASE=' + DATABASE +
                '\n');
        }
        let err = error.toString().split('\n');
        console.log('\n' + err[0]);
        winstonLogger.log("error", '\n' + err[0]);
        return {
            message: err[0],
            return_value: 0
        };
    }
}
async function checkForDatabase(dbCon, dbToCheck) {
    const result = await dbCon.query("SELECT datname FROM pg_catalog.pg_database where datname ='" + dbToCheck + "';");
    if (result && result.rows && result.rows.length > 0) {
        return true;
    }
    else {
        return false;
    }
}
async function checkForTable(dbCon, table) {
    try {
        await dbCon.query(`select * from public.${table};`);
        return true;
    }
    catch (e) {
        return false;
    }
}
async function checkDatabaseForSetup(dbCon) {
    const result = await dbCon.query("SELECT tablename FROM pg_catalog.pg_tables where tablename ='hostcfg';");
    if (result.rows.length >= 1) {
        return true;
    }
    else {
        return false;
    }
}
async function checkDatabaseVersion(dbCon) {
    const result = await dbCon.query("SELECT default_version, installed_version FROM pg_available_extensions where name = 'timescaledb';");
    if (result.rows[0].installed_version !== undefined && (result.rows[0].installed_version === TSDB_MIN_VERSION || (0, publicUrl_js_1.isNewerVersion)(TSDB_MIN_VERSION, result.rows[0].installed_version))) {
        return true;
    }
    else {
        console.log("Update your TimescaleDB");
        console.log("Your TSDB default version:   " + result.rows[0].default_version);
        console.log("Your TSDB installed version: " + result.rows[0].installed_version);
        console.log("Lowest supported version:    " + TSDB_MIN_VERSION);
        return false;
    }
}
async function getInfo(predefined) {
    let infoArray = [];
    if (predefined === "appport") {
        let infoBody = {
            APP_PORT: hasAccessToPrivatePorts() ? 443 : 8443,
        };
        let dbInfoSection = await getSection(rl, infoBody, 'required');
        infoArray.push(dbInfoSection);
    }
    else if (predefined == "db") {
        let infoBody = {
            DB_HOST: '127.0.0.1',
            DB_PORT: '5432',
            DB_USERNAME: 'postgres',
            DB_DATABASE: 'xormon',
            DB_PASSWORD: '',
        };
        let dbInfoSection = await getSection(rl, infoBody, 'required');
        infoArray.push(dbInfoSection);
    }
    else if (predefined == "passwdchange") {
        let infoBody = {
            DB_PASSWORD: '',
        };
        let answer = await question(rl, '\nDo you want to change the password?(y)\n', "");
        if (answer.toString().toLowerCase() === 'y') {
            let dbInfoSection = await getSection(rl, infoBody, 'required');
            infoArray.push(dbInfoSection);
        }
    }
    return infoArray;
}
const checkPublicUrl = async (publicUrl) => {
    const { setPublicUrlToUI } = await Promise.resolve().then(() => require("./common/publicUrl.js"));
    try {
        if (publicUrl !== undefined) {
            await setPublicUrlToUI(publicUrl);
        }
    }
    catch (e) {
        console.error("[PublicUrlSet] error: " + e);
        winstonLogger.log("error", "[PublicUrlSet] error: " + e);
    }
};
function hasAccessToPrivatePorts() {
    const estimated_caps = 'cap_net_bind_service=ep';
    const execSync = require('child_process').execSync;
    const nodejs_binary = execSync('which node', { encoding: 'utf-8' }).replace(/\r?\n|\r|\s/g, "");
    let effective_caps;
    try {
        effective_caps = execSync(`/usr/bin/getcap ${nodejs_binary} 2>/dev/null`, { encoding: 'utf-8' });
    }
    catch (error) {
        try {
            effective_caps = execSync(`/usr/sbin/getcap ${nodejs_binary} 2>/dev/null`, { encoding: 'utf-8' });
        }
        catch (error) {
            console.log("Couldn't find 'getcap', assuming it cannot run on privileged ports.");
            winstonLogger.log("info", "Couldn't find 'getcap', assuming it cannot run on privileged ports.");
            return false;
        }
    }
    if (!effective_caps.includes(estimated_caps)) {
        return false;
    }
    else {
        return true;
    }
}
async function createDatabase(dbCon, DB) {
    try {
        let resp = await dbCon.query('CREATE DATABASE "' + DB + '"');
        return true;
    }
    catch (error) {
        let err = error.toString().split('\n');
        console.log('\n' + err[0]);
        winstonLogger.log("error", err[0]);
        return false;
    }
}
async function dbReadyCheck(dbCon) {
    await dbCon.query('CREATE EXTENSION IF NOT EXISTS timescaledb;');
    await dbCon.query('CREATE SCHEMA IF NOT EXISTS metrics;');
    if (!await checkDatabaseVersion(dbCon)) {
        return false;
    }
    return true;
}
async function setup(dbCon, infoArray, customDbExists, envExists) {
    if (!customDbExists) {
        let dbCreated = await createDatabase(dbCon, infoArray[0].DB_DATABASE);
        if (!dbCreated) {
            return 0;
        }
    }
    let connectionStatus = await connectToDatabase(infoArray[0].DB_USERNAME, infoArray[0].DB_HOST, infoArray[0].DB_PASSWORD, infoArray[0].DB_PORT, infoArray[0].DB_DATABASE, 0);
    await dbCon.end();
    if (connectionStatus.return_value) {
        let customDbCon = connectionStatus.dbConnection;
        try {
            const { getPublicUrl, setPublicUrlToSourceMaps } = await Promise.resolve().then(() => require("./common/publicUrl.js"));
            const publicUrl = await getPublicUrl();
            if (!envExists || !await checkForTable(customDbCon, "versions")) {
                await checkEnvProperties(env, { NODE_ENV: "production" });
                checkPublicUrl(publicUrl);
            }
            await setPublicUrlToSourceMaps(publicUrl);
        }
        catch (e) {
            console.error("[setPublicUrlToSourceMaps] error: " + e);
            winstonLogger.log("error", "[setPublicUrlToSourceMaps] error: " + e);
        }
        let isReady = await dbReadyCheck(customDbCon);
        customDbCon.end();
        if (isReady) {
            if (!envExists) {
                await createEnv(infoArray);
            }
            try {
                await migration();
            }
            catch (error) {
                let err = error.toString().split('\n');
                console.log('\n' + 'Migration failed with error: ' + err[0]);
                winstonLogger.log("error", 'Migration failed with error: ' + err[0]);
                return 0;
            }
            return 1;
        }
        else {
            return 0;
        }
    }
    else {
        console.log("Couldn't connect to database " + infoArray[0].DB_DATABASE);
        winstonLogger.log("error", "Couldn't connect to database " + infoArray[0].DB_DATABASE);
        return 0;
    }
}
const checkEnv = async () => {
    let databaseInfo = {
        DB_USERNAME: null,
        DB_HOST: null,
        DB_PASSWORD: null,
        DB_PORT: null,
        DB_DATABASE: null,
    };
    let appInfo = {
        APP_PORT: null,
    };
    await checkLogsPath();
    let envExists = await fs.existsSync(env);
    if (envExists && process.env.MS_PORT_PREFIX && (+process.env.MS_PORT_PREFIX < 3 || +process.env.MS_PORT_PREFIX > 64)) {
        console.log("Microservice port prefix: " + process.env.MS_PORT_PREFIX + " is out of range (3-64)");
        winstonLogger.log("error", "Microservice port prefix: " + process.env.MS_PORT_PREFIX + " is out of range (3-64)");
        cleanUp(rl, 1);
    }
    let envOK = (envExists && process.env.DB_USERNAME);
    if (process.env.DOCKER && !process.env.APP_SECRET) {
        console.log("APP_SECRET is not configured. Can't proceed.");
        winstonLogger.log("error", "APP_SECRET is not configured. Can't proceed.");
        cleanUp(rl, 1);
    }
    if (!envOK && !(process.env.DOCKER)) {
        let appInfoArray = await getInfo("appport");
        appInfo = {
            APP_PORT: appInfoArray[0].APP_PORT,
        };
        let infoArray = await getInfo("db");
        databaseInfo = {
            DB_USERNAME: infoArray[0].DB_USERNAME,
            DB_HOST: infoArray[0].DB_HOST,
            DB_PASSWORD: infoArray[0].DB_PASSWORD,
            DB_PORT: infoArray[0].DB_PORT,
            DB_DATABASE: infoArray[0].DB_DATABASE,
        };
    }
    else {
        if (!process.env.DOCKER && (+process.env.APP_PORT < 1024 && !hasAccessToPrivatePorts())) {
            const message = `
      Application is unable to run on privileged ports(1024 and lower)

      This can be fixed by executing the commands below:

      su root
      setcap 'cap_net_bind_service=+ep' \`which node\`

      `;
            console.log(message);
            cleanUp(rl, 1);
        }
        if (!process.env.APP_SECRET) {
            let tempSecret = (0, uuid_1.v4)();
            await checkEnvProperties(env, { APP_SECRET: tempSecret });
            process.env.APP_SECRET = tempSecret;
        }
        databaseInfo = {
            DB_USERNAME: process.env.DB_USERNAME,
            DB_HOST: process.env.DB_HOST,
            DB_PASSWORD: await Crypto_1.Crypto.decrypt(process.env.DB_PASSWORD, process.env.APP_SECRET),
            DB_PORT: process.env.DB_PORT,
            DB_DATABASE: process.env.DB_DATABASE,
        };
    }
    let connectionStatus;
    if (databaseInfo.DB_PASSWORD && databaseInfo.DB_PASSWORD !== "") {
        connectionStatus = await connectToDatabase(databaseInfo.DB_USERNAME, databaseInfo.DB_HOST, databaseInfo.DB_PASSWORD, databaseInfo.DB_PORT, "postgres", 0);
    }
    if (!connectionStatus || !connectionStatus.return_value || connectionStatus.message.includes("authentication failed")) {
        connectionStatus = await connectToDatabase(databaseInfo.DB_USERNAME, databaseInfo.DB_HOST, process.env.DB_PASSWORD, databaseInfo.DB_PORT, "postgres", 0);
        let passwordInfo;
        if (connectionStatus && connectionStatus.return_value) {
            databaseInfo.DB_PASSWORD = process.env.DB_PASSWORD;
            passwordInfo = await changePassword(process.env.DB_PASSWORD, envExists);
        }
        if (!connectionStatus || !connectionStatus.return_value || connectionStatus.message.includes("authentication failed")) {
            if (!process.env.DOCKER) {
                passwordInfo = await incorrectPassword(envExists);
            }
            else {
                passwordInfo = {
                    plain: process.env.DB_PASSWORD,
                    return_value: 1
                };
            }
            if ('return_value' in passwordInfo && passwordInfo.return_value) {
                databaseInfo.DB_PASSWORD = passwordInfo.plain;
                connectionStatus = await connectToDatabase(databaseInfo.DB_USERNAME, databaseInfo.DB_HOST, databaseInfo.DB_PASSWORD, databaseInfo.DB_PORT, "postgres", 0);
            }
        }
    }
    if (connectionStatus && connectionStatus.return_value) {
        console.log('\n' + 'Looking for database ' + databaseInfo.DB_DATABASE);
        let customDbExists = await checkForDatabase(connectionStatus.dbConnection, databaseInfo.DB_DATABASE);
        let infoArrayReplacement = [];
        infoArrayReplacement[0] = databaseInfo;
        infoArrayReplacement[1] = appInfo;
        let res = await setup(connectionStatus.dbConnection, infoArrayReplacement, customDbExists, envOK);
        if (res) {
            cleanUp(rl, 0);
        }
        else {
            cleanUp(rl, 1);
        }
    }
    else {
        cleanUp(rl, 1);
    }
};
async function incorrectPassword(toFile) {
    console.log('\n' + 'Incorrect password!');
    let newInfo = await getInfo('passwdchange');
    if (newInfo.length > 0) {
        return await changePassword(newInfo[0].DB_PASSWORD, toFile);
    }
    return {
        return_value: 0,
    };
}
async function changePassword(newPassword, toFile) {
    let encrypted = await Crypto_1.Crypto.encrypt(newPassword, process.env.APP_SECRET);
    if (toFile) {
        let text = await readFile(env);
        let regex = /(?<=DB_PASSWORD=)(.*?)(?=\n)/g;
        let newText = text.replace(regex, encrypted);
        await writeToFile(env, newText);
    }
    return {
        return_value: 1,
        plain: newPassword,
        encrypted: encrypted
    };
}
async function readFile(file) {
    let text = fs.readFileSync(file, 'utf8');
    return text;
}
checkEnv();
//# sourceMappingURL=verify.js.map