package HCP;

use strict;
use warnings;
use FindBin;
use lib "$FindBin::Bin/"; # use lib in this folder
use lib "../../../perl-modules";
use LWP::UserAgent;
use HTTP::Request;
use Data::Dumper;
use Xormon;
use JSON qw(encode_json decode_json);
use Test;
use XormonLogger;
use HTTP::Date;
use MIME::Base64 qw( encode_base64 );
use Digest::MD5 qw( md5_hex );
use SNMP_lib;

my $xormon;
my $cfg;
my $logger;
my $act_time;
my $ip;
my $version;
my $run_conf;
my $token;
my $port = 9090;
my $protocol = "https";
my $debug;
my %api_data;
my $list_perf = list_perf();
my (%device_perf,%perf);
my @tenant_names;
my $sess;

sub collect {
  $xormon = shift;
  $cfg    = shift;

  $debug = exists $cfg->{'debug'} ? $cfg->{'debug'} : 0;

  if ( !exists $cfg->{hostcfg_id} ) {
    Xormon::error("Cannot continue, 'hostcfg_id' is not configured!");
    return;
  }

  $logger = XormonLogger->new('hcp', $cfg->{hostcfg_id}, $cfg->{hostalias});
  $act_time = time;
  $cfg->{timestamp} = $act_time;
  my $data = $xormon->getData();
  my $last_conf_run = exists $data->{last_conf_run} ? $data->{last_conf_run} : $act_time - 3700;

  $token = authentication_token($cfg->{apiUsername}, $cfg->{apiPassword});
  $ip = $cfg->{apiHost};
  $version = get_version();
  $run_conf = $act_time - $last_conf_run >= 3600 && exists $data->{min_perf} && @{$data->{min_perf}{collect_time}} >= 4 || $debug;
  $sess = new_session();

  if ($run_conf) {
    $data = $data || {};
    my $tenant_info = tenants_conn($cfg->{tenants});
    my @tenants_gui = keys %{$tenant_info};
    if (@tenants_gui) {
      tenants(\@tenants_gui);
      namespace($tenant_info);
    } else {
      $logger->info("No tenant in configuration.");
    }

    $data = {
      %{$data},
      last_conf_run => $act_time,
      last_hour_data => {
        device => \%device_perf,
        subsystems => \%perf,
      }
    };
  }

  my $min_perf = exists $data->{min_perf} ? $data->{min_perf} : undef;
  my $save_data = sub {
    $xormon->saveData({
      %{$data},
    });
  };

  if (!$min_perf) {
    $save_data->();
    $logger->info("Init task 'fetch' for '$cfg->{hostalias} completed.'");
    return 'successful';
  }

  if (@{$min_perf->{collect_time}} < 4 && !$debug) {
    $save_data->();
    error_debug("Failed to collect required number of data samples!");
    exit;
  }

  if (!$run_conf) {
    %perf = %{$data->{last_hour_data}{subsystems}};
    %device_perf = %{$data->{last_hour_data}{device}};
  }

  my ($node_5m, $volume_5m) = get_5min_data($min_perf);
  $save_data->();

  volume($volume_5m);
  node($node_5m, $volume_5m);

  if ($run_conf) {
    save_architecture();
    save_conf_data();
  }

  save_performance();

  return 'successful';
}

sub one_min_collect {
  $xormon = shift;
  $cfg    = shift;

  $debug = exists $cfg->{'debug'} ? $cfg->{'debug'} : 1;

  $logger = XormonLogger->new('hcp', $cfg->{hostcfg_id}, $cfg->{hostalias});
  my $data = $xormon->getData();

  my $data_perf = $data->{min_perf}{data} || {};
  my $collect_time = $data->{min_perf}{collect_time} || [];
  my (undef,$min) = localtime;

  $min = substr($min, -1);
  if ($min == 0 || $min == 5) { ## the last fifth sample is collected by fetch task
    return 'successful';
  }

  if ($min == 1 || $min == 6) { ## the last fifth sample is collected by fetch task
    $data_perf = {};
    $collect_time = []
  }

  $act_time = time;
  if (@{$collect_time} && $act_time - $collect_time->[-1] < 60) {
    sleep(60 - ($act_time - $collect_time->[-1])) if !$debug;
  }

  $act_time = time;
  $token = authentication_token($cfg->{apiUsername}, $cfg->{apiPassword});
  my $get = make_request('NODE', $token, $cfg->{apiHost}, 0)->('NODE'); ## Provide one min sample.

  push(@{$collect_time}, $act_time);
  $data_perf->{$act_time} = $get ? $get->{nodes}{node} : undef;
  $xormon->saveData({
    %{$data},
    min_perf => {
      collect_time => $collect_time,
      data => $data_perf,
    }
  });

  return 'successful';
}

sub save_performance {
  my $data2csv = sub {
    my($type, $data) = @_;

    return if !$data;

    my $data_out = "id;time";
    foreach ( @{$list_perf->{$type}{names}} ) {
      $data_out .= ";$_";
    }

    $data_out .= "\n";
    foreach (@{$data}) {
      my $hash_key = $type ne 'DEVICE' ? hash_key($type, $_->{name}) :  $cfg->{'hostcfg_id'};
      $data_out .= "$hash_key;$act_time";

      foreach my $metric ( @{$list_perf->{$type}{names}} ) {
        $data_out .= defined $_->{$metric} ? ";$_->{$metric}" : ";";
      }

      $data_out .= "\n";
      my $rc = $xormon->saveCSV($data_out,"hcp",$type);
      error("Attempt to save data of type '$type' ended with code: $rc") if $rc ne 201;
    }
  };

  $data2csv->('DEVICE', [\%device_perf]);

  foreach my $type (@{[qw( NODE VOLUME TENANT NAMESPACE )]}) {
    $data2csv->($type, $perf{$type});
  }
}

sub hash_key {
  my ($type, $label) = @_;

  return lc(Xormon::create_hash_key($cfg->{'hostcfg_id'} . uc($type) . $label));
}

sub save_architecture {
  my @architecture;
  my $st_serial = $cfg->{'hostcfg_id'};

  my $add_item = sub {
    my $type = shift;
    my $label = shift;
    my $other_props = shift || {};
    my $hash_label = shift || $label;
    my $hash_key = hash_key($type, $hash_label);

    push(@architecture, {
      "hw_type"     => "hcp",
      "hostcfg_id"  => $st_serial,
      "subsystem"   => $type,
      "item_id"     => $hash_key,
      "label"       => $label,
      "parents"     => [$st_serial],
      %{$other_props}
    });
  };

  $add_item->('device', $cfg->{'hostalias'}, {
    "item_id" => $st_serial,
    "parents" => [],
  });

  foreach my $type (@{[qw( node tenant )]}) {
    foreach (@{$perf{uc($type)}}) {
      $add_item->($type, $_->{name});
    }
  }

  foreach (@{$perf{VOLUME}}) {
    $add_item->("VOLUME", $_->{name}, {
      "parents" => [$st_serial, hash_key('NODE', $_->{node})],
    });
  }

  foreach (@{$perf{NAMESPACE}}) {
    my $name = $_->{name};
    $name =~ s/^$_->{tenant}-//;
    $add_item->("NAMESPACE", $name, {
      "parents" => [$st_serial, hash_key('TENANT', $_->{tenant})],
    }, $_->{name});
  }

  $xormon->saveArchitecture(\@architecture);
}

sub save_conf_data {
  my $st_serial = $cfg->{'hostcfg_id'};
  my %conf;
  my @health_status;

  my %hs_table = (
    'Node Status' => 'ok',
    'Node NIC Failure' => 'no',
    'Node NIC Degraded' => 'no',
    'SAN Status' => 'ok',
    'Volume Status' => 'unknown',
    'Pool Status' => 'ok',
    'Components Status' => 'ok',
    'Irreparable Objects Present' => 'no',
    "Unavailable Objects Present" => 'no'
  );

  my %snmp_data = %{get_snmp_data({
    configuration => [qw(hcpVersion serialNumber )],
    node => ["nodeNumber","nodeStatus", "nodeIP","nodeModel", "nodeNicFailure", "nodeSANStatus","nodeNicDegraded", "storageComponentStatus", "storagePoolStatus", "irreparableObjectsPresent", "unavailableObjectsPresent", "bbuBroken", "bbuChargeStatus", "bbuType", "bbuDiagnostics", "bbuDegraded"],
    alerts => [qw(hcpAdminLogEventTimestamp hcpAdminLogEventSeverity hcpAdminLogEventMessage)],
    volume => [qw(storageChannelUnit storageAvailability storageStatus storageNodeNumber)]
  })};

  ## Subsystems Health status
  my %nodes;
  for (my $index = 0; $index < @{$snmp_data{node}{nodeNumber}}; $index++) {
    foreach (keys %{$snmp_data{node}}) {
      $nodes{$snmp_data{node}{nodeNumber}[$index]}{$_} = $snmp_data{node}{$_}[$index];
    }
  }
  
  my $device_hs = "ok";
  my ($device_model, $node_name);
  my @nodes_hs;
  foreach my $node_num (keys %nodes) {
    my $node = $nodes{$node_num};
    my $node_hs = "ok";

    if ($node->{nodeIP} eq $ip) {
      $node_name = $node->{nodeNumber};
      $device_model = $node->{nodeModel};
    }

    my $node_state = { 'Label' => $node_num };
    my $state_check = sub {
      my ($table, $state_text) = @_;
      $node_hs = "error";
      $device_hs = "error";
      $$table = $state_text;
    };

    if ($node->{nodeStatus} ne "4" && $node->{nodeStatus} ne "available") {
      $state_check->(\$hs_table{'Node Status'}, 'nok');
      $node_state->{Status} = 'nok';
    } else {
      $node_state->{Status} = 'ok';
    }

    if ($node->{nodeNicFailure} ne "2" && $node->{nodeNicFailure} ne "no") {
      $state_check->(\$hs_table{'Node NIC Failure'}, 'yes');
      $node_state->{'NIC Failure'} = 'yes';
    } else {
      $node_state->{'NIC Failure'} = 'no';
    }

    if ($node->{nodeSANStatus} eq "0" || $node->{nodeSANStatus} eq "error") {
      $state_check->(\$hs_table{'SAN Status'}, 'nok');
      $node_state->{'SAN Status'} = 'crit';
    } elsif ($node->{nodeSANStatus} eq "2" || $node->{nodeSANStatus} eq "sanWithoutMultipath" || 
      $node->{nodeSANStatus} eq "4" || $node->{nodeSANStatus} eq "sanDegradedToOne" ) {
      $node_state->{'SAN Status'} = 'warn';
    } elsif ($node->{nodeSANStatus} eq "1" || $node->{nodeSANStatus} eq "rain" || 
      $node->{nodeSANStatus} eq "3" || $node->{nodeSANStatus} eq "sanGood" ) {
      $node_state->{'SAN Status'} = 'ok';
    }

    if ($node->{nodeNicDegraded} ne "2" && $node->{nodeNicDegraded} ne "no") {
       $state_check->(\$hs_table{'Node NIC Degraded'}, 'yes');
      $node_state->{'NIC Degraded'} = 'yes';
    } else {
      $node_state->{'NIC Degraded'} = 'no';
    }
     
     
    if ($node->{storageComponentStatus} ne "0" && $node->{storageComponentStatus} ne "good") {
       $state_check->(\$hs_table{'Components Status'}, 'nok');
    }
    
    if ($node->{storagePoolStatus} ne "0" && $node->{storagePoolStatus} ne "good") {
       $state_check->(\$hs_table{'Pool Status'}, 'nok');
    }

    if ($node->{irreparableObjectsPresent} ne "2" && $node->{irreparableObjectsPresent} ne "no") {
       $state_check->(\$hs_table{'Irreparable Objects Present'}, 'yes');
    }

    if ($node->{unavailableObjectsPresent} ne "2" && $node->{unavailableObjectsPresent} ne "no") {
       $state_check->(\$hs_table{"Unavailable Objects Present"}, 'yes');
    }

    $node_state->{'Volume Status'} = 'unknown';

    #volume state
    if ($node_hs ne "error" && exists $snmp_data{volume}{storageNodeNumber}) {
      $hs_table{'Volume Status'} = 'ok';
      $node_state->{'Volume Status'} = 'ok';
      for (my $index = 0; $index < @{$snmp_data{volume}{storageNodeNumber}}; $index++) {
        if ($snmp_data{volume}{storageNodeNumber}[$index] == $node->{nodeNumber} && 
          $snmp_data{volume}{storageStatus}[$index] ne '4' && $snmp_data{volume}{storageStatus}[$index] ne 'available'
        ) {
          $state_check->(\$hs_table{'Volume Status'}, 'nok');
          $node_state->{'Volume Status'} = 'nok';
          last;
        }
      }
    }

    push(@nodes_hs, $node_state);
    push(@health_status, {
      "item_id" => hash_key("NODE", $node->{nodeNumber}),
      "status" => $node_hs
    });
  }

  ## Device conf
  $conf{$st_serial} = {
    serial => $snmp_data{configuration}{serialNumber},
    version => $snmp_data{configuration}{hcpVersion},
    model => $device_model,
    node_name => $node_name,
    hs => \%hs_table,
    node_hs => encode_json(\@nodes_hs)
  };

  my %snmp_d = %{get_snmp_data({
    alerts => ["hcpAdminEvent", "hcpAdminLogEventMessage", "hcpAdminLogEventTimestamp", "hcpNotificationGroup"]
  })};

  # Alerts
  if (exists $snmp_data{alerts} ) {
    for ( my $k = 0; $k < @{ $snmp_data{alerts}{'hcpAdminLogEventSeverity'} }; $k++ ) {
      $conf{$st_serial}{alerts}[$k]{'timestamp'} = $snmp_data{alerts}{'hcpAdminLogEventTimestamp'}[$k];
      $conf{$st_serial}{alerts}[$k]{'severity'} = $snmp_data{alerts}{'hcpAdminLogEventSeverity'}[$k];
      $conf{$st_serial}{alerts}[$k]{'description'} = $snmp_data{alerts}{'hcpAdminLogEventEventMessage'}[$k];

      if ($snmp_data{alerts}{'hcpAdminLogEventSeverity'}[$k] eq "3" || $snmp_data{alerts}{'hcpAdminLogEventSeverity'}[$k] eq "error" ) {
        $device_hs = "error";
      }
    }
  } else {
    $conf{$st_serial}{alerts} = [];
  }

  $conf{$st_serial}{alerts} = encode_json(\@{$conf{$st_serial}{alerts}});


  # Volume lun_id
  foreach (@{$perf{VOLUME}}) {
    my $hash = hash_key("VOLUME", $_->{name});
    $conf{$hash}{'lun_id'} = $_->{name};
  }

  ## Configuration Node table
  my @node_table;
  my %snmp_node = %{get_snmp_data({
    node => ["nodeNumber", 'nodeIP', "nodeMgmtIP", "nodeIPv6", "nodeMgmtIPv6", "nodeDateTime", "nodeBootTime", "nodeHardwareType", "nodeModel", "nodeSANStatus"]
  })};

  my %node_key_map = (
    nodeNumber => 'Number',
    nodeIP => 'IP address',
    nodeMgmtIP => 'Management IP',
    nodeIPv6 => 'IPv6 address',
    nodeMgmtIPv6 => 'Management IPv6',
    nodeDateTime => 'Date time',
    nodeBootTime => 'Boot time',
    nodeHardwareType => 'Hardware type',
    nodeModel => 'Model',
    nodeSANStatus => 'SAN status',
  );

  for (my $index = 0; $index < @{$snmp_node{node}{nodeNumber}}; $index++) {
    my %node;
    foreach (keys %{$snmp_node{node}}) {
      $node{$node_key_map{$_}} = $snmp_node{node}{$_}[$index];
    }

    push(@node_table,\%node);
  }

  $conf{$st_serial}{nodes} = encode_json(\@node_table);

  ## Device Health status 
  push(@health_status, {
    "item_id" => $st_serial,
    "status" =>  $device_hs 
  });

  $xormon->saveConfJSON(\%conf);
  $xormon->saveStatus(\@health_status);
}

sub volume {
  my $volume_5m = shift;

  foreach my $name (keys %{$volume_5m}) {
    my ($node) = split('-',$name);
    our_metrics('VOLUME', {
      %{{name => $name, node => $node}},
      %{$volume_5m->{$name}}
    });
  }
}

sub node {
  my ($node_5m, $volume_5m) = @_;

  foreach my $name (keys %{$node_5m}) {
    my @volumes = grep {$_ =~ m/^$name/} keys %{$volume_5m};
    my %volume_metric;
    my $node_summ = sub {
      my ($metric, $add, $process) = @_;

      if (exists $volume_metric{$metric} && Xormon::isdigit($volume_metric{$metric}) && Xormon::isdigit($add)) {
        $volume_metric{$metric} = $process->();
      } elsif (Xormon::isdigit($add)) {
        $volume_metric{$metric} = $add;
      }
    };

    foreach my $volume_name (@volumes) {
      foreach my $metric (keys %{$list_perf->{VOLUME}{metric}}) {
        my $process_sum = sub { return $volume_metric{$metric} + $volume_5m->{$volume_name}{$metric}; };
        my $process_median = sub {return ($volume_metric{$metric} + $volume_5m->{$volume_name}{$metric}) / 2; };
        $node_summ->($metric, $volume_5m->{$volume_name}{$metric}, $metric ne 'lun_utilization' ? $process_sum : $process_median);
      }
    }

    our_metrics('NODE', {
      %{{name => $name}},
      %{$node_5m->{$name}},
      %volume_metric
    });
  }
}

sub tenants {
  my $tenants_gui = shift;
  my $type = 'TENANT';
  my $call = make_request($type);
  my $get_names = $call->("$type-LIST");
  endJob('Failed to obtain list of tenants.') if !$get_names;

  foreach my $tenant (@{ $get_names->{'name'} }) {
    push( @tenant_names, $tenant ) if grep {$tenant eq $_} @{$tenants_gui};
  }

  endJob('No matching tenant') if scalar @tenant_names == 0;

  foreach my $tenant (@tenant_names) {
    my $get_report = $call->("$type-REPORT", $tenant);
    my $get_info = $call->("$type-INFO", $tenant);
    next if !$get_report || !$get_info;

    make_data($type, {%{$get_report->{chargebackData}[0]}, %{$get_info}}, {%{$list_perf->{$type}{report}}, %{$list_perf->{$type}{config}}});
  }
}

sub namespace {
  my $tenant_info = shift;
  my $type = "NAMESPACE";
  foreach my $tenant (@tenant_names) {
    my $call = make_request($type, $tenant_info->{$tenant}{token}, $tenant_info->{$tenant}{ip});
    my $get_names = $call->("$type-LIST", $tenant);
    if (!$get_names) {
      error_debug("No namespaces list for tenant '$tenant'");
      next;
    }

    foreach my $namespace (@{$get_names->{name}}) {
      my $get_report = $call->("$type-REPORT", $tenant, $namespace);
      my $get_info = $call->("$type-INFO", $tenant, $namespace);
      next if !$get_report || !$get_info;

      make_data($type, {%{$get_report->{chargebackData}[0]}, %{$get_info}}, {%{$list_perf->{$type}{report}}, %{$list_perf->{$type}{config}}});
    }
  }
}

sub connection_test {
  my $client = shift;
  $cfg = shift;

  my $test = Test->new($client);

  my @variables = (qw(hostcfg_id hostalias snmpHost apiUsername apiPassword apiHost tenants ));
  foreach my $variable (@variables){
    unless ( exists $cfg->{$variable} ){
      $test->addSection("Validation of variables", 0 , "Cannot continue, $variable is not configured!");
      return $test->stop();
    }
  }

  $test->addSection("Validation of variables", 1 , "");
  $ip = $cfg->{apiHost};

  my @ports_state;
  my $code_global = 1;
  my $test_conn = sub {
    my $ip = shift;
    my $port = shift;
    my $conntest = shift || \&Xormon::conntest_tcp;

    my ($code, $text) = $conntest->($ip,$port);
    $code_global = 0 if !$code;
    push(@ports_state, "$text\n");
  };

  $test_conn->($cfg->{apiHost}, $port);

  foreach my $tenant (@{$cfg->{tenants}}) {
    $test_conn->($tenant->{host}, $port);
  }

  $test_conn->($cfg->{snmpHost}, $cfg->{snmpPort}, \&Xormon::conntest_udp);

  $test->addSection("API host network connection", $code_global, \@ports_state);

  return $test->stop() if !$code_global;

  $token = authentication_token($cfg->{apiUsername}, $cfg->{apiPassword});
  my ($version_conn, $serial) = get_version(1);

  if (!$version_conn) {
    $test->addSection("API data $cfg->{hostalias}", 0, "ERROR: Couldn't get API data");
  } else {
    $test->addSection("API data $cfg->{hostalias}", 1, [
      "Name : $cfg->{hostalias}\n",
      "Serial : $serial\n",
      "Version : $version_conn\n",
    ]);
  }

  my $tenant_info = tenants_conn($cfg->{tenants});
  my @tenants_gui = keys %{$tenant_info};

  if (@tenants_gui) {
    my $get_names = make_request('TENANT')->('TENANT-LIST', undef, undef, 1);
    if ($get_names) {
      my @tenant_names;
      foreach my $tenant (@{ $get_names->{'name'} }) {
        push( @tenant_names, $tenant ) if grep {$tenant eq $_} @tenants_gui;
      }
  
      foreach my $tenant (@tenant_names) {
        my $get = make_request("NAMESPACE", $tenant_info->{$tenant}{token}, $tenant_info->{$tenant}{ip})->("NAMESPACE-LIST", $tenant, undef, 1);
        if (!$get_names) {
          $test->addSection("API data for tenant $tenant", 0, "ERROR: Couldn't get API data");
          continue;
        }
  
        $test->addSection("API data for tenant $tenant", 1, "API data obtainable.");
      }
    }
  } else {
    $test->addSection("Tenant test", 0, "Add at least one tenant to configuration.");
  }

  $sess = new_session();
  my $snmp_data = {};
  my ( $status, $reason ) = SNMP_lib::collect_hcp( $snmp_data, $sess, 'nodeIP', undef, 1 );

  if ( $status ) {
    $test->addSection("SNMP data", 1, SNMP_lib::build_snmpwalk_cmd($sess));
  } else {
    $test->addSection("SNMP data", 0, [$reason . "\n",SNMP_lib::build_snmpwalk_cmd($sess)]);
  }

  return $test->stop();
}

sub tenants_conn {
  my $tenants = shift;
  my %tenant_info;

  ## make tentants tokens
  foreach (@{$tenants}) {
    $tenant_info{$_->{alias}} = {
      ip => $_->{host},
      token => authentication_token($_->{username}, $_->{password})
    };
  }

  return \%tenant_info;
}

sub make_data {
  my ($type, $raw_data, $list) = @_;
  my $result = {};

  foreach (keys %$list) {
    my $api = $list->{$_}{api};
    my $name = $list->{$_}{name};
    if (!exists $raw_data->{$api}) {
      error_debug("Api key '$api' of type '$type' doesn`t exist");
      $result->{$name} = undef;
      next;
    }

    $result->{$name} = $raw_data->{$api};
  }

  our_metrics($type, $result);
}

sub get_5min_data {
  my $min_perf = shift;
  my $node_5m = {};
  my $volume_5m = {};

  ## get last data sample
  my $collect_time = time;
  if (@{$min_perf->{collect_time}} && $act_time - $min_perf->{collect_time}[-1] < 60) {
    my $num = 60 - ($act_time - $min_perf->{collect_time}[-1]);
    sleep(60 - ($act_time - $min_perf->{collect_time}[-1])) if !$debug;
  }

  $collect_time = time;
  push(@{$min_perf->{collect_time}}, $collect_time);
  my $get = make_request('NODE', undef, undef, 0)->('NODE'); ## Provide one min sample.
  $min_perf->{data}{$collect_time} = $get ? $get->{nodes}{node} : undef;

  my $process = sub {
    my ($type, $obj, $name, $ref_5m) = @_;

    foreach my $metric ( keys %{$list_perf->{$type}{metric}} ) {
      my $metric_api = $list_perf->{$type}{metric}{$metric}{api};
      if (Xormon::isdigit($obj->{$metric_api})) {
        if ($ref_5m->{$name}{$metric}){
          $ref_5m->{$name}{$metric} = ($ref_5m->{$name}{$metric} + $obj->{$metric_api}) / 2;
        } else {
          $ref_5m->{$name}{$metric} = $obj->{$metric_api};
        }
      }
    }
  };

  for ( my $index = @{$min_perf->{collect_time}}; $index > @{$min_perf->{collect_time}} - 5; $index--) {
    my $nodes = $min_perf->{data}{$min_perf->{collect_time}[$index-1]};
    next if !$nodes;

    foreach my $node (@{$nodes}) {
      $process->('NODE', $node, $node->{nodeNumber}, $node_5m);

      foreach my $volume (@{$node->{volumes}{volume}}) {
        my $volume_name = $node->{nodeNumber} . "-" . $volume->{id};
        $process->('VOLUME', $volume, $volume_name, $volume_5m);
      }
    }
  }

  return ($node_5m, $volume_5m);
}

sub device_metric {
  my ($metric, $add, $process) = @_;

  if (exists $device_perf{$metric} && Xormon::isdigit($device_perf{$metric})) {
    $device_perf{$metric} = $process->() if Xormon::isdigit($add);
  } else {
    $device_perf{$metric} = $add if Xormon::isdigit($add);
  }
};

sub our_metrics {
  my ($type, $data) = @_;

  my $valid_data = sub { 
    my $metrics = shift;
    foreach (@{$metrics}) {
      my $ok = Xormon::isdigit($data->{$_}) ? 1 : 0;

      return 0 if !$ok;
    }

    return 1;
  };

  if ($type eq "NODE" || $type eq "VOLUME" ) {
    $data->{capacity_used} = $valid_data->(['capacity_total', 'capacity_free']) ? $data->{capacity_total} - $data->{capacity_free} : undef;
    $data->{capacity_used_percent} = $valid_data->(['capacity_total', 'capacity_free']) ? $data->{capacity_used} / $data->{capacity_total} * 100 : undef;
    $data->{lun_data_rate} = $valid_data->(['lun_read', 'lun_write']) ? $data->{lun_read} + $data->{lun_write} : undef;
  }

  if ($type eq "NODE") {
    $data->{network_rate} = $valid_data->(['network_tx', 'network_rx']) ? $data->{network_tx} + $data->{network_rx} : undef;
    $data->{network_rate_b} = $valid_data->(['network_tx_b', 'network_rx_b']) ? $data->{network_tx_b} + $data->{network_rx_b} : undef;

    foreach (@{[qw( network_tx network_rx network_rate network_tx_b network_rx_b network_rate_b lun_read lun_write lun_data_rate capacity_used capacity_free capacity_total capacity_used_percent swap_out )]}) {
      my $metric = sub { return $device_perf{$_} + $data->{$_}; };
      device_metric($_, $data->{$_}, $metric);
    }

    foreach (@{[qw( cpu_user cpu_system cpu_wait lun_utilization capacity_used_percent )]}) {
      my $metric = sub { return ($device_perf{$_} + $data->{$_}) / 2; };
      device_metric($_, $data->{$_}, $metric);
    }
  }

  if ($type eq 'TENANT' || $type eq 'NAMESPACE') {
    if ($data->{hard_quota}) {
      my $unit = $data->{hard_quota};
      $unit =~ s/^\d+\.\d+\s//;
      $data->{hard_quota} =~ s/\s\w+$//;
      $data->{hard_quota} =~ s/\,//g;
      $data->{hard_quota} *= $unit eq 'TB' ? 2**40 : 2**30;

      ## incoming soft_quota from api is percentage => convert to bytes
      $data->{soft_quota} =  $valid_data->(['soft_quota']) ? $data->{hard_quota} * ($data->{soft_quota} / 100) : undef;
      $data->{capacity_free} = $valid_data->(['size']) ? $data->{hard_quota} - $data->{size} : undef;
    }

    $data->{io_rate} = $valid_data->(['read_io', 'write_io']) ? $data->{read_io} + $data->{write_io} : undef;
    $data->{data_rate} = $valid_data->(['read', 'write']) ? $data->{read} + $data->{write} : undef;
  }

  if ($type eq 'TENANT') {
    foreach (@{[qw( object_total capacity_ingested )]}) {
      my $metric = sub { return $device_perf{$_} + $data->{$_}; };
      device_metric($_, $data->{$_}, $metric);
    }
  }

  push(@{$perf{$type}}, $data);
}

sub authentication_token {
  my ( $usr, $passwd ) = @_;
  $usr    = encode_base64($usr);
  $passwd = md5_hex($passwd);
  chomp($usr);

  return "HCP " . $usr . ":" . $passwd;
}

sub address {
  my ( $api_ip, $type, $line, $parm1, $parm2 ) = @_;

  my $url = "$protocol://$api_ip:$port/mapi";

  if    ( $type eq 'NODE' ) { $url .= '/nodes/statistics'; }
  elsif    ( $type eq 'VERSION' ) { $url .= '/storage/licenses'; }
  elsif    ( $type eq 'TENANT-LIST' ) { $url .= '/tenants'; }
  elsif    ( $type eq 'TENANT-REPORT' ) { $url .= "/tenants/$parm1/chargebackReport" . report_time(); }
  elsif    ( $type eq 'TENANT-INFO' ) { $url .= "/tenants/$parm1"; }
  elsif    ( $type eq 'NAMESPACE-LIST' ) { $url .= "/tenants/$parm1/namespaces"; }
  elsif    ( $type eq 'NAMESPACE-REPORT' ) { $url .= "/tenants/$parm1/namespaces/$parm2/chargebackReport" . report_time(); }
  elsif    ( $type eq 'NAMESPACE-INFO' ) { $url .= "/tenants/$parm1/namespaces/$parm2"; }
  else {  error_debug("Unknown API address '$type'!", $line); }

  return $url;
}

sub report_time {
  my $sec_t = time;
  my ( $sec, $min, $hour, $mday, $mon, $year ) = localtime($sec_t);
  $sec_t -= $sec + 60 * $min;
  ( $sec, $min, $hour, $mday, $mon, $year ) = time_form($sec_t - 1);
  my $end = "&end=$year-$mon-$mday" . "T$hour:59:59-0000";

  ( $sec, $min, $hour, $mday, $mon, $year ) = time_form( $sec_t - 3600 );
  my $start = "?start=$year-$mon-$mday" . "T$hour:00:00-0000";

  return $start . $end . "&granularity=hour";
}

sub time_form {
  my ($sec, $min, $hour, $mday, $mon, $year) = gmtime(shift);

  $year += 1900;
  $mon + 1 < 13 ? $mon += 1 : ( $mon = 1 );

  $mon  = sprintf("%02d", $mon);
  $mday = sprintf("%02d", $mday);
  $hour = sprintf("%02d", $hour);
  $min  = sprintf("%02d", $min);

  return ($sec, $min, $hour, $mday, $mon, $year);
}

sub new_session {
  SNMP_lib::initialize_snmp($cfg);
  SNMP_lib::initialize_mibs('HCP');
  return SNMP_lib::new_session( { 'Timeout' => 1000000 } );
}

sub get_snmp_data {
  my $mibs = shift || {};
  my $inventory_data = {};

  for my $section (keys %{$mibs}) {
    for my $mib (@{$mibs->{$section}}) {
      SNMP_lib::collect_hcp($inventory_data, $sess, $mib, $section);
    }
  }

  return $inventory_data;
}

sub make_ua {
  return LWP::UserAgent->new(
    ssl_opts => {
      SSL_cipher_list => 'DEFAULT:!DH',
      verify_hostname => 0,
      SSL_verify_mode => 0
    }
  );
}

sub new_request {
  my $url = shift;

  return HTTP::Request->new('GET', $url, [
    'Content-Type' => 'application/json',
    'Accept'       => 'application/json',
  ]);
}

sub make_request {
  my $type = shift;
  my $api_token = shift || $token;
  my $api_ip = shift || $ip;
  my $history_peek = shift || 1;

  return sub {
    my (undef, undef, $line) = caller;
    my ( $api_type, $parm1, $parm2, $print_err ) = @_;
    my $url = address($api_ip, $api_type, $line, $parm1, $parm2);
    return request_get($api_token, $url, $type, $line, $print_err, $history_peek);
  };
}

sub request_get {
  my ( $api_token, $url, $type, $line, $print_err, $history_peek ) = @_;

  return $api_data{$url} if exists $api_data{$url} && $history_peek; ## same api call has been done already

  my $ua = make_ua();
  my $req = new_request($url);
  $req->header( 'Authorization' => $api_token );

  my $res = $ua->request($req);
  if ($res->is_error) {
    api_error("Unable to get data type $type from: $url", $line, $res->content, $print_err);
    return;
  }

  my $data = decode_json($res->content);
  $api_data{$url} = $data;

  return $data;
}

sub get_version {
  my $hs_check = shift || 0;
  my (undef, undef, $line) = caller;
  my $url = address($ip, 'VERSION', $line);
  my $ua = make_ua();
  my $req = new_request($url);
  $req->header( 'Authorization' => $token );

  my $res = $ua->request($req);
  if ($res->is_error) {
    if (!$hs_check) {
      $xormon->isConnected(0, "auth failed");
      $xormon->hasData(0, "Unable to get API data");
      endJob('Unable to obtain HCP version!', $line); ## exit
    } else {
      return 0;
    }
  }

  if (!$hs_check) {
    $xormon->isConnected(1);
    $xormon->hasData(1);
  }

  my $result = $res->header('x-hcp-softwareversion');
  $result =~ s/\.\d+$//;
  known_version_check($result) if $debug;

  return $result if !$hs_check;

  my $content = decode_json( $res->content );
  my $serial = exists $content->{'license'}[0]{'serialNumber'} ?  $content->{'license'}[0]{'serialNumber'} : 'unknown';

  return ($result, $serial);
}

sub known_version_check {
  my $version = shift;
  my $new;

  # always change values in array on latest version
  my @latest_version = ( 9, 4, 0 );
  my @version_check  = split( /\./, $version );

  for ( my $k = 0; $k < @latest_version; $k++ ) {
    if    ( $latest_version[$k] < $version_check[$k] ) { $new = 1; last; }
    elsif ( $latest_version[$k] > $version_check[$k] ) { return; }
  }

  if ($new) {
    my $last = join('.', @latest_version);
    $logger->info("New version $last -> $version. Consider testing.");
  }
}

sub api_error {
  my ($msg, $line, $body, $print_error) = @_;

  return if $print_error;

  error($msg, $line);
  $logger->error($body);
}

sub error {
  my (undef, $file, $call_line) = caller;
  my $msg = shift;
  my $line = shift || $call_line;

  $logger->error("$msg : $file : $line\n\n");
}

sub error_debug {
  return if !$debug;

  my (undef, $file, $call_line) = caller;
  my $msg = shift;
  my $line = shift || $call_line;

  $logger->error("$msg: $file : $line\n\n");
}

sub endJob {
  my (undef, $file, $call_line) = caller;

  my $msg = shift;
  my $line = shift || $call_line;
  error($msg, $line);
  $xormon->endJob('fetch', $cfg->{hostalias}, 'error');
  exit;
}

sub list_perf {
  return {
    DEVICE => {
      names => [qw(
        cpu_user cpu_system cpu_wait 
        lun_utilization lun_read lun_write lun_data_rate
        capacity_used capacity_free capacity_total capacity_used_percent
        network_tx network_rx network_rate
        network_tx_b network_rx_b network_rate_b
        swap_out
        object_total capacity_ingested
      )],
    },
    NODE => {
      names => [qw(
        cpu_user cpu_system cpu_wait
        lun_utilization lun_read lun_write lun_data_rate
        capacity_used capacity_free capacity_total capacity_used_percent
        network_tx network_rx network_rate
        network_tx_b network_rx_b network_rate_b
        swap_out
      )],
      metric => {
        cpu_user => { name => 'cpu_user', api => 'cpuUser' },
        cpu_system  => { name => 'cpu_system', api => 'cpuSystem' },
        cpu_wait  => { name => 'cpu_wait', api => 'ioWait' },
        swap_out  => { name => 'swap_out', api => 'swapOut' },
        network_tx  => { name => 'network_tx', api => 'frontEndBytesRead' },
        network_rx  => { name => 'network_rx', api => 'frontEndBytesWritten' },
        network_tx_b  => { name => 'network_tx_b', api => 'backEndBytesRead' },
        network_rx_b => { name => 'network_rx_b', api => 'backEndBytesWritten' },
        ## lun_utilization, lun_read, lun_write, lun_data_rate, capacity_free, capacity_total => agr from volume
        ## capacity_used_percent, lun_data_rate, network_rate, network_rate_b => calc
      },
    },
    VOLUME => {
      names => [qw(
        lun_utilization lun_read lun_write lun_data_rate
        capacity_used capacity_free capacity_total capacity_used_percent
      )],
      metric => {
        lun_utilization  => { name => 'lun_utilization', api => 'diskUtilization' },
        lun_read  => { name => 'lun_read', api => 'blocksRead' },
        lun_write  => { name => 'lun_write', api => 'blocksWritten' },
        capacity_free  => { name => 'capacity_free',  api => 'freeBytes' },
        capacity_total => { name => 'capacity_total', api => 'totalBytes' }
        ## capacity_used, capacity_used_percent, lun_data_rate => calc
      }
    },
    TENANT => {
      names => [qw(
        read write data_rate
        read_io write_io io_rate
        object_total capacity_ingested
        size soft_quota hard_quota capacity_free
      )],
      report => { ## REPORT
        read => { name => 'read',  api => 'bytesOut' },
        write => { name => 'write',  api => 'bytesIn' },
        read_io => { name => 'read_io',  api => 'reads' },
        write_io => { name => 'write_io',  api => 'writes' },
        object_total => { name => 'object_total',  api => 'objectCount' },
        capacity_ingested => { name => 'capacity_ingested',  api => 'ingestedVolume' },
        size => { name => 'size',  api => 'storageCapacityUsed' },
      },
      config => {
        name => { name => 'name', api => 'name' },
        soft_quota => {name => 'soft_quota', api => 'softQuota' },
        hard_quota => {name => 'hard_quota', api => 'hardQuota' },
      }
    },
    NAMESPACE => {
      names => [qw(
        read write data_rate
        read_io write_io io_rate
        object_total capacity_ingested
        size soft_quota hard_quota capacity_free
      )],
      report => { ## REPORT
        read => { name => 'read',  api => 'bytesOut'},
        write => { name => 'write',  api => 'bytesIn'},
        read_io => { name => 'read_io',  api => 'reads'},
        write_io => { name => 'write_io',  api => 'writes'},
        object_total => { name => 'object_total',  api => 'objectCount'},
        capacity_ingested => { name => 'capacity_ingested',  api => 'ingestedVolume'},
        size => { name => 'size',  api => 'storageCapacityUsed'},
        tenant => { name => 'tenant',  api => 'tenantName'},
      },
      config => {
        name => { name => 'name', api => 'name' },
        soft_quota => {name => 'soft_quota', api => 'softQuota' },
        hard_quota => {name => 'hard_quota', api => 'hardQuota' },
      }
    },
  };
}

1;
