use LWP;
use JSON;
use strict;
use warnings;
use Data::Dumper;
use FindBin;                
use lib "$FindBin::Bin/"; # use lib in this folder
use lib "../../../perl-modules";
use Test;
use Kubernetes;
use Xormon;

$SIG{CHLD} = 'IGNORE';

use IO::Socket qw(AF_INET AF_UNIX SOCK_STREAM SHUT_WR);

my $DEFAULT_PORT = 3601;
my $port = Xormon::checkPrefixPort($DEFAULT_PORT);

my $server = IO::Socket->new(
    Domain => AF_INET,
    Type => SOCK_STREAM,
    Proto => 'tcp',
    LocalHost => '0.0.0.0',
    LocalPort => $port,
    ReusePort => 1,
    Listen => 5,
) || die "Can't open socket: $@";
Xormon::log("Service listening on port $port");

# waiting for a new client connection
my $client;
while ($client = $server->accept()) {
    
    my $pid;
    unless ( defined( $pid = fork() ) ) {
        Xormon::error("Wait for fork");
        sleep 5;
    }

    if ($pid == 0) {

        # get information about a newly connected client
        my $client_address = $client->peerhost();
        my $client_port = $client->peerport();
        Xormon::log("Connection from $client_address:$client_port");

        # received data template 
        #
        # { 
        #   "jwt"  : "<backendToken>",
        #   "task" : "<taskName>",
        #   "data" : {
        #     <configuration_from_hostcfg>
        #   }
        # }

        # read up to 65536 characters from the connected client
        my $data = "";
        $client->recv($data, 65536);

        my $decoded = from_json($data);
        my $cfg = $decoded->{data};

        Xormon::log("Received task: ".$decoded->{task});
        my $xormon = Xormon->new("$client_address", $decoded->{jwt}, 'kubernetes', $cfg->{hostcfg_id}, $cfg->{hostalias});
        my $response = '';
        if ($decoded->{task} eq "fetch") {

            $client->send("Task successfuly started");
            $client->shutdown(SHUT_WR);

            # target destinations check
            if (defined $decoded->{'destinations'}) {
                $xormon->setDestinations($decoded->{'destinations'});
            }

            if (ref($decoded->{data}) eq 'ARRAY') {
                for (@{$decoded->{data}}) { 
                    my $cfg = $_;
                    process($xormon, $cfg);
                } 
            } else {
                process($xormon, $cfg);
            }
        } elsif ($decoded->{task} eq "test") {

            apitest($client, $decoded->{data});

         } elsif ($decoded->{task} eq "namespaces") {

            $response = namespaces($decoded->{data});
            
            $client->send($response);
            $client->shutdown(SHUT_WR);

        } else {
            $response = "undefined task";

            $client->send($response);
            $client->shutdown(SHUT_WR);
        }

        exit();

    } else {
        $client->autoflush(1);    
    }
}

sub process {
    my $xormon = shift;
    my $cfg = shift;

    $xormon->startJob('fetch', $cfg->{hostalias}, $$);

    my $connected = Test::connTest($cfg->{host}, $cfg->{port});
    if ($connected == 1) {
        $xormon->isConnected(1);
    } else {
        $xormon->isConnected(0, "TCP connection to " . $cfg->{host} . ":" . $cfg->{port} . " has failed!");
        $xormon->endJob('fetch', $cfg->{hostalias}, "error", $$);
        return ();
    }

    # conf data
    my $rc = configuration($xormon, $cfg);
    
    # status code
    my $status;
    if ($rc eq "201") {
        $status = "successful";
    } elsif ($rc eq "401") {
        $status = "unauthorized";
    } else {
        $status = "error";
    }

    $xormon->endJob('fetch', $cfg->{hostalias}, $status, $$);

    Xormon::log("Task [fetch] for ".$cfg->{hostalias}." is done");

    # perf data background task
    background($xormon, $cfg, $$);
}

sub background {
    my $xormon = shift;
	my $data = shift;
    my $pid = shift;

    my $service_data = $xormon->getData();
    my $last = (defined $service_data->{pid}) ? $service_data->{pid} : ();
    if (defined $last) {
        my $exists = kill 0, $last;
        if ($exists) {
            eval { system("kill -9 $last"); Xormon::log("Killed background worker PID $last") }
        }
    }

    my $skipContainers = defined $data->{skipContainer} && $data->{skipContainer} eq "true" ? 1 : 0;
    my $kubernetes = Kubernetes->new($data->{hostalias}, $data->{host}, $data->{token}, $data->{protocol}, $data->{hostcfg_id}, $skipContainers, $data->{namespacesSelected});
    my $resolution = $kubernetes->metricResolution();
    $resolution = ($resolution->{node} gt $resolution->{pod}) ? $resolution->{node} : $resolution->{pod};

    my %labelToId;
    my $conf = $kubernetes->getConfiguration();

    for my $node_id (keys %{$conf->{label}{node}}) {
        $labelToId{$conf->{label}{node}{$node_id}} = $node_id;
    }
    for my $pod_id (keys %{$conf->{label}{n_pod}}) {
        $labelToId{$conf->{label}{n_pod}{$pod_id}} = $pod_id;
    }
    for my $container_id (keys %{$conf->{label}{container}}) {
        $labelToId{$container_id} = Xormon::create_hash_key_long($container_id);
    }

    my $counters = defined $service_data->{counters} ? $service_data->{counters} : ();

    my $limit = 20 * ( 60 / $resolution );
    for ( 1 .. $limit ) {
        eval {
            Xormon::log("[BackgroundWorker]: start");
            my $perf = $kubernetes->getMetricsData($resolution);
            if ($perf) {
                my $csv = '';

                $xormon->hasData(1);

                # Counter obj
                # "id": {
                #   "metric": {
                #     "time": "time",
                #     "value": "value"
                #   }
                # }

                # Nodes
                my $nodes         = "item_id;time;cpu_usage;memory_usage;cpu;cpu_allocatable;cpu_capacity;memory;memory_allocatable;memory_capacity;pods;pods_allocatable;pods_capacity;container_fs_writes;container_fs_reads;container_fs_writes_bytes;container_fs_reads_bytes;container_fs_write_latency;container_fs_read_latency;container_network_receive_bytes;container_network_transmit_bytes;container_fs_usage_bytes;\n";
                my @nodes_metrics = ( 'cpu', 'cpu_allocatable', 'cpu_capacity', 'memory', 'memory_allocatable', 'memory_capacity', 'pods', 'pods_allocatable', 'pods_capacity', 'container_fs_writes_total', 'container_fs_reads_total', 'container_fs_writes_bytes_total', 'container_fs_reads_bytes_total', 'container_fs_write_seconds_total', 'container_fs_read_seconds_total', 'container_network_receive_bytes_total', 'container_network_transmit_bytes_total', 'container_fs_usage_bytes' );

                foreach my $node ( keys %{ $perf->{node} } ) {
                    if (!defined $labelToId{$node}) {
                        next;
                    }
                    foreach my $time ( keys %{ $perf->{node}{$node} } ) {
                    $nodes .= $labelToId{$node} . ";" . $time . ";";
                    
                    # cpu usage
                    if (defined $perf->{node}{$node}{$time}{'cpu_capacity'} && defined $perf->{node}{$node}{$time}{'cpu'}) {
                        $nodes .= $perf->{node}{$node}{$time}{'cpu'} / ($perf->{node}{$node}{$time}{'cpu_capacity'} / 100);
                        $nodes .= ";"
                    } else {
                        $nodes .= ";"
                    }

                    # memory usage
                     if (defined $perf->{node}{$node}{$time}{'memory_capacity'} && defined $perf->{node}{$node}{$time}{'memory'}) {
                        $nodes .= $perf->{node}{$node}{$time}{'memory'} / ($perf->{node}{$node}{$time}{'memory_capacity'} / 100);
                        $nodes .= ";"
                    } else {
                        $nodes .= ";"
                    }

                    for my $metric (@nodes_metrics) {
                        if ( defined $perf->{node}{$node}{$time}{$metric} ) {
                            if ($metric =~ m/_total/) {
                                # latency seconds to ms
                                if ($metric eq "container_fs_write_seconds_total" || $metric eq "container_fs_read_seconds_total") {
                                    $perf->{node}{$node}{$time}{$metric} = $perf->{node}{$node}{$time}{$metric} * 1000;
                                }

                                # counter value check
                                if (defined $counters->{$node} 
                                    && defined $counters->{$node}{$metric} 
                                    && $perf->{node}{$node}{$time}{$metric} > $counters->{$node}{$metric}{'value'}
                                    ) {

                                    $nodes .= (($perf->{node}{$node}{$time}{$metric} - $counters->{$node}{$metric}{'value'}) / ($time - $counters->{$node}{$metric}{'time'}));
                                    $nodes .= ";";
                                } else {
                                    $nodes .= ";";
                                }
                                
                                my %counterObj = (
                                    "time" => $time,
                                    "value" => $perf->{node}{$node}{$time}{$metric}
                                );
                                $counters->{$node}{$metric} = \%counterObj;
                            } elsif($metric =~ m/memory/) {
                                $nodes .= $perf->{node}{$node}{$time}{$metric} * 1024 * 1024;
                                $nodes .= ";";
                            } else {
                                $nodes .= $perf->{node}{$node}{$time}{$metric} . ";";
                            } 
                        }
                        else {
                            $nodes .= ";";
                        }
                    }
                    $nodes .= "\n";
                    }
                }

                $xormon->saveCSV($nodes, 'kubernetes', 'node');

                # Pods & Containers
                my $pods         = "item_id;time;cpu;cpu_limit;cpu_request;memory;memory_limit;memory_request;container_network_receive_bytes;container_network_transmit_bytes;container_processes;container_fs_writes;container_fs_reads;container_fs_writes_bytes;container_fs_reads_bytes;\n";
                my @pods_metrics = ( 'cpu', 'cpu_limit', 'cpu_request', 'memory', 'memory_limit', 'memory_request', 'container_network_receive_bytes_total', 'container_network_transmit_bytes_total', 'container_processes' );
                my @tmp_pods_metrics = ( 'container_fs_writes_total', 'container_fs_reads_total', 'container_fs_writes_bytes_total', 'container_fs_reads_bytes_total' );

                my $containers         = "item_id;time;cpu;cpu_limit;cpu_request;memory;memory_limit;memory_request;container_fs_writes;container_fs_reads;container_fs_writes_bytes;container_fs_reads_bytes;container_fs_usage_bytes;\n";
                my @containers_metrics = ( 'cpu', 'cpu_limit', 'cpu_request', 'memory', 'memory_limit', 'memory_request', 'container_fs_writes_total', 'container_fs_reads_total', 'container_fs_writes_bytes_total', 'container_fs_reads_bytes_total', 'container_fs_usage_bytes' );

                my %tmp_namespace;

                foreach my $pod ( keys %{ $perf->{pod} } ) {

                    # containers 
                    my %tmp_pod;
                    foreach my $container ( keys %{ $perf->{pod}{$pod}{container} } ) {
                        if (!defined $labelToId{"$pod--$container"}) {
                            next;
                        }
                        my $contKey = "$pod--$container";
                        foreach my $time ( keys %{ $perf->{pod}{$pod}{container}{$container} } ) {
                            $containers .= $labelToId{"$pod--$container"} . ";" . $time . ";";
                            for my $metric (@containers_metrics) {
                                if ( defined $perf->{pod}{$pod}{container}{$container}{$time}{$metric} ) {
                                    if ($metric =~ m/_total/) {
                                        # latency seconds to ms
                                        if ($metric eq "container_fs_write_seconds_total" || $metric eq "container_fs_read_seconds_total") {
                                            $perf->{pod}{$pod}{container}{$container}{$time}{$metric} = $perf->{pod}{$pod}{container}{$container}{$time}{$metric} * 1000;
                                        }

                                        if (defined $counters->{$contKey} && defined $counters->{$contKey}{$metric}) {

                                            # counter value check
                                            if ($perf->{pod}{$pod}{container}{$container}{$time}{$metric} > $counters->{$contKey}{$metric}{'value'}) {

                                                my $value = (($perf->{pod}{$pod}{container}{$container}{$time}{$metric} - $counters->{$contKey}{$metric}{'value'}) / ($time - $counters->{$contKey}{$metric}{'time'}));
                                                $containers .= $value;
                                                $containers .= ";";

                                                if (defined $tmp_pod{$metric}) {
                                                    $tmp_pod{$metric} += $value;
                                                } else {
                                                    $tmp_pod{$metric} = int($value);
                                                }

                                            } else {
                                                $containers .= ";";
                                            }
                                            
                                        } else {
                                            $containers .= ";";
                                        }
                                        my %counterObj = (
                                            "time" => $time,
                                            "value" => $perf->{pod}{$pod}{container}{$container}{$time}{$metric}
                                        );
                                        $counters->{$contKey}{$metric} = \%counterObj;
                                    } elsif($metric =~ m/memory/) {
                                        $containers .= $perf->{pod}{$pod}{container}{$container}{$time}{$metric} * 1024 * 1024;
                                        $containers .= ";";
                                    } else {
                                        $containers .= $perf->{pod}{$pod}{container}{$container}{$time}{$metric} . ";";
                                    }  
                                } else {
                                    $containers .= ";";
                                }
                            }
                            $containers .= "\n";
                        }
                    }

                    # pods 

                    if (!defined $labelToId{$pod}) {
                        next;
                    }
                    foreach my $time ( keys %{ $perf->{pod}{$pod}{data} } ) {
                        $pods .= $labelToId{$pod} . ";" . $time . ";";
                        for my $metric (@pods_metrics) {
                            if ( defined $perf->{pod}{$pod}{data}{$time}{$metric} ) {
                                if ($metric =~ m/_total/) {
                                    if (defined $counters->{$pod} && defined $counters->{$pod}{$metric}) {

                                        # counter value check
                                        if ($perf->{pod}{$pod}{data}{$time}{$metric} > $counters->{$pod}{$metric}{'value'}) {

                                            my $value = (($perf->{pod}{$pod}{data}{$time}{$metric}- $counters->{$pod}{$metric}{'value'}) / ($time - $counters->{$pod}{$metric}{'time'}));
                                            $pods .= $value;
                                            $pods .= ";";

                                            if (defined $tmp_namespace{$perf->{pod}{$pod}{namespace}} && defined $tmp_namespace{$perf->{pod}{$pod}{namespace}}{$metric}) {
                                                $tmp_namespace{$perf->{pod}{$pod}{namespace}}{$metric} += $value;
                                            } else {
                                                $tmp_namespace{$perf->{pod}{$pod}{namespace}}{$metric} = $value;
                                            }

                                        } else {
                                            $pods .= ";";
                                        }

                                    } else {
                                        $pods .= ";";
                                    }
                                    my %counterObj = (
                                        "time" => $time,
                                        "value" => $perf->{pod}{$pod}{data}{$time}{$metric}
                                    );
                                    $counters->{$pod}{$metric} = \%counterObj;
                                } elsif($metric =~ m/memory/) {
                                    $pods .= $perf->{pod}{$pod}{data}{$time}{$metric} * 1024 * 1024;
                                    $pods .= ";";
                                } else {
                                    $pods .= $perf->{pod}{$pod}{data}{$time}{$metric} . ";";
                                }
                            }
                            else {
                                $pods .= ";";
                            }
                        }

                        # metric from containers 
                        for my $metric (@tmp_pods_metrics) {
                            if (defined $tmp_pod{$metric}) {
                                $pods .= $tmp_pod{$metric};
                                $pods .= ";";

                                if (defined $tmp_namespace{$perf->{pod}{$pod}{namespace}} && defined $tmp_namespace{$perf->{pod}{$pod}{namespace}}{$metric}) {
                                    $tmp_namespace{$perf->{pod}{$pod}{namespace}}{$metric} += $tmp_pod{$metric};
                                } else {
                                    $tmp_namespace{$perf->{pod}{$pod}{namespace}}{$metric} = $tmp_pod{$metric};
                                }
                            } else {
                                $pods .= ";";
                            }
                        }

                        $pods .= "\n";
                    }
                }

                #print Dumper($pods);
                #print Dumper($containers);

                $xormon->saveCSV($pods, 'kubernetes', 'pod');

                $xormon->saveCSV($containers, 'kubernetes', 'container');

                # Namespaces
                my $namespaces         = "item_id;time;cpu;memory;container_fs_writes;container_fs_reads;container_fs_writes_bytes;container_fs_reads_bytes;container_network_receive_bytes;container_network_transmit_bytes\n";
                my @namespaces_metrics = ( 'cpu', 'memory' );
                my @tmp_namespaces_metrics = ( 'container_fs_writes_total', 'container_fs_reads_total', 'container_fs_writes_bytes_total', 'container_fs_reads_bytes_total', 'container_network_receive_bytes_total', 'container_network_transmit_bytes_total' );

                foreach my $namespace ( keys %{ $perf->{namespace} } ) {
                    foreach my $time ( keys %{ $perf->{namespace}{$namespace} } ) {
                        $namespaces .= $namespace . ";" . $time . ";";
                        for my $metric (@namespaces_metrics) {
                            if ( defined $perf->{namespace}{$namespace}{$time}{$metric} ) {
                                if($metric =~ m/memory/) {
                                    $namespaces .= $perf->{namespace}{$namespace}{$time}{$metric} * 1024 * 1024;
                                    $namespaces .= ";";
                                } else {
                                    $namespaces .= $perf->{namespace}{$namespace}{$time}{$metric} . ";";
                                } 
                            }
                            else {
                                $namespaces .= ";";
                            }
                        }

                        # metric from pods 
                        for my $metric (@tmp_namespaces_metrics) {
                            if (defined $tmp_namespace{$namespace} && $tmp_namespace{$namespace}{$metric}) {
                                $namespaces .= $tmp_namespace{$namespace}{$metric};
                                $namespaces .= ";";
                            } else {
                                $namespaces .= ";";
                            }
                        }

                        $namespaces .= "\n";
                    }
                }

                #print Dumper($namespaces);

                $xormon->saveCSV($namespaces, 'kubernetes', 'namespace');
            
                my %service_data_new = (
                    "counters" => $counters,
                    "pid" => $pid
                );
                $xormon->saveData(\%service_data_new);
            } else {
                 $xormon->hasData(0, "Unable to get perf data");
            }
        };
        warn $@ if $@;
        sleep($resolution);
    }
}

    

sub configuration {
    my $xormon = shift;
	my $data = shift;
    my @architecture;

    my $skipContainers = defined $data->{skipContainer} && $data->{skipContainer} eq "true" ? 1 : 0;
    my $kubernetes = Kubernetes->new($data->{hostalias}, $data->{host}, $data->{token}, $data->{protocol}, $data->{hostcfg_id}, $skipContainers, $data->{namespacesSelected});
    my $resolution = $kubernetes->metricResolution();
    $resolution = ($resolution->{node} gt $resolution->{pod}) ? $resolution->{node} : $resolution->{pod};

    my $conf = $kubernetes->getConfiguration();

    # cluster
    for my $cluster_id (keys %{$conf->{label}{cluster}}) {
        # architecture
        my %cluster_arch = (
            "item_id" => $cluster_id,
            "label" => $conf->{label}{cluster}{$cluster_id},
            "hw_type" => "kubernetes",
            "subsystem" => "cluster",
            "parents" => [],
            "hostcfg_id" => $data->{hostcfg_id}
        );
        push(@architecture, \%cluster_arch);

        # namespace under cluster
        for my $namespace_id (@{$conf->{architecture}{cluster_namespace}{$cluster_id}}) {
            my %namespace_arch = (
                "item_id" => $namespace_id,
                "label" => $conf->{label}{namespace}{$namespace_id},
                "hw_type" => "kubernetes",
                "subsystem" => "namespace",
                "parents" => [$cluster_id],
                "hostcfg_id" => $data->{hostcfg_id}
            );
            push(@architecture, \%namespace_arch);

            # pods under namespace
            if (defined $conf->{architecture}{namespace_pod}) {
                for my $pod_id (@{$conf->{architecture}{namespace_pod}{$namespace_id}}) {
                    my %pod_arch = (
                        "item_id" => $pod_id,
                        "label" => $conf->{label}{pod}{$pod_id},
                        "hw_type" => "kubernetes",
                        "subsystem" => "pod",
                        "parents" => [$cluster_id, $namespace_id],
                        "hostcfg_id" => $data->{hostcfg_id}
                    );
                    push(@architecture, \%pod_arch);

                    # containers under pod
                    if (defined $conf->{architecture}{pod_container} && defined $conf->{architecture}{pod_container}{$pod_id}) {
                        for my $container_id (@{$conf->{architecture}{pod_container}{$pod_id}}) {
                            my %container_arch = (
                                "item_id" => Xormon::create_hash_key_long($container_id),
                                "label" => $conf->{label}{container}{$container_id},
                                "hw_type" => "kubernetes",
                                "subsystem" => "container",
                                "parents" => [$cluster_id, $namespace_id, $pod_id],
                                "hostcfg_id" => $data->{hostcfg_id}
                            );
                            push(@architecture, \%container_arch);
                        }
                    }
                }
            }
        }
    }

    # nodes
    for my $node_id (keys %{$conf->{label}{node}}) {
        #architecture
        my %node_arch = (
            "item_id" => $node_id,
            "label" => $conf->{label}{node}{$node_id},
            "hw_type" => "kubernetes",
            "subsystem" => "node",
            "parents" => [$data->{hostcfg_id}],
            "hostcfg_id" => $data->{hostcfg_id}
        );
        push(@architecture, \%node_arch);
    }

    # save architecture
    my $rc = $xormon->saveArchitecture(\@architecture);
    if ($rc ne "201") {
        return $rc;
    }

    # configuration
    if (defined $conf->{specification}{node}) {
        $xormon->saveConfJSON($conf->{specification}{node});
    }
    if (defined $conf->{specification}{pod}) {
        $xormon->saveConfJSON($conf->{specification}{pod});
    }

    return $rc;
}

sub apitest {
    my $client = shift;
    my $data = shift;

    my @hostSplit = split(":", $data->{host});

    my $testModule = Test->new($client);
    my $connected = Test::connTest($hostSplit[0], $hostSplit[1]);
    if ($connected == 1) {
        $testModule->addSection("Connection test", 1);
    } else {
        $testModule->addSection("Connection test", 0, "TCP connection to " . $hostSplit[0] . ":" . $hostSplit[1] . " has failed! Open it on the firewall.");
        $testModule->stop();
        return;
    }

    my $kubernetes = Kubernetes->new($data->{hostalias}, $data->{host}, $data->{token}, $data->{protocol}, $data->{uuid});
    my $test = $kubernetes->apiTest();

    if (defined $test->{items}) {
        my $nodes = scalar @{$test->{items}};
        $testModule->addSection("Data test", 1, "Number of nodes: $nodes");
    } else {
        $testModule->addSection("Data test", 0, "Can not list nodes");
    }

    $testModule->stop();

}

sub namespaces {
    my $data = shift;
    my $kubernetes = Kubernetes->new($data->{hostalias}, $data->{host}, $data->{token}, $data->{protocol}, $data->{uuid});
    return encode_json($kubernetes->getNamespaces());
}

$server->close();