use FindBin;
use lib "$FindBin::Bin/"; # use lib in this folder
use lib "../../../perl-modules";
use LWP;
use JSON;
use strict;
use warnings;
use Data::Dumper;
use Xormon;
use LanOthers;
use Cwd qw(cwd);
use Test;

$SIG{CHLD} = 'IGNORE';

use IO::Socket qw(AF_INET AF_UNIX SOCK_STREAM SHUT_WR);

my $DEFAULT_PORT = 3403;
my $port = Xormon::checkPrefixPort($DEFAULT_PORT);

my $server = IO::Socket->new(
  Domain => AF_INET,
  Type => SOCK_STREAM,
  Proto => 'tcp',
  LocalHost => '0.0.0.0',
  LocalPort => $port,
  ReusePort => 1,
  Listen => 5,
) || die "Can't open socket: $@";
Xormon::log("Service listening on port $port");

my $client;

# waiting for a new client connection
while ($client = $server->accept()) {

  my $pid;
  unless ( defined( $pid = fork() ) ) {
    Xormon::error("Wait for fork");
    sleep 5;
  }

  if ($pid == 0) {

    # get information about a newly connected client
    my $client_address = $client->peerhost();
    my $client_port = $client->peerport();
    Xormon::log("Connection from $client_address:$client_port");

    # read up to 2048 characters from the connected client
    my $data = "";
    $client->recv($data, 2048);

    my $decoded = from_json($data);
    my $cfg     = $decoded->{data};

    Xormon::log("Received task: $decoded->{task} for Device: $cfg->{hostalias}");
    my $response = '';
    if ($decoded->{task} eq "fetch") {
      #
      # data collection
      #
      $client->send("Task $decoded->{task} successfuly started");
      $client->shutdown(SHUT_WR);

      my $xormon = Xormon->new("$client_address", $decoded->{jwt}, 'lanothers', $cfg->{hostcfg_id}, $cfg->{hostalias});

      # start task
      $xormon->startJob($decoded->{task}, $cfg->{hostalias}, $$);

      # target destinations check
      if (defined $decoded->{'destinations'}) {
        $xormon->setDestinations($decoded->{'destinations'});
      }

      #
      # collect and send to DB conf and perf data
      #
      my $wrkdir   = set_wrkdir($cfg);
      my $response = LanOthers::collect($xormon,$cfg,$wrkdir);

      # end task
      $xormon->endJob($decoded->{task}, $cfg->{hostalias}, $response, $$);
      Xormon::log("Task '$decoded->{task}' for ".$cfg->{hostalias}." is done. Status: $response");
    }
    elsif ( $decoded->{task} eq "test" ) {
      #
      # connection test
      #
      connection_test($cfg);
      $client->shutdown(SHUT_WR);
      Xormon::log("Task '$decoded->{task}' for ".$cfg->{hostalias}." is done.");
    }
    else {
      $response = "undefined task";

      $client->send($response);
      $client->shutdown(SHUT_WR);

    }

    exit();

  } else {
    $client->autoflush(1);
  }
}

$server->close();

exit();

sub connection_test {
  my $cfg  = shift;
  my $test = Test->new($client);

  #
  # test main variables
  #
  my @variables = ("snmpHost", "snmpVersion");
  foreach my $variable (@variables){
    unless ( exists $cfg->{$variable} ){
      $test->addSection("Device cfg", 0 , "Cannot continue, $variable is not configured!");
      return $test->stop();
    }
  }
  $test->addSection("Device cfg", 1 , "");

  #
  # test IP and port
  #
  my $port = ( exists $cfg->{'snmpPort'} && Xormon::isdigit($cfg->{'snmpPort'}) && $cfg->{'snmpPort'} < 65536 ) ? $cfg->{'snmpPort'} : 161; # default 161
  my ( $code, $text ) = Xormon::conntest_udp($cfg->{snmpHost},$port);
  $test->addSection("Network connection", $code, $text);
  if ( $code == 0 ) { return $test->stop(); }

  #
  # SNMP connection test
  #
  my ($status, $reason, $snmpwalk) = LanOthers::snmp_connection_test($cfg);

  my $response = "";
  my @rows     = ();
  if ( $status ) { # ok=1
    if ( defined $snmpwalk && defined $reason ) {
      push(@rows, "$snmpwalk\n");
      push(@rows, "$reason\n");
    }
    $test->addSection("SNMP connection", 1 , \@rows);
  }
  else { # nok=0
    if ( defined $reason ) {
      push(@rows, "$reason\n");
    }
    if ( defined $snmpwalk ) {
      push(@rows, "\n");
      push(@rows, "This snmpwalk command must work from the command line:\n");
      push(@rows, "$snmpwalk\n");
    }
    #push(@rows, "\n");
    #push(@rows, "Troubleshooting docu: https://xormon.com/SAN-troubleshooting.php\n");

    $test->addSection("SNMP connection", 0 , \@rows);
  }

  return $test->stop();
}

sub set_wrkdir {
  my $cfg = shift;

  unless ( exists $cfg->{'hostcfg_id'} ) { Xormon::error( "Cannot continue, 'hostcfg_id' doesn't exist! $!" . __FILE__ . ":" . __LINE__ ) && exit 0; }

  # microservice's directory
  my $act_dir = cwd;
  my $wrkdir  = "$act_dir/data/$cfg->{'hostcfg_id'}";

  unless ( -d "$act_dir/data" )   { mkdir( "$act_dir/data", 0755 )   || Xormon::error( "Cannot mkdir '$act_dir/data': $!" . __FILE__ . ":" . __LINE__ )   && exit 0; }
  unless ( -d "$wrkdir" )         { mkdir( "$wrkdir", 0755 )         || Xormon::error( "Cannot mkdir '$wrkdir': $!" . __FILE__ . ":" . __LINE__ )         && exit 0; }

  return $wrkdir;
}
