package Nutanix;

use strict;
use warnings;

use HTTP::Request::Common;
use LWP;
use Data::Dumper;
use JSON;
use POSIX qw(strftime ceil);
use Date::Parse;
use Time::Local;

sub new {
  my($self, $protocol, $ip, $port, $username, $password) = @_;
  
  my $o = {};
  $o->{protocol} = $protocol;
  $o->{ip} = $ip;
  $o->{port} = $port;
  $o->{username} = $username;
  $o->{password} = $password;
  bless $o;

  return $o;
}

sub getClusters {
  my ($self, $api) = @_;
  my ($data,$url);
  if ($api eq "v3") {
    $data = "{\"kind\":\"cluster\"}";
    $url = "clusters/list";
    return decode_json_eval($self->restCallv3($url, $data));
  } elsif ($api eq "v2") {
    return decode_json_eval($self->restCallv2("clusters/?projection=BASIC_INFO"," ","GET"));
  } elsif ($api eq "v1") {
    return decode_json_eval($self->restCallv1("clusters/?projection=BASIC_INFO"," ","GET"));
  }
}

sub getNodes {
  my ($self, $api) = @_;
  my ($data,$url);
  if ($api eq "v3") {
    $data = "{\"kind\":\"host\"}";
    $url = "hosts/list";
    return decode_json_eval($self->restCallv3($url, $data));
  } elsif ($api eq "v2") {
    return decode_json_eval($self->restCallv2("hosts/?projection=BASIC_INFO"," ","GET"));
  } elsif ($api eq "v1") {
    return decode_json_eval($self->restCallv1("hosts/?projection=BASIC_INFO"," ","GET"));
  }
}

sub getNodePerf {
  my ($self, $uuid, $metrics, $start, $end, $interval, $api) = @_;
  if ($api eq "v2" || $api eq "v3") {
    return decode_json_eval($self->restCallv2("hosts/$uuid/stats/?metrics=$metrics&start_time_in_usecs=".$start."000000&end_time_in_usecs=".$end."000000&interval_in_secs=$interval"," ","GET"));
  } elsif ($api eq "v1") {
    return decode_json_eval($self->restCallv1("hosts/$uuid/stats/?metrics=$metrics&startTimeInUsecs=".$start."000000&endTimeInUsecs=".$end."000000&intervalInSecs=$interval"," ","GET"));
  }
}

sub getVMs {
  my ( $self, $api, $page_max, $page ) = @_;
  my ( $data, $url );
  my $start = ($page - 1) * $page_max;

  #
  my $offset = 5;
  if  ($start > $offset) {
    $start = $start - 5;
    if ( $api ne "v1" ) {
      $page_max = $page_max + 5;
    }
  }

  if ( $api eq "v3" ) {
    $data = "{\"kind\":\"vm\", \"offset\": $start, \"length\": $page_max}";
    $url  = "vms/list";
    return decode_json_eval( $self->restCallv3( $url, $data ) );
  }
  elsif ( $api eq "v2" ) {
    return decode_json_eval( $self->restCallv2( "vms/?offset=" . $start . "&length=" . $page_max . "&projection=BASIC_INFO", " ", "GET" ) );
  }
  elsif ( $api eq "v1" ) {
    return decode_json_eval( $self->restCallv1( "vms/?count=" . $page_max . "&page=" . $page . "&projection=BASIC_INFO", " ", "GET" ) );
  }
}

sub getVMPerf {
  my ($self, $uuid, $metrics, $start, $end, $interval) = @_;
  return decode_json_eval($self->restCallv1("vms/$uuid/stats/?metrics=$metrics&startTimeInUsecs=".$start."000000&endTimeInUsecs=".$end."000000&intervalInSecs=$interval"," ","GET", 1));
}

sub getVMPerfRequest {
  my ($self, $uuid, $metrics, $start, $end, $interval) = @_;
  return $self->createPerfRequest("GET", $self->restUrlv1("vms/$uuid/stats/?metrics=$metrics&startTimeInUsecs=".$start."000000&endTimeInUsecs=".$end."000000&intervalInSecs=$interval"), $uuid);
}

sub getDisks {
  my ($self, $api, $page_max, $page) = @_;
  if ($api eq "v2" || $api eq "v3") {
    return decode_json_eval($self->restCallv2("disks/?count=".$page_max."&page=".$page."&?projection=BASIC_INFO"," ","GET"));
  } elsif ($api eq "v1") {
    return decode_json_eval($self->restCallv1("disks/?count=".$page_max."&page=".$page."&?projection=BASIC_INFO"," ","GET"));
  }
}

sub getDiskPerf {
  my ($self, $uuid, $metrics, $start, $end, $interval, $api) = @_;
  if ($api eq "v2" || $api eq "v3") {
    return decode_json_eval($self->restCallv2("disks/$uuid/stats/?metrics=$metrics&start_time_in_usecs=".$start."000000&end_time_in_usecs=".$end."000000&interval_in_secs=$interval"," ","GET"));
  } elsif ($api eq "v1") {
    return decode_json_eval($self->restCallv1("disks/$uuid/stats/?metrics=$metrics&startTimeInUsecs=".$start."000000&endTimeInUsecs=".$end."000000&intervalInSecs=$interval"," ","GET"));
  }
}

sub getStorageContainers {
  my ($self, $api) = @_;
  my ($data,$url);
  if ($api eq "v2" || $api eq "v3") {
    return decode_json_eval($self->restCallv2("storage_containers/?projection=BASIC_INFO"," ","GET"));
  } elsif ($api eq "v1") {
    return decode_json_eval($self->restCallv1("containers/?projection=BASIC_INFO"," ","GET"));
  }
}

sub getStorageContainerPerf {
  my ($self, $uuid, $metrics, $start, $end, $interval, $api) = @_;
  if ($api eq "v2" || $api eq "v3") {
    return decode_json_eval($self->restCallv2("storage_containers/$uuid/stats/?metrics=$metrics&start_time_in_usecs=".$start."000000&end_time_in_usecs=".$end."000000&interval_in_secs=$interval"," ","GET"));
  } elsif ($api eq "v1") {
    return decode_json_eval($self->restCallv1("containers/$uuid/stats/?metrics=$metrics&startTimeInUsecs=".$start."000000&endTimeInUsecs=".$end."000000&intervalInSecs=$interval"," ","GET"));
  }
}

sub getStoragePools {
  my ($self, $api) = @_;
  my ($data,$url);
  if ($api eq "v2" || $api eq "v3" || $api eq "v1") {
    return decode_json_eval($self->restCallv1("storage_pools/"," ","GET"));
  }
}

sub getStoragePoolPerf {
  my ($self, $uuid, $metrics, $start, $end, $interval, $api) = @_;
  if ($api eq "v2" || $api eq "v3" || $api eq "v1") {
    return decode_json_eval($self->restCallv1("storage_pools/$uuid/stats/?metrics=$metrics&startTimeInUsecs=".$start."000000&endTimeInUsecs=".$end."000000&intervalInSecs=$interval"," ","GET"));
  }
}

sub getVdisks {
  my ($self, $api, $page_max, $page) = @_;
  if ($api eq "v2" || $api eq "v3") {
    return decode_json_eval($self->restCallv2("virtual_disks/?count=".$page_max."&page=".$page."&?projection=BASIC_INFO"," ","GET"));
  } elsif ($api eq "v1") {
    return decode_json_eval($self->restCallv1("virtual_disks/?count=".$page_max."&page=".$page."&?projection=BASIC_INFO"," ","GET"));
  }
}

sub getVdiskPerf {
  my ($self, $uuid, $metrics, $start, $end, $interval, $api) = @_;
  if ($api eq "v2" || $api eq "v3") {
    return decode_json_eval($self->restCallv2("virtual_disks/$uuid/stats/?metrics=$metrics&start_time_in_usecs=".$start."000000&end_time_in_usecs=".$end."000000&interval_in_secs=$interval"," ","GET"));
  } elsif ($api eq "v1") {
    return decode_json_eval($self->restCallv1("virtual_disks/$uuid/stats/?metrics=$metrics&startTimeInUsecs=".$start."000000&endTimeInUsecs=".$end."000000&intervalInSecs=$interval"," ","GET"));
  }
}

sub getVdiskPerfRequest {
  my ($self, $uuid, $metrics, $start, $end, $interval, $api) = @_;
  if ($api eq "v2" || $api eq "v3") {
    return $self->createPerfRequest("GET", $self->restUrlv2("virtual_disks/$uuid/stats/?metrics=$metrics&start_time_in_usecs=".$start."000000&end_time_in_usecs=".$end."000000&interval_in_secs=$interval"), $uuid);
  } elsif ($api eq "v1") {
    return $self->createPerfRequest("GET", $self->restUrlv1("virtual_disks/$uuid/stats/?metrics=$metrics&startTimeInUsecs=".$start."000000&endTimeInUsecs=".$end."000000&intervalInSecs=$interval"), $uuid);
  }
}

sub getEvents {
  my ($self, $start) = @_;
  return decode_json_eval($self->restCallv2("events/?start_time_in_usecs=".$start."000000"," ","GET"));
}

sub getAlerts {
  my ($self, $start) = @_;
  return decode_json_eval($self->restCallv2("alerts/?start_time_in_usecs=".$start."000000"," ","GET"));
}

sub getItemHealth {
  my ($self, $item) = @_;
  return decode_json_eval($self->restCallv1("$item?projection=health"," ","GET"));
}

sub getHealthCheck {
  my ($self, $item) = @_;
  return decode_json_eval($self->restCallv2("health_checks/$item/"," ","GET"));
}

sub getHealthSummary {
  my ($self) = @_;
  return decode_json_eval($self->restCallv1("ncc/run_summary/"," ","GET"));
}

#v3 for Prism Central
sub restCallv3() {
  my ($self, $url, $data) = @_;

  my $ua = LWP::UserAgent->new(
    timeout => 60,
    ssl_opts => { SSL_cipher_list => 'DEFAULT:!DH', verify_hostname => 0, SSL_verify_mode => 0 },
  );

  my $header = ['Content-Type' => 'application/json; charset=UTF-8'];
  my $request = HTTP::Request->new('POST', $self->{protocol}."://".$self->{ip}.":".$self->{port}."/api/nutanix/v3/$url", $header, $data);

  $request->authorization_basic($self->{username}, $self->{password});
  my $response = $ua->request($request);

  if ($response->is_success) {
    return $response->content;
  }
  else {
    error("ERROR: [/api/nutanix/v3/$url] Can't handle request : ".Dumper($response->content));
    return ();
  }
}

sub restUrlv3() {
  my ($self, $url) = @_;

  return $self->{protocol}."://".$self->{ip}.":".$self->{port}."/api/nutanix/v3/$url";
}

#v2 for Prism Element
sub restCallv2 {
  my ($self, $url, $data, $method) = @_;

  my $ua = LWP::UserAgent->new(
    timeout => 60,
    ssl_opts => { SSL_cipher_list => 'DEFAULT:!DH', verify_hostname => 0, SSL_verify_mode => 0 },
  );

  my $header = ['Content-Type' => 'application/json; charset=UTF-8'];
  my $request = HTTP::Request->new($method, $self->{protocol}."://".$self->{ip}.":".$self->{port}."/PrismGateway/services/rest/v2.0/$url", $header, $data);

  $request->authorization_basic($self->{username}, $self->{password});
  my $response = $ua->request($request);

  if ($response->is_success) {
    return $response->content;
  }
  else {
    error("ERROR: [/PrismGateway/services/rest/v2.0/$url] Can't handle request : ".Dumper($response->content));
    return ();
  }
}

sub restUrlv2() {
  my ($self, $url) = @_;

  return $self->{protocol}."://".$self->{ip}.":".$self->{port}."/PrismGateway/services/rest/v2.0/$url";
}

#v1 for Prism Element
sub restCallv1 {
  my ($self, $url, $data, $method, $test) = @_;

  my $ua = LWP::UserAgent->new(
    timeout => 60,
    ssl_opts => { SSL_cipher_list => 'DEFAULT:!DH', verify_hostname => 0, SSL_verify_mode => 0 },
  );

  my $header = ['Content-Type' => 'application/json; charset=UTF-8'];
  my $request = HTTP::Request->new($method, $self->{protocol}."://".$self->{ip}.":".$self->{port}."/PrismGateway/services/rest/v1/$url", $header, $data);

  $request->authorization_basic($self->{username}, $self->{password});

  if (defined $test) {
    #print Dumper($request);
  }

  my $response = $ua->request($request);

  if (defined $test) {
    #print Dumper($response);
  }

  if ($response->is_success) {
    return $response->content;
  }
  else {
    error("ERROR: [/PrismGateway/services/rest/v1/$url] Can't handle request : ".Dumper($response->content));
    return ();
  }
}

sub restUrlv1() {
  my ($self, $url) = @_;

  return $self->{protocol}."://".$self->{ip}.":".$self->{port}."/PrismGateway/services/rest/v1/$url";
}

sub authTest {
  my ($self) = @_;

  my $ua = LWP::UserAgent->new(
    timeout => 60,
    ssl_opts => { SSL_cipher_list => 'DEFAULT:!DH', verify_hostname => 0, SSL_verify_mode => 0 },
  );

  my $header = ['Content-Type' => 'application/json; charset=UTF-8'];
  my $request = HTTP::Request->new("GET", $self->{protocol}."://".$self->{ip}.":".$self->{port}."/PrismGateway/services/rest/v2.0/clusters/?projection=BASIC_INFO", $header);

  $request->authorization_basic($self->{username}, $self->{password});
  my $response = $ua->request($request);

  if ( $response->code == 401 ) {
    return 0;
  } else {
    return 1;
  }
}

sub createPerfRequest {
  my ($self, $method, $url, $id) = @_;

  my $header = ['Content-Type' => 'application/json; charset=UTF-8'];
  my $request = HTTP::Request->new($method, $url, $header);
  if (defined $id) {
    $request->header( 'uuid' => $id );
  }
  $request->authorization_basic($self->{username}, $self->{password});
  return $request;
}

sub decode_json_eval {
  my ($json, $title) = @_;
  my $hash = ();
  eval {
    $hash = decode_json($json);
  };
  if ($@) {
    my $error = $@;
    my $errTitle = defined $title ? "$title [Error]" : "[Error]";
    #error($error);
    if (length($json) == 0) {
      error("$errTitle [Decode JSON] Empty body");
    } else {
      error("$errTitle [Decode JSON] Raw body: $json");
    }
    return ();
  } else {
    return $hash;
  }
}

sub log {
  my $text     = shift;
  my $act_time = localtime();
  chomp($text);

  print "[$act_time]: $text \n";
  return 1;
} ## end sub error

sub error {
  my $text     = shift;
  my $act_time = localtime();
  chomp($text);

  print STDERR "$act_time: $text : $!\n";
  return 1;
} ## end sub error

sub error_die {
  my $message  = shift;
  my $act_time = localtime();
  print STDERR "$act_time: $message : $!\n";
  exit(1);
} ## end sub error_die

1;
