##!/usr/bin/perl
#
# svcperf.pl
#
# $Revision: 2.3.5 $  2023-06-01
my ($revision) = ( '$Revision: 2.3.5 $' =~ /([\d\.-]+)/ );

# Changes:
# 1.1.1 Add FlashSystem 840/900 support
# 1.1.7 Add FlashSystem 9100 support
# 1.2.x Reduce ssh/scp login sessions
# 2.1.x Implement Net::OpenSSH
# 2.2.x Implement tier_scm
# 2.3.x Does not start copying iostats files from non-config nodes - cpdumps (fw 8.3.1.2 and later)

# XormonNG uses different libs
my $xormonNG = ( $ENV{XormonNG} ) ? $ENV{XormonNG} : 0;
if ($xormonNG) {

  #require FindBin;
  use FindBin;
  use lib "$FindBin::Bin/";    # use lib in this folder
  use lib "../../../perl-modules";
  require Xormon;
  require XML::Simple;
}
else {
  #use Xorux_lib;
  unless ( exists $ENV{BINDIR} ) { error( "BINDIR is not defined! $!" . __FILE__ . ":" . __LINE__ ) && exit 1; }
  require Xorux_lib;
  require "$ENV{BINDIR}/xml.pl";    # it replaces above line and fixes an issue on Perl 5.10.1 on AIX 7100-00-06-1216
}

# Modules
use strict;

use Storable qw(retrieve nstore);
# Bug: Max. recursion depth with nested structures exceeded
$Storable::recursion_limit=-1;
$Storable::recursion_limit_hash=-1;

use Data::Dumper;
use Date::Parse;
use SVC 2.3;
use FlashSystem 2.2.1;

# Constant
use constant SVCPRF101W => 101;
use constant SVCPRF102W => 102;
use constant SVCPRF103W => 103;
use constant SVCPRF104W => 104;
use constant SVCPRF105W => 105;
use constant SVCPRF106W => 106;
use constant SVCPRF107W => 107;
use constant SVCPRF108W => 108;
use constant SVCPRF109W => 109;
use constant SVCPRF110E => 110;
use constant SVCPRF111E => 111;
use constant SVCPRF112E => 112;
use constant SVCPRF113E => 113;
use constant SVCPRF114E => 114;
use constant SVCPRF115E => 115;
use constant SVCPRF116E => 116;
use constant SVCPRF117E => 117;
use constant SVCPRF118E => 118;
use constant SVCPRF119E => 119;
use constant SVCPRF120E => 120;
use constant SVCPRF121E => 121;

# Options and their default value
my $storage;    # SVC/Storwize V7000 alias
my $svc;        # SVC/Storwize V7000 cluster host name (IP address)
my $user            = "admin";    # SVC/Storwize V7000 user name
my $key             = "";         # SSH key filename
my $interval        = 1;          # interval time for data collection
my $dir             = "..";       # Data directory
my $ssh_timeout     = 900;        # SSH Command timeout - 15 min
my $net_openssh     = 1;          # Use Net::OpenSSH Perl Modul
my $password        = "";
my $password_enable = "";

my $debug      = 0;               # Debug mode 0:off, 1:on
my $debug_full = 0;               # Debug mode 0:off, 1:on
my $codevrmn = 0;

if ( defined $ENV{STORAGE_NAME} ) {
  $storage = $ENV{STORAGE_NAME};
}
else {
  message("SVC/Storwize V7000 storage name alias is required.");
  exit(1);
}
if ( defined $ENV{SVC_IP} ) {
  $svc = $ENV{SVC_IP};
}
else {
  message("SVC/Storwize V7000 host name or IP address is required.");
  exit(1);
}
if ( defined $ENV{SVC_USER} )            { $user            = $ENV{SVC_USER} }
if ( defined $ENV{SVC_KEY} )             { $key             = $ENV{SVC_KEY} }
if ( defined $ENV{SVC_INTERVAL} )        { $interval        = $ENV{SVC_INTERVAL} }
if ( defined $ENV{SVC_DIR} )             { $dir             = $ENV{SVC_DIR} }
if ( defined $ENV{DEBUG} )               { $debug           = $ENV{DEBUG} }
if ( defined $ENV{SVC_PASSWORD_ENABLE} ) { $password_enable = $ENV{SVC_PASSWORD_ENABLE} }
if ( defined $ENV{SVC_PASSWORD} )        { $password        = $ENV{SVC_PASSWORD} }

if ( defined $ENV{SVC_DEBUG_FULL} ) {
  if    ( $ENV{SVC_DEBUG_FULL} =~ /^\s*(\d{1})\s*$/ ) { $debug      = $1; $debug_full = $1; }
  elsif ( $ENV{SVC_DEBUG_FULL} eq $storage )          { $debug      = 1; $debug_full = 1; }
  else                                                { $debug_full = 0; }
}
if ( defined $ENV{NET_OPENSSH} ) { $net_openssh = $ENV{NET_OPENSSH} }

#
if ($net_openssh) {

  # use require here to do not affect all storage whe it does not exist and "use" is on the top
  require Xorux::Net::OpenSSH;
  if ( $Xorux::Net::OpenSSH::VERSION < 0.78 ) {
    &error( "Current version of Xorux::Net::OpenSSH is " . $Xorux::Net::OpenSSH::VERSION . ". Use version 0.78 and higher." );
  }
  Xorux::Net::OpenSSH->import(qw(new disconnect capture scp_get pipe_out));
}

# Global variables with their value
my $net_ssh;
my $cmd_ssh;
my $cmd_scp;
my $cmd_ls   = "/bin/ls";
my $cmd_mv   = "/bin/mv";
my $cmd_rm   = "/bin/rm";
my $cmd_grep = "/bin/grep";

$cmd_ssh = $ENV{SSH};

my $iosdir = "$dir/iostats/";    # Directory for iostats files
my $tmpdir = "$dir/";            # Directory for temporary files

# Global variables
my $terminate_pgm;
my $iostatsdumps_name;           # SVC, Storwize V7000 dependent
my ( $node,    $front_panel_id, @stats_files, $stats_file );    # Current scan data
my ( @Nn_last, @Nm_last,        @Nv_last,     @Nd_last );       # Last scan data
my $time_last;
my $iostat_file_missing_max = 3;

my ( $command, $ret );
my $cfg_ref;
my %last_data;
my $last_data_ref = \%last_data;
my %prev_data;
my $prev_data_ref = \%prev_data;
my %mdiskgrp_cap;
my $mdiskgrp_cap = \%mdiskgrp_cap;
my $iostat_file_missing_count;
my %node_front_panel;

# Storage Hardware Types
my %storage_info = ( "is_svc" => 0, "is_storwize" => 0, "is_fs840" => 0, "is_fs9100" => 0, "is_v9000" => 0 );

# SVC Hardware Types
my %svc_g1 = ( "CG8" => 1, "CF8" => 1, "8A4" => 1, "8G4" => 1, "8F4" => 1, "8F2" => 1, "4F2" => 1 );
my %svc_g2 = ( "DH8" => 1, "SV1" => 1 );

if ( defined $password && $password ne "" ) {
  unless ($xormonNG) {
    $password = Xorux_lib::unobscure_password($password);
  }
}

# Main process

# Display options
print( "
svcperf.pl: Starts with the following options ...
\tRevision:                                $revision
\tSVC/Storwize V7000 cluster name alias:   $storage
\tSVC/Storwize V7000 cluster host name:    $svc
\tInterval:                                $interval
\tDirectory for iostats files:             $iosdir
\tDirectory for temporary files:           $tmpdir
\tDebug:                                   $debug
" );

# Open remote session to the storage (Net::OpenSSH) or setup SSH/SCP commands
if ($net_openssh) {
  if ($debug_full) { &message( "Open SSH session - host: " . $svc . ", user: " . $user . " key: " . $key ); }
  $net_ssh = SVC::openssh_new( $svc, "22", $user, $key, $ssh_timeout, $password, $password_enable );
  if ( not defined $net_ssh ) {
    &error( "Unable to connect to remote machine " . $svc, 2 );
    exit 2;
  }
}
else {
  if ($debug_full) { &message( "Setup SSH command - host: " . $svc . ", user: " . $user . " key: " . $key ); }
  $cmd_ssh = SVC::cmdssh_setup( $svc, "22", $user, $key, $cmd_ssh );
  $cmd_scp = SVC::cmdscp_setup( $svc, "22", $user, $key, $cmd_scp );
}

# Examine system name
my $svc_name;
my $lssystem = &LsSystem();

if ( not defined $lssystem->{'name'} ) {
  message("Cannot connect storage.");
  exit 1;
}
$svc_name = $lssystem->{'name'};
if ( "x$svc_name" eq "x" ) {
  message("Cannot get system name.");
  exit 1;
}
if ($debug) { message("System name    : $svc_name"); }

&SetStorageInfo($lssystem);

# Set filenames
my $logfile       = $tmpdir . "../../logs/" . $storage . ".svcperf.log";    # log file
my $tmp_perf_file = $tmpdir . $storage . "_svcperf.data";                   #
my $out_perf_file;
my $tmp_conf_file;
my $tmp_conf_full_file = ${storage} . "_svcconf.datafull";

# Create log file
my $time = localtime();
$time = localtime();
message( "Started at " . $time );

# Read etc/.magic file
unless ($xormonNG) { &Read_magic(); }

# Handling signals
$SIG{'INT'}  = \&Sig_Handler;
$SIG{'TERM'} = \&Sig_Handler;

# Change STOUT buffer attribute
$| = 1;    # disable STDOUT buffer

if ( not $storage_info{'is_fs840'} ) {

  # Examine H/W type and Set iostats dumps file name (NOTE: change to script - same command to clear)
  my $cmd = "svcinfo lsnode -nohdr -delim :";
  if ($debug_full) { message( "Remote Command: " . $cmd ); }
  my $line_LSNODE;
  if ($net_openssh) {
    $line_LSNODE = SVC::openssh_runcmd( $net_ssh, $cmd, $ssh_timeout );
  }
  else {
    $line_LSNODE = SVC::cmdssh_runcmd( $cmd_ssh, $cmd, $ssh_timeout );
  }
  if ( not defined $line_LSNODE ) {
    &error( "Remote Command: " . $cmd . " failed", 1 );
    exit(1);
  }

  # if($debug_full) {print("The svcinfo lsnode command output: $line_LSNODE");}
  my @field_LSNODE = split( ':', $line_LSNODE );
  if ( length( $field_LSNODE[2] ) == 0 && length( $field_LSNODE[15] ) != 0 ) {

    # Storwize V7000 = UPS_serial_number is blank AND enclosure_serial_number is not blank
    $iostatsdumps_name = "dump_name";    # Storwize V7000 iostats dumps file name
    if ($debug_full) { message("H/W type, iostats dumps file name (Storwize V7000): $iostatsdumps_name."); }
  }
  elsif ( length( $field_LSNODE[8] ) == 0 && length( $field_LSNODE[15] ) != 0 ) {

    # Storwize V7000 = UPS_unique_id is blank AND enclosure_serial_number is not blank
    $iostatsdumps_name = "dump_name";    # Storwize V7000 iostats dumps file name
    if ($debug_full) { message("H/W type, iostats dumps file name (Storwize V7000): $iostatsdumps_name."); }
  }
  elsif ( $svc_g1{ $field_LSNODE[9] } ) {
    $iostatsdumps_name = "front_panel_id";    # SVC - Gen1 - with UPS - iostats dumps file name
    if ($debug_full) { message("H/W type, iostats dumps file name (SVC): $iostatsdumps_name."); }
  }
  elsif ( $svc_g2{ $field_LSNODE[9] } ) {
    $iostatsdumps_name = "front_panel_id";    # SVC - Gen2 - with battery - iostats dumps file name
    if ($debug_full) { message("H/W type, iostats dumps file name (SVC): $iostatsdumps_name."); }
  }

  # Start SVC/Storwize V7000 iostats collection
  # IBM FlashSystem 840 not supported command startstats - default 15m
  if ($debug) { message( "Statistics collection: " . $lssystem->{'statistics_status'} . " - Interval: " . $lssystem->{'statistics_frequency'} ); }
  if ( $lssystem->{'statistics_frequency'} != $interval ) {
    my $cmd = "svctask startstats -interval " . $interval;
    if ($debug_full) { message( "Remote Command: " . $cmd ); }
    my $out;
    if ($net_openssh) {
      $out = SVC::openssh_runcmd( $net_ssh, $cmd, $ssh_timeout );
    }
    else {
      $out = SVC::cmdssh_runcmd( $cmd_ssh, $cmd, $ssh_timeout );
    }
    if ( not defined $out ) {
      &error( "Remote Command: " . $cmd . " failed" );
    }
    if ($debug) { message("Start statistics collection: interval $interval"); }
  }

  # Initialize Last uploaded data
  $time_last = "190101_000000";    # (\d{6}_\d{6}) # SVC svqueryclock
  if ($debug_full) { message("IOSTATS dir list:"); }
  my @stats_list = (`$cmd_ls $iosdir | $cmd_grep _stats_`);
  while (1) {
    if ($debug_full) { message( "Find loop - IOSTAT files - Last time: " . $time_last ); }
    my $time_last_save = $time_last;
    if ( scalar(@stats_list) == 0 ) { last; }
    my $stats_onetime = &OneTimeStatsList( \@stats_list );
    if ( scalar( @{$stats_onetime} ) == 0 ) {
      last;
    }
  }

  $cmd = "svcinfo lsnode -nohdr -delim :";
  if ($debug_full) { &message( "Remote Command: " . $cmd ); }
  my $fh;
  if ($net_openssh) {
    $fh = SVC::openssh_pipeout( $net_ssh, [$cmd], $ssh_timeout );
  }
  else {
    $fh = SVC::cmdssh_pipeout( $cmd_ssh, [$cmd], $ssh_timeout );
  }
  if ( not defined $fh ) {
    &error( "Remote Command: " . $cmd . " failed.", 1 );
    return (1);
  }
  while ( my $line_LSNODE = <$fh> ) {

    #if($debug_full) {message("The svcinfo lsnode command output: $line_LSNODE");}
    my @field_LSNODE = split( ':', $line_LSNODE );

    # Check none configuration nodes
    if ( $field_LSNODE[4] eq "online" or $field_LSNODE[4] eq "online_spare" ) {
      $node = $field_LSNODE[0];
      my $front_panel_id = &ID_front_panel($node);
      if ($debug_full) { message( "node: " . $node . " ID: " . $front_panel_id ); }
      if ( not defined $front_panel_id ) { next; }
      foreach my $stats_filename (@stats_list) {
        chomp($stats_filename);
        if ( $stats_filename =~ /^Nn_stats_([\w]+[-\d]*)_(\d+_\d+)$/ ) {
          if ( $1 eq $front_panel_id and $2 eq $time_last ) {
            $Nn_last[$node] = $stats_filename;
            if ($debug_full) { message( "node: " . $node . " ID: " . $front_panel_id . " time: " . $time_last . " iostat name: " . $stats_filename ); }
          }
        }
        elsif ( $stats_filename =~ /^Nm_stats_([\w]+[-\d]*)_(\d+_\d+)$/ ) {
          if ( $1 eq $front_panel_id and $2 eq $time_last ) {
            $Nm_last[$node] = $stats_filename;
            if ($debug_full) { message( "node: " . $node . " ID: " . $front_panel_id . " time: " . $time_last . " iostat name: " . $stats_filename ); }
          }
        }
        elsif ( $stats_filename =~ /^Nv_stats_([\w]+[-\d]*)_(\d+_\d+)$/ ) {
          if ( $1 eq $front_panel_id and $2 eq $time_last ) {
            $Nv_last[$node] = $stats_filename;
            if ($debug_full) { message( "node: " . $node . " ID: " . $front_panel_id . " time: " . $time_last . " iostat name: " . $stats_filename ); }
          }
        }
        elsif ( $stats_filename =~ /^Nd_stats_([\w]+[-\d]*)_(\d+_\d+)$/ ) {
          if ( $1 eq $front_panel_id and $2 eq $time_last ) {
            $Nd_last[$node] = $stats_filename;
            if ($debug_full) { message( "node: " . $node . " ID: " . $front_panel_id . " time: " . $time_last . " iostat name: " . $stats_filename ); }
          }
        }
      }
    }
  }
  close($fh);
  if ($debug_full) { print( "NnSTAT: ", Dumper( \@Nn_last ), "\n" ); }
}    ## if ( $storage_info{'is_fs840'} )

# main loop
$terminate_pgm             = 0;
$iostat_file_missing_count = 0;
my $main_loop_start_time;
my $main_loop_interval_time = 0;

while (1) {
  if ($debug_full) { $time = localtime(); message( "MAIN LOOP START. --- " . $time ); }
  $main_loop_start_time = time();
  &Retrieve_last_conf_data();
  # Get Firmware Code Level  
  $codevrmn = &GetCodeLevelNum;

  if ( $storage_info{'is_fs840'} ) {

    # Flash System 840/900
    if ($debug_full) { 
      message("Start process FlashSystem 840/900 statistic"); 
      message("Code Level: [".$codevrmn."] - ".$cfg_ref->{'svc_system'}->{'code_level'} ); 
    }
    &Retrieve_prev_perf_data();
    my $timestamp = &FlashSystem::Last_timestamps( $prev_data_ref, $iosdir . "current_short/timestamp", $cfg_ref, $debug_full );
    message( "Last Timestamp: " . &FlashSystem::Print_timestamps($timestamp) );
    if ($net_openssh) {
      $out_perf_file = $tmpdir . $storage . "_svcflashperf_";
      if ($debug_full) { &message( "Performance Output Filename prefix: " . $out_perf_file . "." ) }
      &FlashSystem::Upload_iostats_files_openssh( $debug_full, $net_ssh, $ssh_timeout, $iosdir, $cfg_ref );
    }
    else {
      $out_perf_file = $tmpdir . $storage . "_svcflashperf_";
      if ($debug_full) { &message( "Performance Output Filename prefix: " . $out_perf_file . "." ) }
      &FlashSystem::Upload_iostats_files_cmdssh( $debug_full, $cmd_scp, $ssh_timeout, $iosdir, $cfg_ref );
    }
    &FlashSystem::Process_iostats_files( $debug_full, $timestamp, $iosdir, $out_perf_file, $cfg_ref );
    &Store_prev_perf_data();
    if ($debug_full) { message("End process FlashSystem 840/900 statistic"); }
  }
  else {
    # SVC/Storwize/V9000/FS9100
    my @active_nodes;
    my $config_node;
    my %cmd_lsdump;

    if ($debug_full) { 
      message("Start process SVC/Storwize/V9000/FS9100 statistic"); 
      message("Code Level: [".$codevrmn."] - ".$cfg_ref->{'svc_system'}->{'code_level'} ); 
    }

    #my %node_front_panel;

    #if ($debug_full) { message("Command: $ssh svcinfo lsnode -nohdr -delim :"); }
    # $cfg_ref->{svc_node}->{$rec}->{$key}
    my @LSNODE = keys( %{ $cfg_ref->{svc_node} } );

    # Scan all nodes and prepare storage commands                              ...get performance data
    foreach my $rec (@LSNODE) {

      #if($debug_full) {message("The svcinfo lsnode command output: $line_LSNODE");}
      #my @field_LSNODE = split( ':', $line_LSNODE );
      # id,name,UPS_serial_number,WWNN,status,IO_group_id,IO_group_name,config_node,UPS_unique_id,hardware,iscsi_name,iscsi_alias,panel_name,enclosure_id,canister_id,enclosure_serial_number
      if ( $cfg_ref->{svc_node}->{$rec}->{status} eq "online" or $cfg_ref->{svc_node}->{$rec}->{status} eq "online_spare" ) {    # Check none configuration nodes
        if ( $cfg_ref->{svc_node}->{$rec}->{config_node} =~ "no" ) {
          $node = $cfg_ref->{svc_node}->{$rec}->{id};
          my $front_panel_id = &ID_front_panel($node);
          if ($debug_full) { message( "Non-configuration node: " . $node . " ID: " . $front_panel_id ); }
          if ( not defined $front_panel_id ) { next; }
          $cmd_lsdump{$node} .= "svcinfo lsiostatsdumps -nohdr -delim : " . $node . ";";
          push( @active_nodes, $front_panel_id );

          #&List_iostats_create();
          #&List_iostats_copy();
        }
        else {    # Check configuration node
          $node = $cfg_ref->{svc_node}->{$rec}->{id};
          my $front_panel_id = &ID_front_panel($node);
          if ($debug_full) { message( "Configuration node: " . $node . " ID: " . $front_panel_id ); }
          if ( not defined $front_panel_id ) { next; }
          $config_node = $node;
          $cmd_lsdump{$node} .= "svcinfo lsiostatsdumps -nohdr -delim : " . $node . ";";
          push( @active_nodes, $front_panel_id );

          #&List_iostats_create();
          #&List_iostats_get();
        }
      }
    }

    my $stats_files = &List_iostats_create( \%cmd_lsdump, \%node_front_panel );
    List_iostats_upload( $config_node, \%node_front_panel, $stats_files );

    # Check uploaded files
    if ($debug_full) { message("Check uploaded iostat files."); }
    my @stats_list = (`$cmd_ls $iosdir | $cmd_grep _stats_`);
    foreach my $rec (@LSNODE) {

      # Check none configuration nodes
      if ( $cfg_ref->{svc_node}->{$rec}->{status} eq "online" or $cfg_ref->{svc_node}->{$rec}->{status} eq "online_spare" ) {
        $node = $cfg_ref->{svc_node}->{$rec}->{id};
        my $front_panel_id = &ID_front_panel($node);
        if ( not defined $front_panel_id ) { next; }
        if ($debug_full) { message( "node: " . $node . " ID: " . $front_panel_id ); }
        foreach my $stats_filename (@stats_list) {
          chomp($stats_filename);
          if ( $stats_filename =~ /^Nn_stats_([\w]+[-\d]*)_(\d+_\d+)$/ ) {
            if ( $1 eq $front_panel_id and $stats_filename gt $Nn_last[$node] ) {
              $Nn_last[$node] = $stats_filename;
              if ($debug_full) { message( "node: " . $node . " ID: " . $front_panel_id . " time: " . $time_last . " iostat name: " . $stats_filename ); }
            }
          }
          elsif ( $stats_filename =~ /^Nm_stats_([\w]+[-\d]*)_(\d+_\d+)$/ ) {
            if ( $1 eq $front_panel_id and $stats_filename gt $Nm_last[$node] ) {
              $Nm_last[$node] = $stats_filename;
              if ($debug_full) { message( "node: " . $node . " ID: " . $front_panel_id . " time: " . $time_last . " iostat name: " . $stats_filename ); }
            }
          }
          elsif ( $stats_filename =~ /^Nv_stats_([\w]+[-\d]*)_(\d+_\d+)$/ ) {
            if ( $1 eq $front_panel_id and $stats_filename gt $Nv_last[$node] ) {
              $Nv_last[$node] = $stats_filename;
              if ($debug_full) { message( "node: " . $node . " ID: " . $front_panel_id . " time: " . $time_last . " iostat name: " . $stats_filename ); }
            }
          }
          elsif ( $stats_filename =~ /^Nd_stats_([\w]+[-\d]*)_(\d+_\d+)$/ ) {
            if ( $1 eq $front_panel_id and $stats_filename gt $Nd_last[$node] ) {
              $Nd_last[$node] = $stats_filename;
              if ($debug_full) { message( "node: " . $node . " ID: " . $front_panel_id . " time: " . $time_last . " iostat name: " . $stats_filename ); }
            }
          }
        }
      }
    }
    if ($debug_full) { print( "NnSTAT: ", Dumper( \@Nn_last ), "\n" ); }

    # Scan uploaded data and compute
    while (1) {
      if ($debug_full) { message( "IOSTAT files loop start - Last time: " . $time_last . "  Nodes: " . scalar(@active_nodes) ); }
      my $time_last_save = $time_last;
      my @stats_list     = (`$cmd_ls $iosdir | $cmd_grep _stats_`);
      if ( scalar(@stats_list) == 0 ) { last; }
      my $stats_onetime = &OneTimeStatsList( \@stats_list );
      if ( scalar( @{$stats_onetime} ) == 0 ) {
        last;
      }
      if ( &Check_iostats_files( \@active_nodes, $stats_onetime ) ) {

        # All expected IO stats files
        message( "Process IOSTAT files - Time: " . $time_last . "  Nodes: " . scalar(@active_nodes)." Last proc. time: ".$main_loop_interval_time." sec." );
        if ($debug_full) { message( "IOSTAT files:\n" . join( "\n", @{$stats_onetime} ) . "\n" ); }
        $iostat_file_missing_count = 0;
        &Process_iostats_files($stats_onetime);
        if ( &Validate_iostat_data($last_data_ref) == 1 ) { next; }
        if ( &Retrieve_previous_data == 1 )               { next; }
        if ( &Validate_iostat_data($prev_data_ref) == 1 ) { next; }
        if ( &Process_iostat_data == 1 )                  { next; }
      }
      else {
        # Some IO stats file missing, waiting ...
        $iostat_file_missing_count++;
        if ( $iostat_file_missing_count <= $iostat_file_missing_max ) {
          &message( "Some IOSTAT file missing in " . $time_last . " interval. Waiting (" . $iostat_file_missing_count . ")." );
          if ($debug_full) { message( "IOSTAT files:\n" . join( "\n", @{$stats_onetime} ) . "\n" ); }
          $time_last = $time_last_save;
          last;
        }
        else {
          &message( "Some IOSTAT file missing in " . $time_last . " interval. The maximum wait time has been exceeded." );
          if ($debug_full) { message( "IOSTAT files:\n" . join( "\n", @{$stats_onetime} ) . "\n" ); }
        }
      }
    }
    continue {
      # Clear Data
      &ClearData();

      # Clear Accumulated Data
      &SVC::ClearAccumData();
      if ($debug_full) { message("IOSTAT files loop done"); }
    }
  }    ## if ( $storage_info{'is_fs840'} )
  if ($debug_full) { $time = localtime(); message( "MAIN LOOP END. --- " . $time ); }
}
continue {
  # Disconnect SSH Session
  if ($net_openssh) {
    if ($debug_full) { &message("Close SSH session"); }
    my $out = SVC::openssh_disconnect( $net_ssh, $ssh_timeout );
    if ( not defined $out ) {
      &error("Unable to disconnect from remote machine");
    }
  }
  my $max;
  if ( $storage_info{'is_fs840'} ) {
    $max = 160;    # Wait about 480 seconds
  }
  else {
    my $main_loop_end_time      = time();
    $main_loop_interval_time = $main_loop_end_time - $main_loop_start_time;
    if ($debug_full) { message( "Main loop interval time " . $main_loop_interval_time . " sec." ); }
    if ( $main_loop_interval_time >= ( $interval * 60 ) ) {
      $max = 0;    # No wait
    }
    else {
      $max = ( ( $interval * 60 ) - $main_loop_interval_time ) / 3;    # Wait about 150 seconds
    }
  }    ## if ( $storage_info{'is_fs840'} )

  ## Wait
  my $maxsec = 3 * $max;
  if ($debug_full) { message( "Main loop done. Wait about " . $maxsec . " sec." ); }
  for ( my $i = 0; $i < $max; $i++ ) {

    # Terminate program?
    if ($terminate_pgm) {
      if ($debug) { message("Exitting"); }
      exit(3);
    }
    sleep(3);
  }

  ## Continue
  # Read etc/.magic file
  unless ($xormonNG) { &Read_magic(); }

  # Open remote session to the storage (Net::OpenSSH)
  if ($net_openssh) {
    if ($debug_full) { &message( "Open SSH session - host: " . $svc . ", user: " . $user . " key: " . $key ); }
    $net_ssh = SVC::openssh_new( $svc, "22", $user, $key, $ssh_timeout, $password, $password_enable );
    if ( not defined $net_ssh ) {
      &error("Unable to connect to remote machine");
      exit 2;
    }
  }
}

sub message {
  my $msg = shift;
  my $tm  = localtime();
  print( $tm. " - INFO    - svcperf.pl: " . $msg . "\n" );
}

sub warning {
  my ( $msg, $rc ) = @_;
  my $tm = localtime();
  print STDERR ( $tm . " - WARNING - svcperf.pl: " . $msg . "\n" );
}

sub error {
  my ( $msg, $rc ) = @_;
  my $tm = localtime();
  if ( defined $rc ) {
    print STDERR ( $tm . " - ERROR   - svcperf.pl: " . $msg . " - Return code: " . $rc . "\n" );
  }
  else {
    print STDERR ( $tm . " - ERROR   - svcperf.pl: " . $msg . "\n" );
  }
}

sub LsSystem {
  my $line;
  my $delim = ":";
  my %cfg;
  my $cmd = "svcinfo lssystem -delim " . $delim;
  if ($debug_full) { &message( "Remote Command: " . $cmd ); }
  my $fh;
  if ($net_openssh) {
    $fh = SVC::openssh_pipeout( $net_ssh, [$cmd], $ssh_timeout );
  }
  else {
    $fh = SVC::cmdssh_pipeout( $cmd_ssh, [$cmd], $ssh_timeout );
  }
  if ( not defined $fh ) {
    &error( "Remote Command: " . $cmd . " failed.", 1 );
    return (undef);
  }
  while ( $line = <$fh> ) {
    chomp($line);
    if ( $line =~ /^(\w+)$delim(.*)$/ ) {    # system
      $cfg{$1} = $2;
    }
  }
  close($fh);
  return ( \%cfg );
}

sub ID_front_panel {
  my $node = shift;
  if ( defined $node_front_panel{$node} ) {
    return ( $node_front_panel{$node} );
  }
  else {
    my $front_panel_id;

    # Create $front_panel_id
    my $cmd = "svcinfo lsnodevpd -delim : " . $node;
    if ($debug_full) { &message( "Remote Command: " . $cmd ); }
    my $fh;
    if ($net_openssh) {
      $fh = SVC::openssh_pipeout( $net_ssh, [$cmd], $ssh_timeout );
    }
    else {
      $fh = SVC::cmdssh_pipeout( $cmd_ssh, [$cmd], $ssh_timeout );
    }
    if ( not defined $fh ) {
      &error( "Remote Command: " . $cmd . " failed.", 1 );
      return undef;
    }
    while ( my $line_LSNODEVPD = <$fh> ) {

      #if($debug_full) {print("The svcinfo lsnodevpd command output: $line_LSNODEVPD");}
      my @field_LSNODEVPD = split( ':', $line_LSNODEVPD );
      if ( $field_LSNODEVPD[0] =~ $iostatsdumps_name ) {

        #if($debug_full) {message("Check: $field_LSNODEVPD[1]");}
        $front_panel_id = $field_LSNODEVPD[1];
        chomp($front_panel_id);
        $node_front_panel{$node} = $front_panel_id;
      }
    }
    close($fh);
    if ( not defined $node_front_panel{$node} ) {
      warning( "Front panel ID/Dump name for node:".$node." is not defined. IOstat dumps name:". $iostatsdumps_name );
      return undef;
    }
    if ($debug_full) { message("Front panel ID/Dump name: $front_panel_id"); }
    return ( $node_front_panel{$node} );
  }
}

sub List_iostats_create {
  my $cmd_list         = shift;
  my $node_front_panel = shift;

  # create CLI command
  my $cmd;
  if ($net_openssh) {
    $cmd = " ";
  }
  else {
    $cmd = " \"";
  }
  foreach my $node ( keys( %{$cmd_list} ) ) {
    #
    $cmd .= " echo __LSIOSTATSDUMPS__:" . $node . ";";
    $cmd .= $cmd_list->{$node};
  }
  $cmd .= " echo __END__;";
  if ($net_openssh) {
    $cmd .= " ";
  }
  else {
    $cmd .= "\"";
  }

  my %stats_files;
  my $nd;
  my $flag = 0;    # 0 ... End; 1 ... node;
  if ($debug_full) { &message( "Remote Command: " . $cmd ); }
  my $fh;
  if ($net_openssh) {
    $fh = SVC::openssh_pipeout( $net_ssh, [$cmd], $ssh_timeout );
  }
  else {
    $fh = SVC::cmdssh_pipeout( $cmd_ssh, [$cmd], $ssh_timeout );
  }
  if ( not defined $fh ) {
    &error( "Remote Command: " . $cmd . " failed.", 1 );
    return undef;
  }
  while ( my $line = <$fh> ) {
    chomp($line);
    if ( $line =~ /^__END__$/ ) { last; }                     # End of commands output
    elsif ( $line =~ /^__LSIOSTATSDUMPS__:([-\.\w]+)$/ ) {    # First entry by node
      $flag = 1;
      $nd   = $1;
      my @a;
      $stats_files{$nd} = \@a;
      next;
    }
    else {

      # create local node's I/O stats file list array (@stats_files)
      #if($debug_full) {print("The svcinfo lsiostatsdumps command output: $line_LSIOSTATSDUMPS");}
      my @field_LSIOSTATSDUMPS = split( ':', $line );
      chomp( $field_LSIOSTATSDUMPS[1] );
      if ( index( $field_LSIOSTATSDUMPS[1], $node_front_panel->{$nd} ) > 0 ) {
        unshift( @{ $stats_files{$nd} }, $field_LSIOSTATSDUMPS[1] );
      }
    }
  }
      #@stats_files = sort(@stats_files);
      #if ($debug_full) { message("I/O statistics log files: @stats_files"); }
  return ( \%stats_files );
}

sub List_iostats_upload {
  my ( $config_node, $node_front_panel, $stats_files ) = @_;
  my $_cmd_ssh;
  my $upload_files = [];
  my $flg_ssh      = 0;
  my $flg_scp      = 0;

  # Prepare commands
  if ($net_openssh) {
    $_cmd_ssh = " ";
  }
  else {
    $_cmd_ssh = " \"";
  }
  foreach my $node ( keys( %{$node_front_panel} ) ) {

    # Scan all (Node, VDisk, MDisk, Drive) stats files
    while ( defined( $stats_file = shift( @{ $stats_files->{$node} } ) ) ) {

      # Handle Node stats file
      if ( not defined $Nn_last[$node] ) { $Nn_last[$node] = "" }
      if ( $stats_file =~ m/Nn/ && $stats_file gt $Nn_last[$node] ) {
      # Cannot copy data - cpdumps (fw 8.3.1.2 and later)
      if ( $codevrmn < 8030102 && $node != $config_node ) {    # non-config node
          $flg_ssh++;
          $_cmd_ssh .= "svctask cpdumps -prefix /dumps/iostats/" . $stats_file . " " . $node . "\;";
          $_cmd_ssh .= "sleep 4\;";
        }
        $flg_scp++;
        push( @{$upload_files}, "/dumps/iostats/" . $stats_file );

        #$Nn_last[$node] = $stats_file;
      }

      # Handle VDisk stats file
      if ( not defined $Nv_last[$node] ) { $Nv_last[$node] = "" }
      if ( $stats_file =~ m/Nv/ && $stats_file gt $Nv_last[$node] ) {
        if ( $codevrmn < 8030102 && $node != $config_node ) {    # non-config node
          $flg_ssh++;
          $_cmd_ssh .= "svctask cpdumps -prefix /dumps/iostats/" . $stats_file . " " . $node . "\;";
          $_cmd_ssh .= "sleep 4\;";
        }
        $flg_scp++;
        push( @{$upload_files}, "/dumps/iostats/" . $stats_file );

        #$Nv_last[$node] = $stats_file;
      }

      # Handle MDisk stats file
      if ( not defined $Nm_last[$node] ) { $Nm_last[$node] = "" }
      if ( $stats_file =~ m/Nm/ && $stats_file gt $Nm_last[$node] ) {
        if ( $codevrmn < 8030102 && $node != $config_node ) {    # non-config node
          $flg_ssh++;
          $_cmd_ssh .= "svctask cpdumps -prefix /dumps/iostats/" . $stats_file . " " . $node . "\;";
          $_cmd_ssh .= "sleep 4\;";
        }
        $flg_scp++;
        push( @{$upload_files}, "/dumps/iostats/" . $stats_file );

        #$Nm_last[$node] = $stats_file;
      }

      # Handle Drive stats file
      if ( not defined $Nd_last[$node] ) { $Nd_last[$node] = "" }
      if ( $stats_file =~ m/Nd/ && $stats_file gt $Nd_last[$node] ) {
        if ( $codevrmn < 8030102 && $node != $config_node ) {    # non-config node
          $flg_ssh++;
          $_cmd_ssh .= "svctask cpdumps -prefix /dumps/iostats/" . $stats_file . " " . $node . "\;";
          $_cmd_ssh .= "sleep 4\;";
        }
        $flg_scp++;
        push( @{$upload_files}, "/dumps/iostats/" . $stats_file );

        #$Nd_last[$node] = $stats_file;
      }
    }    ## while ( defined( $stats_file ...
  }    ## foreach my $node ( keys(%{$node_front_panel}) )
  if ($net_openssh) {
    $_cmd_ssh .= " ";
  }
  else {
    $_cmd_ssh .= "\"";
  }

  # Copy all iostats files from non-config node to config node
  if ($flg_ssh) {
    if ($debug_full) { &message( "Copy iostat file.      [" . $flg_ssh . "]\nRemote Command: " . $_cmd_ssh ); }
    my $fh;
    if ($net_openssh) {
      $fh = SVC::openssh_pipeout( $net_ssh, [$_cmd_ssh], $ssh_timeout );
    }
    else {
      $fh = SVC::cmdssh_pipeout( $cmd_ssh, [$_cmd_ssh], $ssh_timeout );
    }
    if ( not defined $fh ) {
      &error( "Remote Command: " . $_cmd_ssh . " failed.", 1 );
      return undef;
    }

    while ( my $line = <$fh> ) {
      if ($debug_full) { chomp($line); &message( "  " . $line ); }
    }
    close($fh);
    sleep(5);
  }

  # Upload iostats files for config node
  if ($flg_scp) {
    if ($debug_full) { message( "Download files: " . join( " ", @{$upload_files} ) . ", in to directory " . $iosdir ); }
    my $out;
    if ($net_openssh) {
      $out = SVC::openssh_getfile( $net_ssh, $upload_files, $iosdir, $ssh_timeout );
    }
    else {
      $out = SVC::cmdssh_getfile( $cmd_scp, $upload_files, $iosdir, $ssh_timeout );
    }
    if ( not defined $out ) {
      &error( "Download files: " . join( " ", @{$upload_files} ) . ", failed", 1 );
      return undef;
    }
  }
}

sub OneTimeStatsList {
  my $stats_list = shift;
  my ( @tlist, @a );

  # Get timestamp form filenames
  foreach my $line ( @{$stats_list} ) {
    chomp($line);
    if ( $line =~ /^[-\w]+_(\d{6}_\d{6})$/ ) {
      my $tm = $1;
      push( @tlist, $tm );
    }
  }
  @tlist = sort(@tlist);

  #print STDERR Dumper(\@tlist);
  # Search next time after last time
  my $tm = shift(@tlist);
  while (@tlist) {

    if ( $tm le $time_last ) {
      $tm = shift(@tlist);
    }
    else {
      last;
    }
  }
  if ( $tm lt $time_last ) {
    return \@a;
  }
  if ( $tm eq $time_last ) {
    return \@a;
  }

  # Search file with last time
  foreach my $line ( @{$stats_list} ) {
    chomp($line);
    if ( $line =~ /^[-\w]+_$tm$/ ) {
      push( @a, $line );
    }
  }

  # Set new last time
  $time_last = $tm;
  return \@a;
}

sub Check_iostats_files {
  my ( $nodes, $stats ) = @_;
  my $nd;
  foreach $nd ( @{$nodes} ) {
    if ( not defined $nd ) { next; }
    my $num_files = 0;
    foreach my $stats_name ( @{$stats} ) {
      chomp($stats_name);
      if ( $stats_name =~ /^N[dmnv]_stats_${nd}_.*$/ ) {
        $num_files++;
      }
    }
    if ( $num_files != 4 ) { return 0; }
  }
  return 1;
}

sub Process_iostats_files {
  my $stats = shift;
  if ($debug_full) { &message("Process IOSTAT files:"); }
  foreach my $line_DIR ( @{$stats} ) {
    $stats_file = $line_DIR;
    chomp($stats_file);
    if ($debug_full) { &message( "        " . $stats_file ); }
    &SVC::Data_insert( $iosdir, $stats_file, $last_data_ref, $debug_full );
  }
}

sub Retrieve_previous_data {

  # Chech if exist previous data
  if ( not -r $tmp_perf_file ) {
    if ($debug) { &message("No previous data. Store data to file: $tmp_perf_file.\n Continue.") }
    if ( !defined nstore( $last_data_ref, $tmp_perf_file ) ) {
      message("svcperf.pl: ERROR: Store data to file $tmp_perf_file failed.");
      system( "$cmd_rm -f " . $tmp_perf_file );
      exit SVCPRF110E;
    }
    return 1;
  }

  # Retrieve previous data from temporary file
  eval { $prev_data_ref = retrieve($tmp_perf_file); };
  if ($@) {
    &message( $@ . __FILE__ . ":" . __LINE__ . " file:" . $tmp_perf_file );
    &message("Retrieve data from file $tmp_perf_file failed.");
    system( "$cmd_rm -f " . $tmp_perf_file );
    exit SVCPRF112E;
  }
  if ( !defined $prev_data_ref ) {
    &message("Retrieve data from file $tmp_perf_file failed.");
    system( "$cmd_rm -f " . $tmp_perf_file );
    exit SVCPRF112E;
  }
  return 0;
}

sub Validate_iostat_data {
  my $data_ref = shift;

  # Check previous data validity
  my @nodes = keys( %{ $data_ref->{svc_node} } );
  if ( scalar(@nodes) == 0 ) {
    if ($debug) { message("Data is not valid. Continue.") }
    return 1;
  }
  if ( not defined $data_ref->{svc_node}->{ $nodes[0] }->{timestamp} || "x$data_ref->{svc_drive}->{$nodes[0]}->{timestamp}" eq "x" ) {
    return 1;
  }
}

sub Process_iostat_data {
  my @prev_nodes = keys( %{ $prev_data_ref->{svc_node} } );
  my $prev_time  = str2time( $prev_data_ref->{svc_node}->{ $prev_nodes[0] }->{timestamp} );
  my @last_nodes = keys( %{ $last_data_ref->{svc_node} } );
  my $last_time  = str2time( $last_data_ref->{svc_node}->{ $last_nodes[0] }->{timestamp} );

  if ( $last_time < $prev_time ) {
    if ($debug) { message("svcperf.pl: Previous data is not valid. Exiting.") }
    system( "$cmd_rm -f " . $tmp_perf_file );
    exit SVCPRF117E;
  }
  if ( $last_time == $prev_time ) {
    if ($debug_full) { &message("No new data. Continue.") }
    return 1;
  }

  # Store latest data
  if ($debug_full) { &message( "svcperf.pl: Store actual data to file: " . $tmp_perf_file . "." ) }
  if ( !defined nstore( $last_data_ref, $tmp_perf_file ) ) {
    message("ERROR: Store data to file $tmp_perf_file failed.");
    system( "$cmd_rm -f " . $tmp_perf_file );
    exit SVCPRF111E;
  }

  # Set output filename
  my ( $sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst ) = localtime($last_time);
  $year += 1900;
  $mon++;
  my $date = sprintf( "%4d%2.2d%2.2d",   $year, $mon, $mday );
  my $time = sprintf( "%2.2d%2.2d%2.2d", $hour, $min, $sec );
  $out_perf_file = $tmpdir . $storage . "_svcperf_" . $date . "_" . $time . ".out";
  if ($debug_full) { &message( "Performance Output Filename: " . $out_perf_file . "." ) }

  # Get Latest Config Data
  $tmp_conf_file = &GetConfigFilename();
  if ($debug) { message( "Config Filename: " . $tmpdir . $tmp_conf_file ) }
  if ( !defined $tmp_conf_file || !-r $tmpdir . $tmp_conf_file ) {
    message("Config File not found");
    exit SVCPRF114E;
  }
  eval { $cfg_ref = retrieve( $tmpdir . $tmp_conf_file ); };
  if ($@) {
    &message( $@ . __FILE__ . ":" . __LINE__ . " file:" . $tmpdir . $tmp_conf_file );
    &message( "Retrieve data from file " . $tmpdir . $tmp_conf_file . " failed." );
    system( "$cmd_rm -f " . $tmpdir . $tmp_conf_file );
    exit SVCPRF119E;
  }

  # Check configuration data validity
  if ( !scalar( %{$cfg_ref} ) ) {
    if ($debug) { message("Configuration data is not valid. Exiting.") }
    system( "$cmd_rm -f " . $tmpdir . $tmp_conf_file );
    exit SVCPRF119E;
  }

  if ($debug_full) { &message("Configuration and Previous performance data is valid. Continue.") }

  # Get Capacity data
  #&GetCapacityData();

  # Count performance data
  $ret = &SVC::CountPerfData( $out_perf_file, $prev_data_ref, $last_data_ref, $cfg_ref->{svc_pool}, $cfg_ref, $debug_full );
  if ($ret) {
    &warning("Latest data is not valid. Continue.");
    return 1;
  }
  return 0;
}

sub GetCapacityData {
  #
  &MDiskGroupCapacityData();
}

sub MDiskGroupCapacityData {
  undef %mdiskgrp_cap;
  my $tier_level = 1;
  my $record_ref;
  #
  my $cmd;
  if ($net_openssh) {
    $cmd = " ";
    $cmd .= "svcinfo lsmdiskgrp -nohdr |while read id line; do svcinfo lsmdiskgrp -bytes -delim : \$id;echo '*****';done";
    $cmd .= " ";
  }
  else {
    $cmd = " \"";
    $cmd .= "svcinfo lsmdiskgrp -nohdr |while read id line; do svcinfo lsmdiskgrp -bytes -delim : \\\$id;echo '*****';done";
    $cmd .= "\"";
  }

  if ($debug_full) { &message( "Remote Command: " . $cmd ); }
  my $fh;
  if ($net_openssh) {
    $fh = SVC::openssh_pipeout( $net_ssh, [$cmd], $ssh_timeout );
  }
  else {
    $fh = SVC::cmdssh_pipeout( $cmd_ssh, [$cmd], $ssh_timeout );
  }
  if ( not defined $fh ) {
    &error( "Remote Command: " . $cmd . " failed.", 1 );
    return undef;
  }

  while ( my $line = <$fh> ) {
    chomp($line);
    if ( $line eq "*****" ) {

      # Record last line
      if ( not defined $record_ref->{name} and defined $record_ref->{parent_mdisk_grp_name} ) {    # FS840
        $record_ref->{name} = $record_ref->{parent_mdisk_grp_name};
      }
      $mdiskgrp_cap->{ $record_ref->{name} } = $record_ref;
      next;
    }
    my @a = split( ':', $line );
    if ( $a[0] eq "id" ) {

      # New record first line
      my %rec;
      $record_ref = \%rec;
      &InitMdiskGrpRecord($record_ref);
      $record_ref->{ $a[0] } = $a[1];
    }
    elsif ( $a[0] eq "site_id" && !defined $a[1] ) {
      $record_ref->{ $a[0] } = "";
    }
    elsif ( $a[0] eq "site_name" && !defined $a[1] ) {
      $record_ref->{ $a[0] } = "";
    }
    elsif ( defined $a[0] && !defined $a[1] ) {

      # Value is not set - suppress
      next;
    }
    elsif ( $a[0] eq "tier" && $a[1] eq "generic_ssd" || $a[1] eq "ssd" || $a[1] eq "tier0_flash" ) {    # set TIER Level 0
      $tier_level = 0;
      $record_ref->{tier_0} = $a[1];
    }
    elsif ( $a[0] eq "tier" && $a[1] eq "tier1_flash" ) {                                                # set TIER Level 0
      $tier_level = "0n";
      $record_ref->{tier_0n} = $a[1];
    }
    elsif ( $a[0] eq "tier" && $a[1] eq "tier_scm" ) {                                                   # set TIER Level 0
      $tier_level = "scm";
      $record_ref->{tier_scm} = $a[1];
    }
    elsif ( $a[0] eq "tier" && $a[1] eq "generic_hdd" || $a[1] eq "enterprise" || $a[1] eq "tier_enterprise" ) {    # set TIER Level 1
      $tier_level = 1;
      $record_ref->{tier_1} = $a[1];
    }
    elsif ( $a[0] eq "tier" || $a[1] eq "nearline" || $a[1] eq "tier_nearline" ) {                                  # set TIER Level 2
      $tier_level = 2;
      $record_ref->{tier_2} = $a[1];
    }
    elsif ( $a[0] eq "tier_mdisk_count" ) {
      $record_ref->{ "tier_" . $tier_level . "_mdisk_count" } = $a[1];
    }
    elsif ( $a[0] eq "tier_capacity" ) {
      $record_ref->{ "tier_" . $tier_level . "_capacity" } = $a[1];
    }
    elsif ( $a[0] eq "tier_free_capacity" ) {
      $record_ref->{ "tier_" . $tier_level . "_free_capacity" } = $a[1];
    }
    else {
      $record_ref->{ $a[0] } = $a[1];
    }
  }
  close($fh);
}

sub ListMdiskGrp {
  #
  my @list;
  my $cmd = "svcinfo lsmdiskgrp -nohdr -delim :";
  if ($debug_full) { &message( "Remote Command: " . $cmd ); }
  my $fh;
  if ($net_openssh) {
    $fh = SVC::openssh_pipeout( $net_ssh, [$cmd], $ssh_timeout );
  }
  else {
    $fh = SVC::cmdssh_pipeout( $cmd_ssh, [$cmd], $ssh_timeout );
  }
  if ( not defined $fh ) {
    &error( "Remote Command: " . $cmd . " failed.", 1 );
    return undef;
  }
  while ( my $line_LSMDISKGRP = <$fh> ) {
    my @field_LSMDISKGRP = split( ':', $line_LSMDISKGRP );
    #
    push( @list, $field_LSMDISKGRP[0] );
  }
  return ( join( " ", @list ) );
}

sub InitMdiskGrpRecord {
  my $c = shift;
  $c->{tier_0}               = "ssd";
  $c->{tier_0_mdisk_count}   = 0;
  $c->{tier_0_capacity}      = 0;
  $c->{tier_0_free_capacity} = 0;
  $c->{tier_1}               = "enterprise";
  $c->{tier_1_mdisk_count}   = 0;
  $c->{tier_1_capacity}      = 0;
  $c->{tier_1_free_capacity} = 0;
  $c->{tier_2}               = "nearline";
  $c->{tier_2_mdisk_count}   = 0;
  $c->{tier_2_capacity}      = 0;
  $c->{tier_2_free_capacity} = 0;
}

sub GetConfigFilename {
  my @a;
  foreach my $line_LS (`$cmd_ls $tmpdir`) {
    chomp($line_LS);

    # Select Config Data file
    if ( $line_LS =~ /^${storage}_svcconf_\d+_\d+.data$/ ) {
      push( @a, $line_LS );
    }
  }
  my @b = sort @a;
  return ( pop @b );
}

sub ClearData {
  undef %last_data;
  undef %prev_data;
  undef %mdiskgrp_cap;
}

sub Localtime_str {

  # OUT: 150425_095000
  my ( $sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst ) = localtime();
  my $y   = $year - 100;
  my $m   = $mon + 1;
  my $str = sprintf( "%02d%02d%02d_%02d%02d%02d", $y, $m, $mday, $hour, $min, $sec );
  return ($str);
}

sub Read_magic {
  my $magic_file = $ENV{BINDIR} . "/../etc/.magic";
  if ( -r $magic_file ) {
    my $ret = open( MAGIC, $magic_file );
    if ( $ret != 1 ) { warning( "Cannot open magic file: " . $magic_file ); return; }
    my $debug_full_pre = $debug_full;
    while ( my $line = <MAGIC> ) {
      chomp($line);
      if ( $line =~ /^\s*#.*$/ ) { next; }
      if ( $line =~ /^.*SVC_DEBUG_FULL\s*=\s*(\d{1})\s*$/ ) {
        $debug      = $1;
        $debug_full = $1;
      }
      elsif ( $line =~ /^.*SVC_DEBUG_FULL\s*=\s*([\w -\.]*)\s*$/ ) {
        if ( $1 eq $storage ) {
          $debug      = 1;
          $debug_full = 1;
        }
      }
    }
    if ( $debug_full_pre == 0 && $debug_full ) {
      message("Enable full debug mode");
    }
    elsif ( $debug_full_pre && $debug_full == 0 ) {
      message("Disble full debug mode");
    }
    close(MAGIC);
  }
}

sub cmd_exec {
  my $cmd = shift;
  $ret = 1;
  my $out;
  my $count = 0;    # let it run only 5 times ( 10 x 80 sec timeout in ssh = 800secs)
  while ( $ret && !$terminate_pgm ) {
    $count++;

    #if($debug_full) {message("Run: $cmd");}
    ( $ret, $out ) = &runcmd("$cmd 2>&1");
    if ( $ret && $count < 5 ) {
      for my $line ( @{$out} ) {
        chomp($line);
        if ( $line =~ /No such file or directory/ ) {
          return 2;
        }
        else {
          warning( "Command output: " . $line );
        }
      }
      warning("Wait 5s and Retry $cmd");
      sleep 5;
      next;
    }
    if ( $ret && $count == 5 ) {
      error("Failed: $cmd : too many attempts : $count");
      return 1;
    }
  }
  return $ret;
}

sub runcmd {
  my $cmd = "@_";
  my $ret_value;
  my $msg;
  my $rc = 0;
  my @out;
  #
  if ($debug_full) { message("runcmd: Execute: @_"); }
  ( @out = `@_` );
  $ret_value = $?;
  if ( $ret_value == -1 ) {
    $msg = sprintf "runcmd: Failed to execute: $!\n";
    if ($debug_full) { message($msg); }
    $rc = -1;
  }
  elsif ( $ret_value & 127 ) {
    $msg = sprintf "runcmd: Command \'%s\' died with signal %d\n", $cmd, ( $ret_value & 127 );
    if ($debug_full) { message($msg); }
    $rc = ( $ret_value & 127 );
  }
  else {
    $msg = sprintf "runcmd: Command \'%s\' exited with value %d\n", $cmd, $ret_value >> 8;
    if ($debug_full) { message($msg); }
    $rc = ( $ret_value >> 8 );
  }
  return ( $rc, \@out );
}

sub SetStorageInfo {
  my $lssys  = shift;
  my $isprod = &ProductName( $lssys->{product_name} );
  if ( defined $isprod ) {
    $storage_info{$isprod} = 1;
  }
  if ($debug_full) {
    my $msg;
    my @a;
    foreach my $k ( sort( keys(%storage_info) ) ) {
      push( @a, $k . "=" . $storage_info{$k} );
    }
    message( "HW Type Info: " . join( ",", @a ) . "\n" );
  }
}

sub ProductName {
  my $prod = shift;
  if ( defined $prod ) {

    # IBM Storwize ('product_name' => 'IBM Storwize V7000')
    return ("is_storwize") if ( $prod =~ m/^IBM\s+Storwize\s+V/ );

    # IBM Storwize ('product_name' => 'Lenovo Storage V3700 V2 XP')
    return ("is_storwize") if ( $prod =~ m/^Lenovo\s+Storage\s+V3700/ );

    # IBM FlashSystem ('product_name' => 'IBM SAN Volume Controller')
    return ("is_svc") if ( $prod =~ m/^IBM\s+SAN\s+Volume\s+Controller/ );

    # IBM FlashSystem V9000 ('product_name' => 'IBM FlashSystem V9000')
    return ("is_v9000") if ( $prod =~ m/^IBM\s+FlashSystem\s+V9000$/ );

    # IBM FlashSystem 9100 ('product_name' => 'IBM FlashSystem 9100')
    return ("is_fs9100") if ( $prod =~ m/^IBM\s+FlashSystem\s+9100$/ );

    # Inspur AS5600/5500/2600/2200 ('product_name' => 'Inspur AS2600G2') --> it uses this xml file name: mcs.config.cron.xml_9793365-2
    return ("is_inspur") if ( $prod =~ m/^Inspur\s+AS/ );

    # IBM FlashSystem 840/900
    return ("is_fs840") if ( $prod =~ m/^IBM\s+FlashSystem\s+840$/
      || $prod =~ m/^IBM\s+FlashSystem\s+900$/ );
    return ("is_unknown");
  }
  return (undef);
}

sub Retrieve_prev_perf_data {

  # Retrieve previous data from temporary file
  eval { $prev_data_ref = retrieve($tmp_perf_file); };
  if ($@) {
    &message( $@ . __FILE__ . ":" . __LINE__ . " file:" . $tmp_perf_file );
    &message("Retrieve data from file $tmp_perf_file failed.");
    &runcmd( "$cmd_rm -f " . $tmp_perf_file );
    exit SVCPRF112E if ( not $storage_info{is_fs840} );
  }
  if ( !defined $prev_data_ref ) {
    &message("Retrieve data from file $tmp_perf_file failed.");
    &runcmd( "$cmd_rm -f " . $tmp_perf_file );
    exit SVCPRF112E if ( not $storage_info{is_fs840} );
  }
}    ## Retrieve_prev_perf_data

sub Retrieve_last_conf_data {

  # Get Latest Config Data
  $tmp_conf_file = &GetConfigFilename();
  if ($debug) { message( "Config Filename: " . $tmpdir . $tmp_conf_file ) }
  if ( !defined $tmp_conf_file || !-r $tmpdir . $tmp_conf_file ) {
    message("Config File not found");
    exit SVCPRF114E;
  }
  eval { $cfg_ref = retrieve( $tmpdir . $tmp_conf_file ); };
  if ($@) {
    &message( $@ . __FILE__ . ":" . __LINE__ . " file:" . $tmpdir . $tmp_conf_file );
    &message( "Retrieve data from file " . $tmpdir . $tmp_conf_file . " failed." );
    system( "$cmd_rm -f " . $tmpdir . $tmp_conf_file );
    exit SVCPRF119E;
  }

  # Check configuration data validity
  if ( !scalar( %{$cfg_ref} ) ) {
    if ($debug) { message("Configuration data is not valid. Exiting.") }
    system( "$cmd_rm -f " . $tmpdir . $tmp_conf_file );
    exit SVCPRF119E;
  }

  if ($debug_full) { &message("Configuration and Previous performance data is valid. Continue.") }

}

sub Store_prev_perf_data {
  if ($debug_full) { &message( "Store data to file " . $tmp_perf_file ); }
  if ( !defined nstore( $prev_data_ref, $tmp_perf_file ) ) {
    error( "SVCPRF121E Store data to file $tmp_perf_file failed.", SVCPRF121E );
    runcmd( "$cmd_rm -f " . $tmp_perf_file );
    exit SVCPRF121E;
  }
}

###############################################################################
#
# Sig_Handler
#
# 1. Catch SIGINT and SIGTERM
# 2. Close SSH Session
# 3. Ternimate process
#
###############################################################################

sub Sig_Handler {
  my $sig = shift;
  $terminate_pgm = 1;
  if ($debug_full) { &message( "Terminating by SIG" . $sig . "\n" ); }
  if ($net_openssh) {
    if ($debug_full) { &message("Close SSH session"); }
    my $out = SVC::openssh_disconnect( $net_ssh, $ssh_timeout );
    if ( not defined $out ) {
      if ($debug_full) { &error("Unable to disconnect from remote machine"); }
    }
  }
}

sub epoch2datestr {

  # Vystup: 2012-04-25 09:50:00 GMT+01:00
  my $tm = shift;
  my ( $sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst ) = localtime($tm);
  my $y   = $year + 1900;
  my $m   = $mon + 1;
  my $tz  = tzoffset();
  my $str = sprintf( "%4s-%02s-%02s %02s:%02s:%02s %s", $y, $m, $mday, $hour, $min, $sec, $tz );
  return ($str);
}

sub tzoffset {
  my $sign       = "";
  my $t          = time();
  my $local_time = localtime($t);
  my $gm_time    = gmtime($t);
  my $local      = str2time($local_time);
  my $utc        = str2time($gm_time);
  my $tzo        = ( ( $local - $utc ) / 3600 );
  if ( $tzo > 0 ) { $sign = "+"; }
  return sprintf( "%s%s%02d:00", "GMT", $sign, $tzo );
}

sub GetCodeLevelNum {
  #
  my %code_level;
  if ( defined $cfg_ref->{'svc_system'}->{'code_level'} && 
               $cfg_ref->{'svc_system'}->{'code_level'} =~ /^(\d+)\.(\d+)\.(\d+)\.(\d+)\s+\(build\s+(.*)\).*$/ ) {
    return( $1 * 1000000 + $2 * 10000 + $3 * 100 + $4 );
  }
  else {
    return(0);
  }
}