package Unity;

use strict;
use warnings;

use FindBin;
use lib "$FindBin::Bin/"; # use lib in this folder
use lib "../../../perl-modules";
use Data::Dumper;
use Xormon qw(VrDump);
use XormonLogger; # for this microservice
use Cmd;
use Parsing qw(string2arrayLikeUnity);
use Test;
use JSON;
use Time::Piece;
use POSIX qw(strftime);

use constant 'DATA_LOADING_CONFIGURATION_FILENAME' => './data_loading_configuration.json';
use constant 'ONLY_CONFIGURED_METRICS' => 0;
use Time::Local;
use Time::Zone;
use POSIX qw(tzset);

my $tzoffset;

BEGIN {
  our $start_time = time();
}

my $xormon;
my $cfg;
my $service_name;
my $logger;
my $cfg_time = 1;
my %data;
my %uid_cache;
my @architecture;
my $cmd;
my $act_time;
my @availableMetrics;
my $DEBUG;
my $PERF_COUNT;
my $CMD_TIMEOUT;
my $PERF_SAMPLE_RATE;


# constructor, initializes the loggeer and the $cmd basic settings
sub init {
  my $config = shift;
  $logger = XormonLogger->new('unity', $config->{'hostcfg_id'}, $config->{'hostalias'});

  # setting variables from the configuration, instead of constants
  $DEBUG = $config->{'debug'} ? $config->{'debug'} : 0;
  $PERF_COUNT = $config->{'perf_count'} ? $config->{'perf_count'} : 1;
  $CMD_TIMEOUT = $config->{'timeout'} ? $config->{'timeout'} : 300;
  $PERF_SAMPLE_RATE = $config->{'perf_sample_rate'} ? $config->{'perf_sample_rate'} : 300;
  # time shift setting - 0 => always UTC
  $tzoffset = $config->{'tzoffset'} ? $config->{'tzoffset'} : 0;


  # basic settings of the client cli application
  $cmd = Cmd->new($config->{'cmd'}, $DEBUG);
  $cmd->setTimeout($CMD_TIMEOUT);
  $cmd->set_logger($logger);
  my $regex = qr/^(.*-p\s)(\S+)(\s.*$)/mp;
  my $subst = '"$1password_removed$3"';
  $cmd->set_password_remove_regex($regex, $subst);
  $cmd->addCoreParam('-d', $config->{'host'}, 1);
  $cmd->addCoreParam('-u', $config->{'username'}, 1);
  $cmd->addCoreParam('-p', $config->{'password'}, 1);
  $cmd->addCoreParam('-silent', '', 0);
  $cmd->addCoreParam('-gmtoff', $tzoffset, 1);
  $cmd->addCoreParam('-noHeader', '', 0);
  return 1;
}


# basic function to collect data from unity storage
# performs basic setup and collection of all data
# @Param
#   @var $xormon XORMON - xormon service
#   @var $cfg hash - microservice configuration data
#   @var $service_name - string -micro service name
#
sub collect {
  $xormon = shift;
  $cfg = shift;
  $service_name = shift;
  $act_time = time();

  # Loading metrics configuration
  my %data_loading_configuration = %{Xormon::read_json(DATA_LOADING_CONFIGURATION_FILENAME, 0)};

  $logger->info("Starting collect conf information") if( $DEBUG >= 1 );
  # collection of configuration data for the device

  collect_conf($data_loading_configuration{'conf'}, $data_loading_configuration{'globalMetricSettings'});
  if( keys(%data) > 0 ) {

    $logger->info("Starting affter conf process") if( $DEBUG >= 1 );
    after_conf_uni($data_loading_configuration{'after_conf'});

    $logger->info("Adding subsystems to parents") if( $DEBUG >= 1 );
    add_subsystem_to_parent($data_loading_configuration{'add_subsystem_to_parent'});

    $logger->info("retrieve available metrics for the device") if( $DEBUG >= 1 );
    collect_availableMetrics($data_loading_configuration{'availableMetrics'});

    if( @availableMetrics > 0 ) {
      $logger->info("collection of performance data of the device") if( $DEBUG >= 1 );
      collect_perf($data_loading_configuration{'perf'}, $data_loading_configuration{'globalMetricSettings'});
    }

    $logger->info("Starting data aggregation") if( $DEBUG >= 1 );
    aggregate_uni($data_loading_configuration{'aggregate'});

    $logger->info("Starting affter aggregate process") if( $DEBUG >= 1 );
    after_aggregate_uni($data_loading_configuration{'after_aggregate'});

    $logger->info("Starting a generation of architectures") if( $DEBUG >= 1 );
    $data{'architecture'} = \@{generate_architecture($data_loading_configuration{'architecture'})};

    save_hash_to_json(\%data) if( $DEBUG >= 1 );
    $logger->info("Clean up final data") if( $DEBUG >= 1 );
    clean_up($data_loading_configuration{'clean_up'});

    # for some unknown reason these handles cannot be deleted from the clean_up function.
    # I have not been able to discover why. so I delete directly in the code :(
    delete($data{'CONF'}{'device_tmp'});
    delete($data{'PERF'}{'unity'}{'volume'}{'spa'});
    delete($data{'PERF'}{'unity'}{'volume'}{'spb'});
    delete($data{'PERF'}{'unity'}{'drive'}{'spa'});
    delete($data{'PERF'}{'unity'}{'drive'}{'spb'});

    $logger->info("final structure change") if( $DEBUG >= 1 );
    final_structure_change($data_loading_configuration{'final_structure_change'});
    $logger->info("generating health status") if( $DEBUG >= 1 );
    health_status($data_loading_configuration{'health_status'});

    save_hash_to_json(\%data, "_clean_up") if( $DEBUG >= 1 );
  }
  return \%data;
}

# collection of configuration data for the device
# @param
#  @var $config_ref - hash reference from the "conf" section - from the metric configuration
#  @var $globalMetricSettings_ref - hash reference from the "globalMetricSettings" section - from the metric configuration
sub collect_conf {
  my $config_ref = shift;
  my $globalMetricSettings_ref = shift;
  my @config = @{$config_ref};
  for my $item_ref ( @config ) {
    collect_uni($item_ref, $globalMetricSettings_ref);
  }
  return 1;
}

# function populates the @availableMetrics field with the metrics available on the device
# @var $config_ref - hash reference from the "availableMetrics" section - from the metric configuration
sub collect_availableMetrics {
  my $config_ref = shift;
  if( ref $config_ref ) {
    my %config = %{$config_ref};
    my $result_ref = collect_uni(\%config);
    if( ref($result_ref) eq 'HASH' ) {
      @availableMetrics = sort keys %{$result_ref};
    }
  }
  return 1;
}

# collection of performance data for the device
# @param
#  @var $config_ref - hash reference from the "perf" section - from the metric configuration
#  @var $globalMetricSettings_ref - hash reference from the "globalMetricSettings" section - from the metric configuration
# needs the @availableMetrics array to be loaded - using the collect_availableMetrics function
sub collect_perf {
  my $config_ref = shift;
  my $globalMetricSettings_ref = shift;
  if( ref $config_ref ) {
    my @config = @{$config_ref};
    for my $item_ref ( @config ) {
      my %item = %{$item_ref};
      my $path = preparePath($item{'cmd'});
      $item{'cmd'} = "/metrics/value/hist -path \"" . $path . "\" show -count " . $PERF_COUNT . " -interval " . $PERF_SAMPLE_RATE . " -output nvp";
      collect_uni(\%item, $globalMetricSettings_ref);
    }
  }
  return 1;
}

# collect funtion
sub collect_uni {
  my ( $config_ref, $globalMetricSettings_ref ) = @_;
  my %config = %{$config_ref};
  my %globalMetricSettings;
  if( ref($globalMetricSettings_ref) eq 'HASH' ) {
    %globalMetricSettings = %{$globalMetricSettings_ref};
  }
  if( defined($config{'description'}) ) {
    $logger->info($config{'description'})
  }

  my %output;
  my %preparedData;
  $cmd->resetParams();
  my $ID;
  my $idName = $config{'id'} // 'id';
  my %sections;

  if( exists($config{'cmd'}) ) {
    $cmd->addParam($config{'cmd'}, '', 0);
    my ( $string, $error ) = $cmd->run();
    if( !defined($error) ) {
      my $out = string2arrayLikeUnity($string);
      for my $item ( @{$out} ) {
        my ( $section, $name, $value ) = @{$item};
        $name = lc($name); # Převést na malá písmena
        $name =~ s/[ \/]/_/g;
        $sections{$section}++;
        if( $name eq $idName ) {
          $ID = $value;
        }
        my %settingMetrics;

        if( exists($config{'items'}{$name}) ) {
          %settingMetrics = %{$config{'items'}{$name}};
        }
        elsif( exists($globalMetricSettings{$name}) ) {
          %settingMetrics = %{$globalMetricSettings{$name}};
        }
        else {
          if( ONLY_CONFIGURED_METRICS ) {
            next;
          }
        }

        if( %settingMetrics ) {

          if( exists($settingMetrics{'copy'}) ) {
            my @newItem = ( $section, $settingMetrics{'copy'}, $value );
            push(@{$out}, \@newItem);
          }

          if( exists($settingMetrics{'callback'}) ) {
            my @callbacks = split(',', $settingMetrics{'callback'});
            for my $callback ( @callbacks ) {
              my @params = ();
              if( defined($callback) ) {
                no strict 'refs';
                # I can only send the ID to the callback function, I don't have all the items of the section
                $value = &{$callback}($value, $ID, @params);
                use strict 'refs';
              }
            }
          }

          if( exists($settingMetrics{'rename'}) ) {
            $name = $settingMetrics{'rename'};
          }

          if( exists($settingMetrics{'returnBytes'}) ) {
            $value = Xormon::returnBytes($value, $settingMetrics{'returnBytes'});
          }

          if( exists($settingMetrics{'type'}) ) {
            my $type = $settingMetrics{'type'};
            if( $type =~ /^int.*/ || $type =~ /^float.*/ ) {
              $value = Xormon::formatNumberByType($type, $value);
            }
          }

        }

        if( defined $value && $value ne '' ) {
          $preparedData{$section}{$name} = $value;
        }
      }
    }
    else {
      $logger->error("Error when executing a command: " . $config{'cmd'} . "\n" . $error);
      return 0;
    }

  }

  if( exists($config{'addItems'}) ) {
    for my $addItem_ref ( @{$config{'addItems'}} ) {
      my %addItem = %{$addItem_ref};
      my $itemName = $addItem{'name'};
      if( exists($addItem{'params'}) ) {
        my %callbackParams;
        for my $item ( @{$addItem{'params'}} ) {
          if( $item =~ /^@(.*)/ ) {
            my $field = $1;
            for my $section ( keys %sections ) {
              if( !exists $callbackParams{$section} ) {
                $callbackParams{$section} = ();
              }
              push(@{$callbackParams{$section}}, $preparedData{$section}{$field});
            }
          }
        }

        for my $section ( keys %sections ) {
          my $callback = $addItem{'callback'};
          no strict 'refs';
          my $result = &{$callback}($callbackParams{$section});
          use strict 'refs';
          if( defined $result && $result ne '' ) {
            # We do not overwrite the value if it already exists.
            # The value is only counted if we don't already have it from the device.
            if( !ref($preparedData{$section}{$itemName}) ) {
              $preparedData{$section}{$itemName} = $result;
            }
          }
        }
      }
    }
  }

  for my $key ( keys(%preparedData) ) {
    my %dataItem = %{$preparedData{$key}};
    if( defined($ID) && !exists($config{'structure'}) ) {
      $output{$dataItem{$idName}} = \%dataItem;
    }
    elsif( exists($config{'structure'}) ) {
      if( exists($config{'uid'}) ) {
        if( !defined($dataItem{'uid'}) ) {
          # if uid exists we do not overwrite
          my $unique = create_unique_key($config{'uid'}, \%dataItem);
          $dataItem{'uid'} = getUID($unique);
        }
      }
      else {
        if( !exists($dataItem{'uid'}) and exists($config{'subsystem'}) and exists($config{'id_field'}) ) {
          $dataItem{'uid'} = $uid_cache{$config{'subsystem'}}{$dataItem{$config{'id_field'}}};
        }
      }

      if( exists($config{'subsystem'}) and exists($config{'id_field'}) and exists($config{'uid_field'}) and exists($dataItem{$config{'id_field'}}) and exists($dataItem{$config{'uid_field'}}) ) {
        $uid_cache{$config{'subsystem'}}{$dataItem{$config{'id_field'}}} = $dataItem{$config{'uid_field'}};
      }
      %data = %{create_hash_structure($config{'structure'}, \%data, \%dataItem)};
    }
    else {
      %output = %dataItem;
    }
  }
  return \%output;
}

#used to add a subsystem to a parent system - for example drive -> pool - needed for example for architecture generation
sub add_subsystem_to_parent {
  my $config_ref = shift;
  my @config = @{$config_ref};
  for my $cfg_ref ( @config ) {
    my %cfg = %{$cfg_ref};
    if( defined($cfg{'description'}) ) {
      $logger->info($cfg{'description'})
    }
    my %subsystem = %{get_ref_from_sructure($cfg{"subsystem_path"})};
    my %parent_subsystem = %{get_ref_from_sructure($cfg{"parent_subsystem_path"})};
    for my $key ( keys(%subsystem) ) {
      my %item = %{$subsystem{$key}};
      my $uid = $item{$cfg{"uid_field"}};
      my $parent_id_field = $item{$cfg{"parent_id_field"}};
      if( defined $parent_id_field ) {
        my @parent_ids = split(", ", $parent_id_field);
        for my $parent_id ( @parent_ids ) {
          if( exists($uid_cache{$cfg{"parent_subsystem_name"}}{$parent_id}) ) {
            my $parent_uid = $uid_cache{$cfg{"parent_subsystem_name"}}{$parent_id};
            if( defined($parent_uid) ) {
              if( exists($parent_subsystem{$parent_uid}) ) {
                $parent_subsystem{$parent_uid}{$cfg{"target_field_name"}}{$uid} = $uid;
                if( exists $cfg{'parent_uids_field'} ) {
                  $subsystem{$uid}{$cfg{'parent_uids_field'}}{$parent_uid} = $parent_uid;
                }
              }
            }
            else {
              $logger->error("missing parent_uid for subsystem: '" . $cfg{"parent_subsystem_name"} . "' and parent_id: '" . $parent_id . "'");
              if( $DEBUG ) {
                my $dump = &VrDump(\%uid_cache, '%uid_cache', 0, 1, 1);
                $logger->info($dump);
              }
            }
          }
        }
      }
    }
  }
  return 1;
}

# function that processes the aggregate section of the config - passes unique items and calls the appropriate callback functions
sub aggregate_uni {
  my $config_ref = shift;
  my @config = @{$config_ref};
  for my $cfg_ref ( @config ) {
    my %cfg = %{$cfg_ref};
    if( defined($cfg{'description'}) ) {
      $logger->info($cfg{'description'})
    }
    my $aggregate_function = $cfg{'aggregate_function'};
    no strict 'refs';
    &{$aggregate_function}(\%cfg);
    use strict 'refs';
  }
  return 1;
}
#function that processes the after_aggregate section of the config - passes unique items and calls the appropriate callback functions
sub after_aggregate_uni {
  my $config_ref = shift;
  my @config = @{$config_ref};
  for my $cfg_ref ( @config ) {
    my %cfg = %{$cfg_ref};
    my $aggregate_function = $cfg{'after_aggregate_function'};
    if( defined($cfg{'description'}) ) {
      $logger->info($cfg{'description'})
    }
    no strict 'refs';
    &{$aggregate_function}(\%cfg);
    use strict 'refs';
  }
  return 1;
}

#function that processes the after_conf section of the config - passes unique items and calls the appropriate callback functions
sub after_conf_uni {
  my $config_ref = shift;
  my @config = @{$config_ref};
  for my $cfg_ref ( @config ) {
    my %cfg = %{$cfg_ref};
    my $aggregate_function = $cfg{'function'};
    if( defined($cfg{'description'}) ) {
      $logger->info($cfg{'description'})
    }
    no strict 'refs';
    &{$aggregate_function}(\%cfg);
    use strict 'refs';
  }
  return 1;
}

# used for calculations and creation of new keys over aggregated data
sub after_aggregate_uid_timestamp {
  my $cfg_ref = shift;
  my %cfg = %{$cfg_ref};
  my @structure_source = split(",", $cfg{'structure_source'});
  my %hash = %data;
  for my $key ( @structure_source ) {
    if( $key =~ /^@(.*)/ ) {
      for my $uid_item ( keys %hash ) {
        for my $time_item ( keys %{$hash{$uid_item}} ) {
          my %record = %{$hash{$uid_item}{$time_item}};
          for my $addItem_ref ( @{$cfg{'addItems'}} ) {
            my %addItem = %{$addItem_ref};
            my @param_value;
            for my $parametr_field ( @{$addItem{'params'}} ) {
              if( $parametr_field =~ /^@(.*)/ ) {
                my $field = $1;
                push(@param_value, $hash{$uid_item}{$time_item}{$field});
              }
              else {
                push(@param_value, $parametr_field);
              }
            }
            my $callback = $addItem{'callback'};
            no strict 'refs';
            my $result = &{$callback}(\@param_value);
            use strict 'refs';
            if( defined $result && $result ne '' ) {
              my %dataItem = ( $addItem{'name'} => $result, 'uid' => $uid_item, 'timestamp' => $time_item );
              %data = %{create_hash_structure($cfg{'structure_desc'}, \%data, \%dataItem)};
            }
          }
        }
      }
      last;
    }
    else {
      if( ref($hash{$key}) eq 'HASH' ) {
        %hash = %{$hash{$key}};
      }
      else {
        %hash = ();
      }
    }
  }
  return 1;
}

#is used to aggregate data for the same item (same uid) if it is placed under multiple nodes.
sub aggregate_uid_timestamp {
  my $cfg_ref = shift;
  my %cfg = %{$cfg_ref};
  my @structure_source = split(",", $cfg{'structure_source'});
  my %hash = %data;
  for my $key ( @structure_source ) {
    if( $key eq '*' ) {
      my %new_data;
      for my $sp_item ( keys %hash ) {
        for my $uid_item ( keys %{$hash{$sp_item}} ) {
          for my $time_item ( keys %{$hash{$sp_item}{$uid_item}} ) {
            if( ref($hash{$sp_item}{$uid_item}{$time_item}) eq 'HASH' ) {
              for my $item ( keys %{$hash{$sp_item}{$uid_item}{$time_item}} ) {
                push(@{$new_data{$uid_item}{$time_item}{$item}}, $hash{$sp_item}{$uid_item}{$time_item}{$item});
              }
            }
          }
        }
      }

      for my $addItem_ref ( @{$cfg{'addItems'}} ) {
        my %addItem = %{$addItem_ref};
        for my $uid_item ( keys %new_data ) {
          for my $time_item ( keys %{$new_data{$uid_item}} ) {
            my @param_value;
            for my $parametr_field ( @{$addItem{'params'}} ) {
              if( $parametr_field =~ /^@(.*)/ ) {
                my $field = $1;
                push(@param_value, $new_data{$uid_item}{$time_item}{$field});
              }
              else {
                push(@param_value, $parametr_field);
              }
            }
            my $callback = $addItem{'callback'};
            no strict 'refs';
            my $result = &{$callback}(@param_value);
            use strict 'refs';
            if( defined $result && $result ne '' ) {
              my %dataItem = ( $addItem{'name'} => $result, 'uid' => $uid_item, 'timestamp' => $time_item );
              %data = %{create_hash_structure($cfg{'structure_desc'}, \%data, \%dataItem)};
            }
          }
        }
      }

      last;
    }
    else {
      if( ref($hash{$key}) eq 'HASH' ) {
        %hash = %{$hash{$key}};
      }
      else {
        %hash = ();
      }
    }
  }
  return 1;
}

#is used to aggregate data for the superordinate subsystem from data of several subordinate subsystems.
sub aggregate_system_from_subsystem {
  my $cfg_ref = shift;
  my %cfg = %{$cfg_ref};
  my %structure_source = %{get_ref_from_sructure($cfg{'structure_source'})};
  my %subsystem_perf_data = %{get_ref_from_sructure($cfg{'subsytem_perf_data'})};
  my @structure = split(",", $cfg{'structure_source'});
  my $subsystem_uids_key = pop(@structure);
  if( $subsystem_uids_key =~ /^@(.*)/ ) {
    $subsystem_uids_key = $1;
  }

  for my $uid ( keys %structure_source ) {
    my $subsystem_uids = $structure_source{$uid}{$subsystem_uids_key};
    my %new_data;
    for my $subsystem_uid ( keys(%{$subsystem_uids}) ) {
      for my $timestamp ( keys(%{$subsystem_perf_data{$subsystem_uid}}) ) {
        my %perf_data = %{$subsystem_perf_data{$subsystem_uid}{$timestamp}};
        for my $item_name ( keys %perf_data ) {
          push(@{$new_data{$timestamp}{$item_name}}, $perf_data{$item_name});
        }
      }
    }
    for my $item_ref ( @{$cfg{'addItems'}} ) {
      my %item = %{$item_ref};
      for my $timestamp ( keys(%new_data) ) {
        my @param_value;
        for my $parametr_field ( @{$item{'params'}} ) {
          if( $parametr_field =~ /^@(.*)/ ) {
            my $field = $1;
            push(@param_value, $new_data{$timestamp}{$field});
          }
          else {
            push(@param_value, $parametr_field);
          }
        }
        if( @param_value ) {
          my $callback = $item{'callback'};
          no strict 'refs';
          my $result = &{$callback}(@param_value);
          use strict 'refs';
          if( defined $result && $result ne '' ) {
            my %dataItem = ( $item{'name'} => $result, 'uid' => $uid, 'timestamp' => $timestamp );
            %data = %{create_hash_structure($cfg{'structure_desc'}, \%data, \%dataItem)};
          }
        }
      }
    }
  }
  return 1;
}

#is used to aggregate data for the superordinate subsystem from data of several subordinate subsystems.
sub aggregate_devices_from_pool_fs {
  my $cfg_ref = shift;
  my %cfg = %{$cfg_ref};
  my %structure_pools_source = %{get_ref_from_sructure($cfg{'structure_pools_source'})};
  my %structure_fs_source = %{get_ref_from_sructure($cfg{'structure_fs_source'})};
  my %pools_perf_data = %{get_ref_from_sructure($cfg{'pools_perf_data'})};
  my %fs_perf_data = %{get_ref_from_sructure($cfg{'fs_perf_data'})};

  my @pools_structure = split(",", $cfg{'structure_pools_source'});
  my $pools_uids_key = pop(@pools_structure);
  if( $pools_uids_key =~ /^@(.*)/ ) {
    $pools_uids_key = $1;
  }

  my @fs_structure = split(",", $cfg{'structure_fs_source'});
  my $fs_uids_key = pop(@fs_structure);
  if( $fs_uids_key =~ /^@(.*)/ ) {
    $fs_uids_key = $1;
  }

  my %structure_source = ( %structure_pools_source, %structure_fs_source );
  for my $uid ( keys %structure_source ) {
    my %new_data;

    my $pools_uids = $structure_pools_source{$uid}{$pools_uids_key};
    for my $pool_uid ( keys(%{$pools_uids}) ) {
      for my $timestamp ( keys(%{$pools_perf_data{$pool_uid}}) ) {
        my %perf_data = %{$pools_perf_data{$pool_uid}{$timestamp}};
        for my $item_name ( keys %perf_data ) {
          push(@{$new_data{$timestamp}{$item_name}}, $perf_data{$item_name});
        }
      }
    }

    my $fs_uids = $structure_fs_source{$uid}{$fs_uids_key};
    for my $fs_uid ( keys(%{$fs_uids}) ) {
      for my $timestamp ( keys(%{$fs_perf_data{$fs_uid}}) ) {
        my %perf_data = %{$fs_perf_data{$fs_uid}{$timestamp}};
        for my $item_name ( keys %perf_data ) {
          push(@{$new_data{$timestamp}{$item_name}}, $perf_data{$item_name});
        }
      }
    }

    for my $item_ref ( @{$cfg{'addItems'}} ) {
      my %item = %{$item_ref};
      for my $timestamp ( keys(%new_data) ) {
        my @param_value;
        for my $parametr_field ( @{$item{'params'}} ) {
          if( $parametr_field =~ /^@(.*)/ ) {
            my $field = $1;
            push(@param_value, $new_data{$timestamp}{$field});
          }
          else {
            push(@param_value, $parametr_field);
          }
        }
        if( @param_value ) {
          my $callback = $item{'callback'};
          no strict 'refs';
          my $result = &{$callback}(@param_value);
          use strict 'refs';
          if( defined $result && $result ne '' ) {
            my %dataItem = ( $item{'name'} => $result, 'uid' => $uid, 'timestamp' => $timestamp );
            %data = %{create_hash_structure($cfg{'structure_desc'}, \%data, \%dataItem)};
          }
        }
      }
    }
  }
  return 1;
}


#is used to copy data. It is used to transfer configuration data to the performance
sub aggregate_copy_data {
  my $cfg_ref = shift;
  my %cfg = %{$cfg_ref};
  my %source = %{get_ref_from_sructure($cfg{'structure_source'})};
  my %desc = %{get_ref_from_sructure($cfg{'structure_desc'})};
  for my $uid ( keys %source ) {
    my %source_item = %{$source{$uid}};
    for my $cfg_item_ref ( @{$cfg{'items'}} ) {
      my %cfg_item = %{$cfg_item_ref};
      my $value = $source_item{$cfg_item{'from'}};
      for my $timestamp ( keys %{$desc{$uid}} ) {
        $desc{$uid}{$timestamp}{$cfg_item{'to'}} = $value;
      }
    }
  }
  return 1;
}

# adds the event log events as an array, in the conf data of the device
sub add_events_to_device {
  my $cfg_ref = shift;
  my %cfg = %{$cfg_ref};
  my %source = %{get_ref_from_sructure($cfg{'structure_source'})};
  my %desc = %{get_ref_from_sructure($cfg{'structure_desc'})};
  my @events;
  for my $uid ( keys %source ) {
    my %source_item = %{$source{$uid}};
    if( $source_item{'severity'} ne 'info' and $source_item{'severity'} ne 'notice' ) {
      push(@events, \%source_item);
    }
  }
  my $events_json = JSON::to_json(\@events);
  for my $uid ( keys(%desc) ) {
    $desc{$uid}{$cfg{'target_events_field'}} = $events_json;
  }
  return 1;
}

# generate configuration data for naspool
sub create_conf_naspool {
  my $cfg_ref = shift;
  my %cfg = %{$cfg_ref};
  my %subsystems = %{get_ref_from_sructure($cfg{'nas_conf_path'})};
  my %naspool_data;
  for my $nas_uid ( keys(%subsystems) ) {
    my $pool_uid;
    my $pool_name;
    if( exists($subsystems{$nas_uid}{$cfg{'pool_uid_field'}}) ) {
      $pool_uid = $subsystems{$nas_uid}{$cfg{'pool_uid_field'}};
    }
    if( defined($pool_uid) and exists($subsystems{$nas_uid}{$cfg{'pool_name_hash'}}{$pool_uid}) ) {
      $pool_name = $subsystems{$nas_uid}{$cfg{'pool_name_hash'}}{$pool_uid};
    }
    my $naspool_uid = create_unique_key($cfg{'uid'}, $subsystems{$nas_uid});
    $naspool_uid = getUID($naspool_uid);
    $naspool_data{$naspool_uid}{"uid"} = $naspool_uid;
    $naspool_data{$naspool_uid}{"pool"} = $pool_uid;
    if( defined($pool_name) ) {
      $naspool_data{$naspool_uid}{$cfg{'name_field_name'}} = $cfg{'pool_name_prefix'} . $subsystems{$nas_uid}{$cfg{'name_field_name'}} . " - " . $pool_name;
    }
    $naspool_data{$naspool_uid}{"nas"}{$nas_uid} = $nas_uid;
  }
  for my $data_item_ref ( values(%naspool_data) ) {
    my %data_item = %{$data_item_ref};
    if( exists($data_item{$cfg{'name_field_name'}}) and exists($data_item{'uid'}) ) {
      $uid_cache{$cfg{'subsystem'}}{$data_item{$cfg{'name_field_name'}}} = $data_item{'uid'};
    }
    %data = %{create_hash_structure($cfg{'target_conf_path'}, \%data, \%data_item)};
  }
  return 1;
}

# creates configuration data, adds pools to device
sub create_conf_device_pools_fs {
  my $cfg_ref = shift;
  my %cfg = %{$cfg_ref};
  my %pools = %{get_ref_from_sructure($cfg{'pools_conf_path'})};
  my %fs = %{get_ref_from_sructure($cfg{'fs_conf_path'})};
  my %devices = %{get_ref_from_sructure($cfg{'target_conf_path'})};
  my %pools_uid;
  my %fss_uid;
  for my $pool_uid ( keys(%pools) ) {
    $pools_uid{$pool_uid} = $pool_uid;
  }

  for my $fs_uid ( keys(%fs) ) {
    $fss_uid{$fs_uid} = $fs_uid;
  }
  for my $device_uid ( keys(%devices) ) {
    $devices{$device_uid}{$cfg{'target_pools_field_name'}} = \%pools_uid;
    $devices{$device_uid}{$cfg{'target_fs_field_name'}} = \%fss_uid;
  }

  return 1;
}

# creates a node_name field (node names separated by a comma and a space)
sub create_device_node_name {
  my $cfg_ref = shift;
  my %cfg = %{$cfg_ref};
  my %nodes = %{get_ref_from_sructure($cfg{'nodes_conf_path'})};
  my %devices = %{get_ref_from_sructure($cfg{'target_conf_path'})};
  my $node_name;
  for my $node_uid ( keys(%nodes) ) {
    if( defined($node_name) ) {
      $node_name .= ", ";
    }
    $node_name .= $nodes{$node_uid}{$cfg{'node_name_field'}};
  }

  for my $device_uid ( keys(%devices) ) {
    $devices{$device_uid}{$cfg{'target_field_name'}} = $node_name;
  }
  return 1;
}

# copies part of the structure in %data to a new location in %data
# {
#   "function"        : "copy_new_structure",
#                       "source_structure": "CONF,device_tmp",
#                       "target_structure": "CONF,device",
#                       "uid_field"       : "uid"
# }
sub copy_new_structure {
  my $cfg_ref = shift;
  my %cfg = %{$cfg_ref};
  my %source = %{get_ref_from_sructure($cfg{'source_structure'})};
  my $uid = $source{$cfg{'uid_field'}};
  my %data_item;
  my %clone = %source;
  $data_item{$uid} = \%clone;
  %data = %{create_hash_structure($cfg{'target_structure'}, \%data, \%data_item)};
  return 1;
}

#generates architectures
sub generate_architecture {
  my $config_ref = shift;
  my @config = @{$config_ref};
  my $hw_type;
  $logger->info("Generating architectures");
  for my $i ( 0 .. $#config ) {
    my %arch_def = %{$config[$i]};
    if( exists($arch_def{'hw_type'}) ) {
      $hw_type = $arch_def{'hw_type'}
    }

    if( $arch_def{"subsystem"} eq "device" ) {
      my %arch = (
        "hw_type"    => $hw_type,
        "subsystem"  => $arch_def{'subsystem'},
        "hostcfg_id" => $cfg->{'hostcfg_id'},
        "item_id"    => $cfg->{'hostcfg_id'},
        "label"      => $cfg->{'hostalias'},
        "parents"    => [],
      );
      push(@architecture, \%arch);
    }
    else {
      my %subsystems = %{get_ref_from_sructure($arch_def{'structure'})};
      for my $key ( keys %subsystems ) {
        my %arch = (
          "hw_type"    => $hw_type,
          "subsystem"  => $arch_def{'subsystem'},
          "hostcfg_id" => $cfg->{'hostcfg_id'},
        );
        $arch{"item_id"} = $key;
        if( exists($subsystems{$key}{$arch_def{'label_field'}}) ) {
          $arch{"label"} = $subsystems{$key}{$arch_def{'label_field'}};
        }
        else {
          $arch{"label"} = $key;
        }
        if( exists $arch_def{'wwn_field'} and exists($subsystems{$key}{$arch_def{'wwn_field'}}) ) {
          $arch{"wwn"} = $subsystems{$key}{$arch_def{'wwn_field'}};
        }
        for my $paren_name ( @{$arch_def{'parents'}} ) {
          if( $paren_name eq "device" ) {
            push(@{$arch{"parents"}}, $cfg->{'hostcfg_id'})
          }
          else {
            if( $paren_name =~ /^@(.*)/ ) {
              $paren_name = $1;
            }
            if( exists($subsystems{$key}{$paren_name}) ) {
              if( ref($subsystems{$key}{$paren_name}) eq 'HASH' ) {
                push(@{$arch{"parents"}}, keys %{$subsystems{$key}{$paren_name}})
              }
              else {
                push(@{$arch{"parents"}}, $subsystems{$key}{$paren_name});
              }
            }
          }
        }
        if( exists($arch_def{'children'}) ) {
          for my $children_path ( @{$arch_def{'children'}} ) {
            my %children_hash = %{get_ref_from_sructure($children_path, $subsystems{$key})};
            for my $child_uid ( keys %children_hash ) {
              push(@{$arch{"children"}}, $child_uid);
            }
          }
        }
        push(@architecture, \%arch);
      }
    }
  }
  return \@architecture;
}
# deletes unnecessary data from %data
sub clean_up {
  my $config_ref = shift;
  my @config = @{$config_ref};
  my %global_action;
  $logger->info("Clean up");

  for my $cfg_ref ( @config ) {
    my %cfg = %{$cfg_ref};
    if( $cfg{'structure'} eq "global" ) {
      for my $remove_field ( @{$cfg{'remove'}} ) {
        $global_action{$remove_field} = 'remove';
      }
    }
    else {
      my %action = %{\%global_action};
      for my $remove_field ( @{$cfg{'remove'}} ) {
        $action{$remove_field} = 'remove';
      }
      my %hash = %{get_ref_from_sructure($cfg{'structure'})};
      clean_up_item(\%hash, \%cfg, \%action);
    }
  }
  return 1;
}
# deletes one item
sub clean_up_item {
  my $item = shift;
  my $cfg_ref = shift;
  my $action_ref = shift;
  my %cfg = %{$cfg_ref};
  my %action = %{$action_ref};
  if( ref($item) eq 'HASH' ) {
    for my $key ( keys %{$item} ) {
      if( exists($action{$key}) and $action{$key} eq "remove" ) {
        delete $item->{$key};
      }
      if( ref($item->{$key}) eq 'HASH' ) {
        clean_up_item(\%{$item->{$key}}, \%cfg, \%action);
      }
    }
  }
  return 1;
}

# perform the final transfer of the structure
sub final_structure_change {
  my $config_ref = shift;
  my %cfg = %{$config_ref};
  $logger->info("final structure change");
  for my $structure ( @{$cfg{'source_stucture'}} ) {
    my %data_item = %{get_ref_from_sructure($structure)};
    %data = %{create_hash_structure($cfg{'target_structure'}, \%data, \%data_item)};
  }
  return 1;
}

# creates a new structure according to the specified path
sub create_hash_structure {
  my $structureDefinition = shift;
  my $output_ref = shift;
  my $items_ref = shift;
  my %output = %{$output_ref};
  my %items = %{$items_ref};
  my @structureKeys = split(",", $structureDefinition);

  my $key;
  my $ref = \%output;
  for my $i ( 0 .. $#structureKeys ) {
    my $col = $structureKeys[$i];
    if( $col =~ /^@(.*)$/ ) {
      my $colName = $1;
      $key = $items{$colName};
    }
    else {
      $key = $col;
    }

    if( defined($key) ) {
      if( $i == $#structureKeys ) {
        $ref->{$key} = { %{$ref->{$key} // {}}, %items };
      }
      else {
        $ref->{$key} = {} unless( exists $ref->{$key} );
      }

      if( exists($ref->{$key}) ) {
        $ref = $ref->{$key};
      }
      else {
        $ref = {};
      }
    }
  }
  return \%output;
}

# returns a reference to the structure according to the specified path
sub get_ref_from_sructure {
  my $structureDefinition = shift;
  my $optional_hash_ref = shift;
  my @structureKeys = split(",", $structureDefinition);
  my $key;
  my %ref;
  if( ref $optional_hash_ref eq 'HASH' ) {
    %ref = %{$optional_hash_ref};
  }
  else {
    %ref = %data;
  }

  for my $i ( 0 .. $#structureKeys ) {
    my $col = $structureKeys[$i];
    if( $col =~ /^@(.*)$/ or $col eq '*' ) {
      last
    }
    else {
      $key = $col;
    }
    if( exists($ref{$key}) ) {
      %ref = %{$ref{$key}};
    }
    else {
      %ref = ();
    }
  }
  return \%ref;
}

# Creates a unique key according to the configuration (readable form, encrypted separately)
sub create_unique_key {
  my $keyDefinition = shift;
  my $items_ref = shift;
  my %items = %{$items_ref};
  my @uniqueKeys = split(",", $keyDefinition);

  my $unique = '';
  for my $i ( 0 .. $#uniqueKeys ) {
    my $col = $uniqueKeys[$i];
    if( $col =~ /^@(.*)$/ ) {
      my $key = $1;
      if( exists($items{$key}) ) {
        $unique .= $items{$key};
      }
    }
    else {
      $unique .= $col;
    }
  }
  return $unique;
}

# filters the @availableMetrics array, metrics according to the pattern and returns them as a string separated by a comma and a space
sub preparePath {
  my $requiredMetrics = shift;
  my @metricsFound = grep(/$requiredMetrics/, @availableMetrics);
  return join(", ", @metricsFound);
}

#creates structure for healt status
sub health_status {
  my $cfg_ref = shift;
  my %cfg = %{$cfg_ref};
  my @health_status;
  my %source = %{get_ref_from_sructure($cfg{"structure_source"})};

  for my $uid ( keys(%source) ) {
    my $status = $source{$uid}{$cfg{"status_field_name"}};
    my $detail = $source{$uid}{$cfg{"status_detail_field_name"}};

    if( defined $status ) {
      no strict 'refs';
      $status = &{$cfg{"status_normalize_callback"}}($status);
      use strict 'refs';
      my %data_item;
      $data_item{"item_id"} = $uid;
      # $data_item{"type"}    = "health";
      $data_item{"status"} = $status;
      if( defined($detail) ) {
        no strict 'refs';
        $detail = &{$cfg{"status_detail_callback"}}($detail);
        use strict 'refs';
        $data_item{"data"} = $detail;
      }
      else {
        $data_item{"data"} = '';
      }

      push(@health_status, \%data_item);
    }
  }
  $data{$cfg{"structure_target"}} = \@health_status;
  return 1;
}


#### callback function

# for unity is the size in bytes format (conversion to higher units)
# the function removes the text in brackets and leaves only the value in bytes
sub normalizeSize {
  my $text = shift;
  $text =~ s/^(.*)(\s.*)$/$1/;
  return $text;
}

# removes the % sign and leaves only the numeric value
sub normalizePercent {
  my $text = shift;
  $text =~ s/%//g;
  return $text;
}

#converts microseconds to milliseconds
sub conver_usec2msec {
  my $usec = shift;
  my $msec = $usec / 1000;
  return $msec;
}

# adds a pool to the NAS
sub addPool2Nas {
  my $poolName = shift;
  my $ID = shift;
  my $pool_uid = getUID("pool$poolName");
  my $nas_uid = getUID("nas$ID");

  my $pool_name = $data{'CONF'}{'pool'}{$pool_uid}{'name'};
  # I can't control the existence of the NAS because it doesn't exist yet.
  # However, I am close to the definition of NAS so it must come into existence.
  $data{'CONF'}{'nas'}{$nas_uid}{'pools'}{$pool_uid} = $pool_name;

  return ( $pool_uid );
}
# adds drive to pool
sub addDrive2Pool {
  my $poolName = shift;
  my $ID = shift;
  my $pool_uid = getUID("pool$poolName");
  my $drive_uid = getUID("drive$ID");

  # I can't control the existence of the NAS because it doesn't exist yet.
  # However, I am close to the definition of NAS so it must come into existence.
  $data{'CONF'}{'pool'}{$pool_uid}{'drives'}{$drive_uid} = $drive_uid;

  return ( $pool_uid );
}

# convert date fe format YYYY-mm-dd HH:mm:ss to timestamp
sub date2timestamp {
  my $datetime = shift;

  my ( $year, $month, $day, $hour, $min, $sec ) = $datetime =~ /(\d+)-(\d+)-(\d+) (\d+):(\d+):(\d+)/;
  # Perl's Time::Local module expects months in the range 0-11, so we subtract 1 from the month
  $month -= 1;
  my $timestamp = timegm($sec, $min, $hour, $day, $month, $year);
  return ( $timestamp )
}

# converts text to lower case
sub lowerCase {
  my $value = shift;
  return lc($value);
}

# multiplies all numbers passed in the parameter array by each other
sub number_multiply {
  my $array_ref = shift;
  my $result;
  for my $value ( values @{$array_ref} ) {
    if( !defined($result) ) {
      $result = $value;
    }
    else {
      $result *= $value;
    }
  }
  return $result;
}

# divide the first number passed in the field by the second and then divides the result by the next number
sub number_divide {
  my $array_ref = shift;
  my $result;
  if( ref($array_ref) eq 'ARRAY' ) {
    if( @{$array_ref} > 1 ) {
      for my $value ( values @{$array_ref} ) {
        if( Xormon::isdigit($value) ) {
          if( !defined($result) ) {
            $result = $value;
          }
          else {
            if( $value != 0 ) {
              $result /= $value;
            }
            else {
              $result = '';
              last;
            }
          }
        }
        else {
          $result = '';
          last;
        }
      }
    }
    else {
      $result = '';
    }
  }
  return $result;
}

# multiplies all the numbers passed to the function in the array
sub number_add {
  my $array_ref = shift;
  my $result;
  if( ref($array_ref) eq 'ARRAY' ) {
    for my $value ( values @{$array_ref} ) {
      if( defined($value) ) {
        if( !defined($result) ) {
          $result = $value;
        }
        else {
          $result += $value;
        }
      }
    }
  }
  return $result;
}

# sequentially subtracts all values passed to the function in the array from each other
sub number_subtract {
  my $array_ref = shift;
  my $result;
  for my $value ( values @{$array_ref} ) {
    if( defined($value) ) {
      if( !defined($result) ) {
        $result = $value;
      }
      else {
        $result -= $value;
      }
    }
  }
  return $result;
}

# successively subtracts all values passed to the function in the array from each other. if the result is a negative number returns zero
sub number_subtract_negative_zero {
  my $array_ref = shift;
  my $result;
  $result = number_subtract($array_ref);
  if( $result <= 0 ) {
    $result = 0;
  }
  return $result;
}

# calculates the bound diameter, the function accepts only arrays of two elements
sub weighted_avg_single_params {
  my $param_ref = shift;
  my @param = @{$param_ref};
  return weighted_avg($param[0], $param[1]);
}

#calculates the weighted average, the function accepts two arrays, an array of values and an array of the number of elements
sub weighted_avg {
  my $value_ref = shift;
  my $count_ref = shift;
  if( ref($value_ref) eq 'ARRAY' and ref($count_ref) eq 'ARRAY' ) {
    my @values = @{$value_ref};
    my @counts = @{$count_ref};
    my @data_for_calculation;
    for my $i ( 0 .. $#values ) {
      my %item = ( 'value' => $values[$i], 'count' => $counts[$i] );
      push(@data_for_calculation, \%item);
    }
    my $weighted_avg = Xormon::calculate_weighted_avg(\@data_for_calculation, 0);
    if( !defined($weighted_avg) ) {
      $weighted_avg = 0;
    }
    return $weighted_avg;
  }
  else {
    return undef;
  }
}

# calculates the capacity of subscription overrun in bytes
sub calculate_capacity_oversubscr {
  my $array_ref = shift;
  my @array = @{$array_ref};
  my $result = 0;
  if( defined($array[0]) and defined($array[1]) ) {
    my $capacity_total = $array[0];
    my $oversubscr_percent = $array[1];
    $result = $capacity_total / 100 * $oversubscr_percent;
  }
  return $result;
}
# calculates the capacity of subscription overrun in percentage
sub calculate_capacity_oversubscr_percent {
  my $array_ref = shift;
  my @array = @{$array_ref};
  my $result = 0;
  if( defined($array[0]) and defined($array[1]) ) {
    my $capacity_total = $array[0];
    my $capacity_oversubscr = $array[1];
    $result = ( $capacity_oversubscr / $capacity_total ) * 100;
  }
  return $result;
}

# calculates the used capacity in percentage
sub calculate_used_percent {
  my $array_ref = shift;
  my @array = @{$array_ref};
  my $result = 0;
  if( defined($array[0]) and defined($array[1]) ) {
    my $capacity_total = $array[0];
    my $capacity_used = $array[1];
    $result = ( $capacity_used / $capacity_total ) * 100;
  }
  return $result;
}

# always returns hostcfg_id, ignores passed parameters
sub get_hostcfg_id {
  my @unuset_params = @_;
  return $cfg->{'hostcfg_id'};
}

# standardizes health_status: OK -> ok, Warning -> warning, everything else to error
sub health_status_normalize {
  my $value = shift;
  my $state;
  if( defined($value) ) {
    if( $value =~ /^OK.*/ ) {
      $state = "ok";
    }
    elsif( $value =~ /Warning/ ) {
      $state = "warning"
    }
    else {
      $state = "error";
    }
  }
  return $state;
}

# remove quotation marks from the text - only at the beginning and end of the text
sub remove_quotation_marks {
  my $value = shift;
  $value =~ s/^"(.*)"$/$1/; #removed quotation marks
  return $value;
}

#format WWN, returns the last 16 characters. Unity use WWN 32 character
sub formatWWN{
  my $value = shift;
  $value = Xormon::formatWWN($value);
  $value = substr($value, -16);
  return $value;
}


# tools

# encrypts the unique call to Xormon::create_hash_key and returns the unique uuid
sub getUID {
  my $uniqueKey = shift;
  my $uid = lc(Xormon::create_hash_key($cfg->{'hostcfg_id'} . $uniqueKey));
  return $uid;
}

# start display with passed parameters
sub runCmd {
  my $params = shift;
  my $cmd = $cfg->{'cmd'} . ' ' . $params;
  my @out = `$cmd`;
  return @out;
}

# the connection test performs a basic query using uemcli to determine if there is a functional connection to the device
sub test_connection {
  my $test = shift;

  $cmd->addParam("/sys/general show -filter 'System name,Model,Product serial number'", '', 0);
  my ( $string, $error ) = $cmd->run();
  if( !defined($error) ) {
    $string =~ s/^\d+:\s+//;   # remove '1:'
    $string =~ s/^\s+//gm;     # remove spaces at the beginning of a line
    $string =~ s/\s+=\s+/: /g; # change format
    $test->addSection("Connection test", 1, $string);
    return 1;
  }
  else {
    $test->addSection("Connection test", 0, $error);
    return 0;
  }
}

# the connection test performs a basic query using uemcli to determine if there is a functional connection to the device
sub test_ip_connection {
  my $test = shift;
  my $config = shift;
  my $port = 443; # the port does not change

  my ( $code, $text ) = Xormon::conntest_tcp($config->{'host'}, $port);
  $test->addSection("Device network connection", $code, $text);
  if( $code != 0 ) {
    return 1;
  }
  else {
    return 0;
  }
}

# test whether unity is configured
sub test_configuration {
  my $test = shift;
  my $config = shift;
  # test configuration values
  my @variables = ( "hostcfg_id", "host", "username", "password" );
  foreach my $variable ( @variables ) {
    unless( exists $config->{$variable} ) {
      $test->addSection("Configuration check", 0, "Cannot continue, $variable is not configured!");
      return 0;
    }
  }
  $test->addSection("Configuration check", 1, "");
  return 1;
}

# tests whether the uemcli is installed
sub test_uemcli_instaled {
  my $test = shift;
  my $config = shift;
  # uemcli application installation test
  my $cmd_test_application_installation = $config->{'cmd'} . " -version";
  my $version = `$cmd_test_application_installation`;
  if( defined($version) ) {
    $test->addSection("Installed the uemcli application", 1, $version);
    return 1;
  }
  else {
    $test->addSection("Installed the uemcli application",
                      0,
                      "uemcli cmd failed, make sure you have installed Dell EMC Navisphere product");
    return 0;
  }
}

# private debug function, saves the hash as a json file in the tmp/ directory
sub save_hash_to_json {
  my $hash_ref = shift;
  my $file_sufix = shift;

  if( !defined($file_sufix) ) {
    $file_sufix = '';
  }

  my $t = localtime($Unity::start_time);
  my $filename = "../../logs/unity/" . $cfg->{'hostcfg_id'} . "/data_" . $t->strftime('%Y%m%d_%H-%M-%S') . $file_sufix . '.json';

  if( !-e $filename ) {
    Xormon::save_as_json($filename, $hash_ref);
  }
  return 1;
}

1;
