=head1 NAME
unity-service - a microservice to collect data from EMC Unity and store it in xormon
=cut
use LWP;
use JSON;
use strict;
use warnings;
use Data::Dumper;
use FindBin;
use lib "$FindBin::Bin/"; # use lib in this folder
use lib "../../../perl-modules";
use Xormon qw(VrDump);
use Test;

# for this microservice
use constant 'SERVICE_NAME' => "unity";
use constant 'DEFAULT_UEMCLI_PATH' => "uemcli";
use Unity;

$SIG{'CHLD'} = 'IGNORE';

use IO::Socket qw(AF_INET AF_UNIX SOCK_STREAM SHUT_WR);

# microservice port
my $port = 3232;

my $server = IO::Socket->new(
  'Domain'    => AF_INET,
  'Type'      => SOCK_STREAM,
  'Proto'     => 'tcp',
  'LocalHost' => '0.0.0.0',
  'LocalPort' => $port,
  'ReusePort' => 1,
  'Listen'    => 5,
)            || die "Can't open socket: $@";
Xormon::log("Service listening on port $port");

my $client;

# waiting for a new client connection
while( $client = $server->accept() ) {

  my $pid;
  unless( defined( $pid = fork() ) ) {
    Xormon::error("Wait for fork");
    sleep 5;
  }

  if( $pid == 0 ) {

    # get information about a newly connected client
    my $client_address = $client->peerhost();
    my $client_port = $client->peerport();
    Xormon::log("Connection from $client_address:$client_port");

    # received data template
    #
    # {
    #   "jwt"  : "<backendToken>",
    #   "task" : "<taskName>"
    #   "data" : {
    #     <configuration_from_hostcfg>
    #   }
    # }

    # read up to 2048 characters from the connected client
    my $data = "";
    $client->recv($data, 2048);

    my $decoded = from_json($data);

    Xormon::log("Received task: " . $decoded->{'task'});
    my $response = '';
    my $cfg = $decoded->{'data'};
    if( defined $cfg->{'uemcli'} && -f "$cfg->{'uemcli'}" ) {
      $cfg->{'cmd'} = $cfg->{'uemcli'};
    }
    else {
      $cfg->{'cmd'} = DEFAULT_UEMCLI_PATH;
    }

    my $xormon = Xormon->new("$client_address", $decoded->{jwt}, SERVICE_NAME, $cfg->{hostcfg_id}, $cfg->{hostalias});

    if( $decoded->{'task'} eq "fetch" ) {

      $client->send("Task successfuly started");
      $client->shutdown(SHUT_WR);

      $xormon->startJob($decoded->{'task'}, $cfg->{'hostalias'}, $$);

      ############################### AGENT SECTION ###############################
      my $status;
      my $test_uemcli_exist = `which $cfg->{'cmd'}`;
      if( $test_uemcli_exist ne '' ) {
        Unity::init($cfg);
        my $rc;
        my $data_ref = Unity::collect($xormon, $cfg, SERVICE_NAME);
        my %service_data_in;
        my $service_data_in_ref = $xormon->getData();
        if( ref($service_data_in_ref) eq 'HASH' ) {
          %service_data_in = %{$xormon->getData()};
        }
        my $no_data = 0;
        my $last_perf_timestamp = time();
        my %data = %{$data_ref};
        if( keys(%data) > 0 ) {
          $rc = $xormon->isConnected(1);
          $rc = $xormon->saveJSON2($data{'PERF'});
          if( $rc eq "201" ) {
            $status = "successful";
            $no_data = 0;
          }
          elsif( $rc eq "401" ) {
            $status = "unauthorized";
            $no_data = 1;
          }
          else {
            $status = "error";
            $no_data = 1;
          }
          $rc = $xormon->saveArchitecture($data{'architecture'});
          if( $rc != 201 ) {
            $no_data = 1;
          }
          $rc = $xormon->saveConfJSON($data{'configuration'});
          if( $rc != 201 ) {
            $no_data = 1;
          }
          $rc = $xormon->saveStatus($data{'health_status'});
        }
        else {
          $xormon->isConnected(0, "Connection failed!");
          $no_data = 1;
        }
        if( $no_data == 0 ) {
          my %service_data_out;
          $service_data_out{ 'last_perf_timestamp' } = $last_perf_timestamp;
          my $save_data = $xormon->saveData(\%service_data_out);
          $xormon->hasData(1);
        }
        else {
          if( !defined($service_data_in{'last_perf_timestamp'}) ) {
            $xormon->hasData(0, "no available stats");
          }
          elsif( ( $last_perf_timestamp - $service_data_in{'last_perf_timestamp'} ) > 60 * 60 * 3 ) {
            $xormon->hasData(0,
                             "last available stats are older than 3 hours: " . $service_data_in{last_perf_timestamp});
          }
        }
      }
      else {
        $status = 'error';
        Xormon::error("uemcli cmd failed, make sure you have installed Dell EMC Navisphere product");
      }
      ############################### END AGENT SECTION ###############################


      $xormon->endJob($decoded->{'task'}, $cfg->{'hostalias'}, $status, $$);

      Xormon::log("Task [fetch] for " . $cfg->{'hostalias'} . " is done");

    }
    elsif( $decoded->{'task'} eq "test" ) {

      Unity::init($cfg);
      my $test = Test->new($client);
      my $rc;

      $rc = Unity::test_configuration($test, $cfg);
      if( $rc == 0 ) {
        $test->stop();
      }

      $rc = Unity::test_uemcli_instaled($test, $cfg);
      if( $rc == 0 ) {
        $test->stop();
      }

      $rc = Unity::test_ip_connection($test, $cfg);
      if( $rc == 0 ) {
        $test->stop();
      }

      $rc = Unity::test_connection($test, $cfg);
      if( $rc == 0 ) {
        $test->stop();
      }

      $test->stop();

    }
    else {
      $response = "undefined task";
      $client->send($response);
      $client->shutdown(SHUT_WR);

    }

    exit();

  }
  else {
    $client->autoflush(1);
  }
}

$server->close();
1;
