package XIV;

use strict;
use warnings;

use FindBin;
use lib "$FindBin::Bin/"; # use lib in this folder
use lib "../../../perl-modules";

use Xormon;
use Data::Dumper;
use JSON qw(encode_json decode_json);
use Date::Parse; # str2time
use POSIX qw(mktime);
use XormonLogger;

my $xormon;
my $cfg;
my $wrkdir;
my $sample_rate = 300; # default
my %capacity;
my %hosts;
my $logger;

sub process_collected_data {
  $xormon = shift;
  $cfg    = shift;
  $wrkdir = shift;

  my $status = "successful"; # status code

  #
  # test main variables
  #
  unless ( exists $cfg->{'hostcfg_id'} ) { Xormon::error("Cannot continue, 'hostcfg_id' is not configured!") && return "error"; }
  unless ( exists $cfg->{'hostalias'} )  { Xormon::error("Cannot continue, 'hostalias' is not configured!")  && return "error"; }
  $logger = XormonLogger->new( 'xiv', $cfg->{hostcfg_id}, $cfg->{hostalias} );

  #if ( exists $cfg->{'sampleRate'} && Xormon::isdigit($cfg->{'sampleRate'}) ) { $sample_rate = $cfg->{'sampleRate'} * 60; } # from minutes to seconds

  opendir( my $dir, "$wrkdir" ) || $logger->error( "Cannot open directory '$wrkdir' " . __FILE__ . ":" . __LINE__ ) && return "error";
  my @files = grep !/^\.\.?$/, readdir($dir);
  closedir($dir);
  chomp @files;

  my @conf_files = grep( /^$cfg->{'hostalias'}\_xivconf\_20.*out\.tmp$/, @files );
  my @perf_files = grep( /^$cfg->{'hostalias'}\_xivperf\_20.*out\.tmp$/, @files );

  #
  # Hostcfg status
  #
  if ( scalar(@perf_files) > 0 ) {
    $xormon->isConnected(1);
    $xormon->hasData(1);
  }
  else {
    $xormon->isConnected(0, "perf file has not been created");
    $xormon->hasData(0, "no available stats");
  }

  # useless for now (conf must be parsed every MS run)
  #
  # get microservice data
  #
  #my $service_data_in  = $xormon->getData();
  #my $service_data_out = {};
  #
  # save microservice data
  #
  #$xormon->saveData(\%{ $service_data_out });

  #
  # conf files, send architecture and conf data (every MS run, bcs there are capacity stats which are used in save perf)
  #
  if ( scalar(@conf_files) > 0 ) {
    chomp @conf_files;

    my $idx = 0;
    foreach my $file ( sort { lc $b cmp lc $a } @conf_files ) {
      $idx++;
      if ( $idx == 1 ) { # use only the last conf file
        $logger->info("parse conf file: $wrkdir/$file");
        my $data = parse_conf("$wrkdir/$file");
        send_conf($data);
      }
      remove_file("$wrkdir/$file");
    }
  }

  #
  # perf files
  #
  if ( scalar(@perf_files) > 0 ) {
    chomp @perf_files;

    foreach my $file ( sort { lc $a cmp lc $b } @perf_files ) {
      $logger->info("parse perf file: $wrkdir/$file");
      my $data     = parse_perf("$wrkdir/$file");
      my $perf_out = prepare_perf($data);
      my $rc       = $xormon->saveJSON2($perf_out);
      $logger->info("save perf: $rc");

      if ( $rc eq "201" ) {
        $status = "successful";
      }
      elsif ( $rc eq "401" ) {
        $status = "unauthorized";
        last;
      }
      else {
        $status = "error";
        last;
      }

      remove_file("$wrkdir/$file");
    }

    # stop if data saving fails
    if ( $status ne 'successful' ) { return $status; }
  }

  #
  # health status
  #
  # XIV doesn't have HS in stor

  return $status;
}

sub prepare_perf {
  my $data_in = shift;

  my %data_out;

  foreach my $timestamp ( sort { $a <=> $b } keys %{ $data_in } ) {
    #
    # VOLUME
    #
    if ( exists $data_in->{$timestamp}{volume} ) {
      foreach my $line_idx ( sort { $a <=> $b } keys %{ $data_in->{$timestamp}{volume} } ) {
        my $var = $data_in->{$timestamp}{volume}{$line_idx};

        #print Dumper $var;
        #'Time' => '2017:09:18T13:30:51.000000+0200',
        #'Volume ID' => '102899',
        #'Interval (s)' => '605',
        #'Volume Name' => 'vio_3p_18',
        #'Pool ID' => '102780',
        #'Pool Name' => 'VIO_3P',
        #'Read Transfer Size (kB)' => '2.235',
        #'Write Transfer Size (kB)' => '4.427',
        #'Overall Transfer Size (kB)' => '4.133',
        #'Overall Response Time (ms)' => '0.000',

        if ( exists $var->{'Volume ID'} ) {
          my $id = lc( Xormon::create_hash_key($cfg->{'hostcfg_id'}."volume".$var->{'Volume ID'}) );

          # it was calculated from counters, so here must be a counter overflow check (value must be higher than 0)
          if ( exists $var->{'Read IO Rate (IO/s)'} && Xormon::isdigit($var->{'Read IO Rate (IO/s)'}) && $var->{'Read IO Rate (IO/s)'} > 0 ) {
            $data_out{xiv}{volume}{$id}{$timestamp}{read_io}   = Xormon::formatNumberByType("float4",$var->{'Read IO Rate (IO/s)'});
          }
          if ( exists $var->{'Write IO Rate (IO/s)'} && Xormon::isdigit($var->{'Write IO Rate (IO/s)'}) && $var->{'Write IO Rate (IO/s)'} > 0 ) {
            $data_out{xiv}{volume}{$id}{$timestamp}{write_io}  = Xormon::formatNumberByType("float4",$var->{'Write IO Rate (IO/s)'});
          }
          if ( exists $var->{'Total IO Rate (IO/s)'} && Xormon::isdigit($var->{'Total IO Rate (IO/s)'}) && $var->{'Total IO Rate (IO/s)'} > 0 ) {
            $data_out{xiv}{volume}{$id}{$timestamp}{io_rate}   = Xormon::formatNumberByType("float4",$var->{'Total IO Rate (IO/s)'});
          }
          if ( exists $var->{'Read Data Rate (KB/s)'} && Xormon::isdigit($var->{'Read Data Rate (KB/s)'}) && $var->{'Read Data Rate (KB/s)'} > 0 ) {
            $data_out{xiv}{volume}{$id}{$timestamp}{read}      = Xormon::formatNumberByType("float4",Xormon::returnBytes($var->{'Read Data Rate (KB/s)'},"kB"));
          }
          if ( exists $var->{'Write Data Rate (KB/s)'} && Xormon::isdigit($var->{'Write Data Rate (KB/s)'}) && $var->{'Write Data Rate (KB/s)'} > 0 ) {
            $data_out{xiv}{volume}{$id}{$timestamp}{write}     = Xormon::formatNumberByType("float4",Xormon::returnBytes($var->{'Write Data Rate (KB/s)'},"kB"));
          }
          if ( exists $var->{'Total Data Rate (KB/s)'} && Xormon::isdigit($var->{'Total Data Rate (KB/s)'}) && $var->{'Total Data Rate (KB/s)'} > 0 ) {
            $data_out{xiv}{volume}{$id}{$timestamp}{data_rate} = Xormon::formatNumberByType("float4",Xormon::returnBytes($var->{'Total Data Rate (KB/s)'},"kB"));
          }
          if ( exists $var->{'Read Response Time (ms)'} && Xormon::isdigit($var->{'Read Response Time (ms)'}) && $var->{'Read Response Time (ms)'} > 0 ) {
            $data_out{xiv}{volume}{$id}{$timestamp}{resp_t_r}  = Xormon::formatNumberByType("float4",$var->{'Read Response Time (ms)'});
          }
          if ( exists $var->{'Write Response Time (ms)'} && Xormon::isdigit($var->{'Write Response Time (ms)'}) && $var->{'Write Response Time (ms)'} > 0 ) {
            $data_out{xiv}{volume}{$id}{$timestamp}{resp_t_w}  = Xormon::formatNumberByType("float4",$var->{'Write Response Time (ms)'});
          }

          # add capacity stats from conf file
          if ( exists $capacity{volume}{$id}{capacity_total} && Xormon::isdigit($capacity{volume}{$id}{capacity_total}) ) { $data_out{xiv}{volume}{$id}{$timestamp}{capacity_total} = $capacity{volume}{$id}{capacity_total} };

          my $vol_stats = $data_out{xiv}{volume}{$id}{$timestamp};

          # calculate resp_t
          if ( exists $vol_stats->{resp_t_r} && Xormon::isdigit($vol_stats->{resp_t_r}) && exists $vol_stats->{resp_t_w} && Xormon::isdigit($vol_stats->{resp_t_w}) && exists $vol_stats->{read_io} && Xormon::isdigit($vol_stats->{read_io}) && exists $vol_stats->{write_io} && Xormon::isdigit($vol_stats->{write_io}) ) {
            my $resp_t = Xormon::calculate_weighted_avg( [ { count => $vol_stats->{read_io}, value => $vol_stats->{resp_t_r} }, { count => $vol_stats->{write_io}, value => $vol_stats->{resp_t_w} } ] );
            if ( defined $resp_t ) { $data_out{xiv}{volume}{$id}{$timestamp}{resp_t} = Xormon::formatNumberByType("float4",$resp_t); }
          }
          # calculate block size
          if ( exists $vol_stats->{io_rate} && Xormon::isdigit($vol_stats->{io_rate}) && exists $vol_stats->{data_rate} ) {
            if ( $vol_stats->{io_rate} == 0 ) {
              $data_out{xiv}{volume}{$id}{$timestamp}{block_size} = 0;
            }
            else {
              $data_out{xiv}{volume}{$id}{$timestamp}{block_size} = Xormon::formatNumberByType("float4",$vol_stats->{data_rate} / $vol_stats->{io_rate});
            }
          }
          if ( exists $vol_stats->{read_io} && Xormon::isdigit($vol_stats->{read_io}) && exists $vol_stats->{read} ) {
            if ( $vol_stats->{read_io} == 0 ) {
              $data_out{xiv}{volume}{$id}{$timestamp}{block_size_r} = 0;
            }
            else {
              $data_out{xiv}{volume}{$id}{$timestamp}{block_size_r} = Xormon::formatNumberByType("float4",$vol_stats->{read} / $vol_stats->{read_io});
            }
          }
          if ( exists $vol_stats->{write_io} && Xormon::isdigit($vol_stats->{write_io}) && exists $vol_stats->{write} ) {
            if ( $vol_stats->{write_io} == 0 ) {
              $data_out{xiv}{volume}{$id}{$timestamp}{block_size_w} = 0;
            }
            else {
              $data_out{xiv}{volume}{$id}{$timestamp}{block_size_w} = Xormon::formatNumberByType("float4",$vol_stats->{write} / $vol_stats->{write_io});
            }
          }

          # add volume uuid to pool
          if ( exists $var->{'Pool ID'} ) {
            my $pool_id = lc( Xormon::create_hash_key($cfg->{'hostcfg_id'}."pool".$var->{'Pool ID'}) );
            push( @{ $data_in->{$timestamp}{pool}{$pool_id}{volumes} }, $id );

            if ( exists $var->{'Pool Name'} ) {
              $data_in->{$timestamp}{pool}{$pool_id}{label} = $var->{'Pool Name'};
            }
          }
        }
      }
    }

    #
    # POOL
    #
    if ( exists $data_in->{$timestamp}{pool} ) {
      foreach my $id ( keys %{ $data_in->{$timestamp}{pool} } ) {
        my $var = $data_in->{$timestamp}{pool}{$id};

        # calculate data from the assigned volumes stats
        #
        my @met_aggregate    = ("io_rate","read_io","write_io","read","write","data_rate");
        my @met_weighted_avg = ( { count_key => 'read_io', value_key => 'resp_t_r' }, { count_key => 'write_io', value_key => 'resp_t_w' }, { count_key => 'io_rate', value_key => 'resp_t' } );

        my $values2avg;
        my $values2weighted_avg;

        if ( exists $var->{'volumes'} ) {
          foreach my $volume_id ( @{ $var->{'volumes'} } ) {
            if ( exists $data_out{xiv}{volume}{$volume_id} && exists $data_out{xiv}{volume}{$volume_id}{$timestamp} ) { # must be like this
              # aggregate stats
              foreach my $met_act ( @met_aggregate ) {
                if ( exists $data_out{xiv}{volume}{$volume_id}{$timestamp}{$met_act} ) {
                  if ( exists $data_out{xiv}{pool}{$id}{$timestamp}{$met_act} ) {
                    $data_out{xiv}{pool}{$id}{$timestamp}{$met_act} += $data_out{xiv}{volume}{$volume_id}{$timestamp}{$met_act};
                  }
                  else {
                    $data_out{xiv}{pool}{$id}{$timestamp}{$met_act} = $data_out{xiv}{volume}{$volume_id}{$timestamp}{$met_act};
                  }
                }
              }
              # weighted avg stats
              foreach my $met_act ( @met_weighted_avg ) {
                if ( exists $data_out{xiv}{volume}{$volume_id}{$timestamp}{ $met_act->{count_key} } && exists $data_out{xiv}{volume}{$volume_id}{$timestamp}{ $met_act->{value_key} } ) {
                  push(@{ $values2weighted_avg->{ $met_act->{value_key} } }, { count => $data_out{xiv}{volume}{$volume_id}{$timestamp}{ $met_act->{count_key} }, value => $data_out{xiv}{volume}{$volume_id}{$timestamp}{ $met_act->{value_key} } } );
                }
              }
            }
          }
        }

        # calculate weighted avg
        foreach my $met_act ( @met_weighted_avg ) {
          if ( exists $values2weighted_avg->{ $met_act->{value_key} } ) {
            my $value = Xormon::calculate_weighted_avg($values2weighted_avg->{ $met_act->{value_key} });
            #print "$met_act->{value_key} = $value\n";
            if ( defined $value ) { $data_out{xiv}{pool}{$id}{$timestamp}{ $met_act->{value_key} } = $value; }
          }
        }
        # calculate block size
        if ( exists $data_out{xiv}{pool}{$id}{$timestamp}{io_rate} && exists $data_out{xiv}{pool}{$id}{$timestamp}{data_rate} ) {
          if ( $data_out{xiv}{pool}{$id}{$timestamp}{io_rate} == 0 ) {
            $data_out{xiv}{pool}{$id}{$timestamp}{block_size} = 0;
          }
          else {
            $data_out{xiv}{pool}{$id}{$timestamp}{block_size} = $data_out{xiv}{pool}{$id}{$timestamp}{data_rate} / $data_out{xiv}{pool}{$id}{$timestamp}{io_rate};
          }
        }
        if ( exists $data_out{xiv}{pool}{$id}{$timestamp}{read_io} && exists $data_out{xiv}{pool}{$id}{$timestamp}{read} ) {
          if ( $data_out{xiv}{pool}{$id}{$timestamp}{read_io} == 0 ) {
            $data_out{xiv}{pool}{$id}{$timestamp}{block_size_r} = 0;
          }
          else {
            $data_out{xiv}{pool}{$id}{$timestamp}{block_size_r} = $data_out{xiv}{pool}{$id}{$timestamp}{read} / $data_out{xiv}{pool}{$id}{$timestamp}{read_io};
          }
        }
        if ( exists $data_out{xiv}{pool}{$id}{$timestamp}{write_io} && exists $data_out{xiv}{pool}{$id}{$timestamp}{write} ) {
          if ( $data_out{xiv}{pool}{$id}{$timestamp}{write_io} == 0 ) {
            $data_out{xiv}{pool}{$id}{$timestamp}{block_size_w} = 0;
          }
          else {
            $data_out{xiv}{pool}{$id}{$timestamp}{block_size_w} = $data_out{xiv}{pool}{$id}{$timestamp}{write} / $data_out{xiv}{pool}{$id}{$timestamp}{write_io};
          }
        }
        # format aggregated values
        foreach my $metric_key ( sort keys %{ $data_out{xiv}{pool}{$id}{$timestamp} } ) {
          $data_out{xiv}{pool}{$id}{$timestamp}{$metric_key} = Xormon::formatNumberByType("float4",$data_out{xiv}{pool}{$id}{$timestamp}{$metric_key});
        }

        # add capacity stats from conf file
        if ( exists $capacity{pool}{$id}{capacity_total}        && Xormon::isdigit($capacity{pool}{$id}{capacity_total}) )        { $data_out{xiv}{pool}{$id}{$timestamp}{capacity_total}        = $capacity{pool}{$id}{capacity_total} };
        if ( exists $capacity{pool}{$id}{capacity_free}         && Xormon::isdigit($capacity{pool}{$id}{capacity_free}) )         { $data_out{xiv}{pool}{$id}{$timestamp}{capacity_free}         = $capacity{pool}{$id}{capacity_free} };
        if ( exists $capacity{pool}{$id}{capacity_used}         && Xormon::isdigit($capacity{pool}{$id}{capacity_used}) )         { $data_out{xiv}{pool}{$id}{$timestamp}{capacity_used}         = $capacity{pool}{$id}{capacity_used} };
        if ( exists $capacity{pool}{$id}{capacity_used_percent} && Xormon::isdigit($capacity{pool}{$id}{capacity_used_percent}) ) { $data_out{xiv}{pool}{$id}{$timestamp}{capacity_used_percent} = $capacity{pool}{$id}{capacity_used_percent} };
      }
    }


    #
    # DEVICE
    #
    if ( exists $data_out{xiv}{volume} ) {
      my $id = $cfg->{'hostcfg_id'};

      # calculate data from the volumes stats
      #
      my @met_aggregate    = ("io_rate","read_io","write_io","read","write","data_rate");
      my @met_weighted_avg = ( { count_key => 'read_io', value_key => 'resp_t_r' }, { count_key => 'write_io', value_key => 'resp_t_w' }, { count_key => 'io_rate', value_key => 'resp_t' } );

      my $values2avg;
      my $values2weighted_avg;

      foreach my $volume_id ( %{ $data_out{xiv}{volume} } ) {
        if ( exists $data_out{xiv}{volume}{$volume_id} && exists $data_out{xiv}{volume}{$volume_id}{$timestamp} ) { # must be like this
          # aggregate stats
          foreach my $met_act ( @met_aggregate ) {
            if ( exists $data_out{xiv}{volume}{$volume_id}{$timestamp}{$met_act} && Xormon::isdigit($data_out{xiv}{volume}{$volume_id}{$timestamp}{$met_act}) ) {
              if ( exists $data_out{xiv}{device}{$id}{$timestamp}{$met_act} ) {
                $data_out{xiv}{device}{$id}{$timestamp}{$met_act} += $data_out{xiv}{volume}{$volume_id}{$timestamp}{$met_act};
              }
              else {
                $data_out{xiv}{device}{$id}{$timestamp}{$met_act} = $data_out{xiv}{volume}{$volume_id}{$timestamp}{$met_act};
              }
            }
          }
          # weighted avg stats
          foreach my $met_act ( @met_weighted_avg ) {
            if ( exists $data_out{xiv}{volume}{$volume_id}{$timestamp}{ $met_act->{count_key} } && exists $data_out{xiv}{volume}{$volume_id}{$timestamp}{ $met_act->{value_key} } ) {
              push(@{ $values2weighted_avg->{ $met_act->{value_key} } }, { count => $data_out{xiv}{volume}{$volume_id}{$timestamp}{ $met_act->{count_key} }, value => $data_out{xiv}{volume}{$volume_id}{$timestamp}{ $met_act->{value_key} } } );
            }
          }
        }
      }

      # calculate weighted avg
      foreach my $met_act ( @met_weighted_avg ) {
        if ( exists $values2weighted_avg->{ $met_act->{value_key} } ) {
          my $value = Xormon::calculate_weighted_avg($values2weighted_avg->{ $met_act->{value_key} });
          #print "$met_act->{value_key} = $value\n";
          if ( defined $value ) { $data_out{xiv}{device}{$id}{$timestamp}{ $met_act->{value_key} } = $value; }
        }
      }
      # format aggregated values
      foreach my $metric_key ( sort keys %{ $data_out{xiv}{device}{$id}{$timestamp} } ) {
        $data_out{xiv}{device}{$id}{$timestamp}{$metric_key} = Xormon::formatNumberByType("float4",$data_out{xiv}{device}{$id}{$timestamp}{$metric_key});
      }

      # add capacity stats from conf file
      if ( exists $capacity{device}{capacity_total}        && Xormon::isdigit($capacity{device}{capacity_total}) )        { $data_out{xiv}{device}{$id}{$timestamp}{capacity_total}        = $capacity{device}{capacity_total} };
      if ( exists $capacity{device}{capacity_free}         && Xormon::isdigit($capacity{device}{capacity_free}) )         { $data_out{xiv}{device}{$id}{$timestamp}{capacity_free}         = $capacity{device}{capacity_free} };
      if ( exists $capacity{device}{capacity_used}         && Xormon::isdigit($capacity{device}{capacity_used}) )         { $data_out{xiv}{device}{$id}{$timestamp}{capacity_used}         = $capacity{device}{capacity_used} };
      if ( exists $capacity{device}{capacity_used_percent} && Xormon::isdigit($capacity{device}{capacity_used_percent}) ) { $data_out{xiv}{device}{$id}{$timestamp}{capacity_used_percent} = $capacity{device}{capacity_used_percent} };
    }


    #
    # HOST (aggregated stats from volumes)
    #
    if ( scalar(keys %hosts) > 0 ) {
      foreach my $id ( keys %hosts ) {
        # calculate data from the volumes stats
        #
        my @met_aggregate    = ("io_rate","read_io","write_io","read","write","data_rate");
        my @met_weighted_avg = ( { count_key => 'read_io', value_key => 'resp_t_r' }, { count_key => 'write_io', value_key => 'resp_t_w' }, { count_key => 'io_rate', value_key => 'resp_t' } );

        my $values2avg;
        my $values2weighted_avg;

        if ( exists $hosts{$id}{volumes} ) {
          foreach my $volume_id ( keys %{ $hosts{$id}{volumes} } ) {
            if ( exists $data_out{xiv}{volume}{$volume_id} && exists $data_out{xiv}{volume}{$volume_id}{$timestamp} ) { # must be like this
              # aggregate stats
              foreach my $met_act ( @met_aggregate ) {
                if ( exists $data_out{xiv}{volume}{$volume_id}{$timestamp}{$met_act} && Xormon::isdigit($data_out{xiv}{volume}{$volume_id}{$timestamp}{$met_act}) ) {
                  if ( exists $data_out{xiv}{host}{$id}{$timestamp}{$met_act} ) {
                    $data_out{xiv}{host}{$id}{$timestamp}{$met_act} += $data_out{xiv}{volume}{$volume_id}{$timestamp}{$met_act};
                  }
                  else {
                    $data_out{xiv}{host}{$id}{$timestamp}{$met_act} = $data_out{xiv}{volume}{$volume_id}{$timestamp}{$met_act};
                  }
                }
              }
              # weighted avg stats
              foreach my $met_act ( @met_weighted_avg ) {
                if ( exists $data_out{xiv}{volume}{$volume_id}{$timestamp}{ $met_act->{count_key} } && exists $data_out{xiv}{volume}{$volume_id}{$timestamp}{ $met_act->{value_key} } ) {
                  push(@{ $values2weighted_avg->{ $met_act->{value_key} } }, { count => $data_out{xiv}{volume}{$volume_id}{$timestamp}{ $met_act->{count_key} }, value => $data_out{xiv}{volume}{$volume_id}{$timestamp}{ $met_act->{value_key} } } );
                }
              }
            }
          }

          # calculate weighted avg
          foreach my $met_act ( @met_weighted_avg ) {
            if ( exists $values2weighted_avg->{ $met_act->{value_key} } ) {
              my $value = Xormon::calculate_weighted_avg($values2weighted_avg->{ $met_act->{value_key} });
              #print "$met_act->{value_key} = $value\n";
              if ( defined $value ) { $data_out{xiv}{host}{$id}{$timestamp}{ $met_act->{value_key} } = $value; }
            }
          }
          # format aggregated values
          foreach my $metric_key ( sort keys %{ $data_out{xiv}{host}{$id}{$timestamp} } ) {
            $data_out{xiv}{host}{$id}{$timestamp}{$metric_key} = Xormon::formatNumberByType("float4",$data_out{xiv}{host}{$id}{$timestamp}{$metric_key});
          }
        }
      }
    }

  }

  #print "#################### PERF\n";
  #print Dumper \%data_out;
  #print Dumper \%{$data_out{xiv}{pool}};
  #print Dumper \%{$data_out{xiv}{device}};
  #print Dumper \%{$data_out{xiv}{host}};


  return \%data_out;
}

sub parse_perf {
  my $perf_file = shift;

  open( IN, "<$perf_file" ) || $logger->error( "Couldn't open file $perf_file $!" . __FILE__ . ":" . __LINE__ ) && return 0;
  my @lines = <IN>;
  close(IN);

  my %data_in;
  my $section    = "";
  my $act_header = "";
  my $stat_time  = "";

  my $line_idx = 0;
  foreach my $line (@lines) {
    chomp $line;
    $line_idx++;

    #print $line."\n";

    $line =~ s/^\s+//g;
    $line =~ s/\s+$//g;

    if ( $line =~ m/^----------/ ) { next; }

    if ( $line =~ m/Volume Level Statistics/ )         { $section    = "volume"; next; }
    if ( $line =~ m/^Volume ID,Time,Interval \(s\),/ ) { $act_header = $line;    next; }

    if ( $line =~ m/Interval End:/ ) {
      # Interval End:     2017:09:18T13:40:55.000000+0200
      $line =~ s/Interval End:\s+//;
      $line =~ s/GMT.*//;
      $stat_time = str2time($line);
      unless ( Xormon::isdigit($stat_time) ) { $stat_time = ""; }
      #print "####### $section : $stat_time : ".localtime($stat_time)." : line=$line\n";
      ################################# test only
      #$stat_time = time();
      ################################# test only
      next;
    }
    if ( $line =~ m/^Interval\s+/ ) { next; }
    # clear out type when empty line
    if ( $line =~ m/^$/ ) { $section = ""; $act_header = ""; }

    if ( defined $stat_time && $stat_time ne '' ) {
      $data_in{$stat_time}{$section}{$line_idx} = Xormon::parse_CSV_line_by_header(",",$act_header,$line);
    }
  }
  #print Dumper \%data_in;

  return \%data_in;
}

sub send_conf {
  my $data = shift;

  my @architecture;
  my %conf;

  #
  # DEVICE
  #
  # Architecture
  my %device_arch = (
    "item_id"    => $cfg->{'hostcfg_id'},
    "hostcfg_id" => $cfg->{'hostcfg_id'},
    "label"      => $cfg->{'hostalias'},
    "hw_type"    => "xiv",
    "subsystem"  => "device",
    "parents"    => []
  );
  push(@architecture, \%device_arch);
  # conf
  if ( exists $cfg->{'host'} )                    { $conf{ $cfg->{'hostcfg_id'} }{'ip'}             = $cfg->{'host'}; }
  if ( exists $data->{device}{'Machine Name'} )   { $conf{ $cfg->{'hostcfg_id'} }{'machine_name'}   = $data->{device}{'Machine Name'}; }
  if ( exists $data->{device}{'Machine Serial'} ) { $conf{ $cfg->{'hostcfg_id'} }{'serial'}         = $data->{device}{'Machine Serial'}; }
  if ( exists $data->{device}{'Machine Type'} )   { $conf{ $cfg->{'hostcfg_id'} }{'model'}          = $data->{device}{'Machine Type'}; }
  #if ( exists $data->{device}{'Machine Number'} ) { $conf{ $cfg->{'hostcfg_id'} }{'machine_number'} = $data->{device}{'Machine Number'}; } # looks like it's missing in the last conf files, it was presented in the old ones
  if ( exists $data->{device}{'Code Level'} )     { $conf{ $cfg->{'hostcfg_id'} }{'version'}        = $data->{device}{'Code Level'}; }
  if ( exists $data->{device}{'Total Hard Capacity'} ) {
    $data->{device}{'Total Hard Capacity'} =~ s/^\s+//;
    $data->{device}{'Total Hard Capacity'} =~ s/\s+TiB//;
    #$conf{ $cfg->{'hostcfg_id'} }{'capacity_total'} = Xormon::formatNumberByType("int",Xormon::returnBytes($data->{device}{'Total Hard Capacity'},"TiB"));
    $capacity{device}{capacity_total} = Xormon::formatNumberByType("int",Xormon::returnBytes($data->{device}{'Total Hard Capacity'},"TiB"));
  }
  elsif ( exists $data->{device}{'Total Capacity'} ) {
    $data->{device}{'Total Capacity'} =~ s/^\s+//;
    $data->{device}{'Total Capacity'} =~ s/\s+TiB//;
    #$conf{ $cfg->{'hostcfg_id'} }{'capacity_total'} = Xormon::formatNumberByType("int",Xormon::returnBytes($data->{device}{'Total Capacity'},"TiB"));
    $capacity{device}{capacity_total} = Xormon::formatNumberByType("int",Xormon::returnBytes($data->{device}{'Total Capacity'},"TiB"));
  }
  if ( exists $data->{device}{'Total Hard Free Capacity'} ) {
    $data->{device}{'Total Hard Free Capacity'} =~ s/^\s+//;
    $data->{device}{'Total Hard Free Capacity'} =~ s/\s+TiB//;
    #$conf{ $cfg->{'hostcfg_id'} }{'capacity_free'} = Xormon::formatNumberByType("int",Xormon::returnBytes($data->{device}{'Total Hard Free Capacity'},"TiB"));
    $capacity{device}{capacity_free} = Xormon::formatNumberByType("int",Xormon::returnBytes($data->{device}{'Total Hard Free Capacity'},"TiB"));
  }
  elsif ( exists $data->{device}{'Total Free Capacity'} ) {
    $data->{device}{'Total Free Capacity'} =~ s/^\s+//;
    $data->{device}{'Total Free Capacity'} =~ s/\s+TiB//;
    #$conf{ $cfg->{'hostcfg_id'} }{'capacity_free'} = Xormon::formatNumberByType("int",Xormon::returnBytes($data->{device}{'Total Free Capacity'},"TiB"));
    $capacity{device}{capacity_free} = Xormon::formatNumberByType("int",Xormon::returnBytes($data->{device}{'Total Free Capacity'},"TiB"));
  }
  if ( exists $capacity{device}{capacity_total} && exists $capacity{device}{capacity_free} ) {
    $capacity{device}{capacity_used}         = $capacity{device}{capacity_total} - $capacity{device}{capacity_free};
    $capacity{device}{capacity_used_percent} = $capacity{device}{capacity_used} / ( $capacity{device}{capacity_total} / 100 );
  }
  #'Total Soft Capacity' => '	71.953 TiB',
  #'Total Soft Free Capacity' => '	6.500 TiB',

  #
  # POOL
  #
  if ( exists $data->{pool} ) {
    foreach ( sort keys %{ $data->{pool} } ) {
      my $var = $data->{pool}{$_};
      if ( exists $var->{id} && exists $var->{name} ) {
        my $id = lc( Xormon::create_hash_key($cfg->{'hostcfg_id'}."pool".$var->{id}) );
        # Architecture
        my %arch = (
          "item_id"    => $id,
          "label"      => $var->{name},
          "hw_type"    => "xiv",
          "subsystem"  => "pool",
          "hostcfg_id" => $cfg->{'hostcfg_id'},
          "parents"    => [$cfg->{'hostcfg_id'}]
        );
        # conf
        if ( exists $var->{'capacity'} ) { # probably in GiB
          #$conf{$id}{'capacity_total'} = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'capacity'},"GiB"));
          $capacity{pool}{$id}{capacity_total} = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'capacity'},"GiB"));
        }
        if ( exists $var->{'free_capacity'} ) { # probably in GiB
          #$conf{$id}{'capacity_free'} = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'free_capacity'},"GiB"));
          $capacity{pool}{$id}{capacity_free} = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'free_capacity'},"GiB"));
        }
        if ( exists $capacity{pool}{$id}{capacity_total} && exists $capacity{pool}{$id}{capacity_free} ) {
          $capacity{pool}{$id}{capacity_used}         = $capacity{pool}{$id}{capacity_total} - $capacity{pool}{$id}{capacity_free};
          $capacity{pool}{$id}{capacity_used_percent} = $capacity{pool}{$id}{capacity_used} / ( $capacity{pool}{$id}{capacity_total} / 100 );
        }

        push(@architecture, \%arch);
      }
    }
  }

  #
  # VOLUME
  #
  if ( exists $data->{volume} ) {
    foreach ( sort keys %{ $data->{volume} } ) {
      my $var = $data->{volume}{$_};
      #if ( exists $var->{vdisk_UID} && exists $var->{name} ) {
      if ( exists $var->{volume_id} && exists $var->{name} ) {
        my $id = lc( Xormon::create_hash_key($cfg->{'hostcfg_id'}."volume".$var->{volume_id}) );
        #my $id = lc( $var->{vdisk_UID} );
        # Architecture
        my %arch = (
          "item_id"    => $id,
          "label"      => $var->{name},
          "hw_type"    => "xiv",
          "subsystem"  => "volume",
          "hostcfg_id" => $cfg->{'hostcfg_id'},
          "parents"    => [$cfg->{'hostcfg_id'}]
        );

        #'pool_name' => 'C1Q',
        #'id' => 'IBM.2810-6001584-113372',
        #'vdisk_UID' => '0017380006300377'

        #if ( exists $var->{'id'} ) { $conf{$id}{'id'} = $var->{'id'}; }
        if ( exists $var->{'pool_id'} ) {
          #$conf{$id}{'pool_id'} = $var->{'pool_id'};
          push(@{ $arch{parents} }, lc( Xormon::create_hash_key($cfg->{'hostcfg_id'}."pool".$var->{'pool_id'}) ));
        }
        if ( exists $var->{'capacity'} ) { # probably in GiB
          #$conf{$id}{'capacity_total'} = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'capacity'},"GiB"));
          $capacity{volume}{$id}{capacity_total} = Xormon::formatNumberByType("int",Xormon::returnBytes($var->{'capacity'},"GiB"));
        }
        if ( exists $var->{'vdisk_UID'} ) {
          # WWN for mapping
          $arch{wwn} = Xormon::formatWWN($var->{'vdisk_UID'});
          # conf
          $conf{$id}{'lun_id'} = $var->{'vdisk_UID'};
        }

        push(@architecture, \%arch);
      }
    }
  }

  #
  # HOST
  #
  if ( exists $data->{host} ) {

    foreach ( sort keys %{ $data->{host} } ) {
      my $var = $data->{host}{$_};
      if ( exists $var->{id} && exists $var->{name} ) {
        my $id = lc( Xormon::create_hash_key($cfg->{'hostcfg_id'}."host".$var->{id}) );
        # Architecture
        my %arch = (
          "item_id"    => $id,
          "label"      => $var->{name},
          "hw_type"    => "xiv",
          "subsystem"  => "host",
          "hostcfg_id" => $cfg->{'hostcfg_id'},
          "parents"    => [$cfg->{'hostcfg_id'}]
        );
        # conf
        #if ( exists $var->{'id'} ) { $conf{$id}{'id'} = $var->{'id'}; }

        #'host_id' => 'IBM.2810-6001573-0000000000019F91',
        #'Volume Names' => 'b1p-db_1 b1p-db_2 b1p-db-h_1 b1p-db-h_2 b1p-db-h_3 b1p-db_3',

        $hosts{$id}{label} = $var->{name};

        if ( exists $var->{'Volume IDs'} ) {
          foreach my $vol_id ( split(/\s+/, $var->{'Volume IDs'}) ) {
            my $volume_id = lc( Xormon::create_hash_key($cfg->{'hostcfg_id'}."volume".$vol_id));
            push(@{ $arch{children} }, $volume_id);
            $hosts{$id}{volumes}{$volume_id} = $volume_id;
          }
        }

        push(@architecture, \%arch);
      }
    }
  }

  #print "#################### Architecture\n";
  #print Dumper @architecture;
  #print "#################### CONF\n";
  #print Dumper \%conf;

  #save architecture
  my $rc_arch = $xormon->saveArchitecture(\@architecture);
  $logger->info("save architecture: $rc_arch");

  #save configuration
  my $rc_conf = $xormon->saveConfJSON(\%conf);
  $logger->info("save conf: $rc_conf");

  return 1;
}

sub parse_conf {
  my $conf_file = shift;

  open( IN, "<$conf_file" ) || $logger->error( "Couldn't open file $conf_file $!" . __FILE__ . ":" . __LINE__ ) && return 0;
  my @lines = <IN>;
  close(IN);

  my %data_in;
  my $section    = "";
  my $act_header = "";

  my $line_idx = 0;
  foreach my $line (@lines) {
    chomp $line;
    $line_idx++;

    $line =~ s/^\s+//g;
    $line =~ s/\s+$//g;

    if ( $line =~ m/^----------/ )          { next; }

    if ( $line =~ m/^$/ && $section ne "device" ) { $section = "NA";        next; }
    if ( $line =~ m/Configuration Data/ )         { $section = "device";    next; }
    if ( $line =~ m/Pool Level Configuration/ )   { $section = "pool";      next; }
    if ( $line =~ m/Volume Level Configuration/ ) { $section = "volume";    next; }
    if ( $line =~ m/Host Level Configuration/ )   { $section = "host";      next; }

    # main cfg, Configuration Data
    if ( $section eq "device" ) {
      if ( $line =~ m/^Machine Name:/ ) {
        $line =~ s/^Machine Name:\s*//;
        if ( $line ne '' ) {
          $data_in{device}{'Machine Name'} = $line;
        }
      }
      if ( $line =~ m/^Machine Type-Model:/ ) {
        $line =~ s/^Machine Type-Model:\s*//;
        if ( $line ne '' ) {
          $data_in{device}{'Machine Type'} = $line;
        }
      }
      if ( $line =~ m/^Machine Number:/ ) {
        $line =~ s/^Machine Number:\s*//;
        if ( $line ne '' ) {
          $data_in{device}{'Machine Number'} = $line;
        }
      }
      if ( $line =~ m/^Machine Serial:/ ) {
        $line =~ s/^Machine Serial:\s*//;
        if ( $line ne '' ) {
          $data_in{device}{'Machine Serial'} = $line;
        }
      }
      if ( $line =~ m/^Code Level:/ ) {
        $line =~ s/^Code Level:\s*//;
        if ( $line ne '' ) {
          $data_in{device}{'Code Level'} = $line;
        }
      }
      if ( $line =~ m/^Total Hard Capacity\s*:/ ) {
        $line =~ s/^Total Hard Capacity\s*:\s*//;
        if ( $line ne '' ) {
          $data_in{device}{'Total Hard Capacity'} = $line;
        }
      }
      if ( $line =~ m/^Total Hard Free Capacity:/ ) {
        $line =~ s/^Total Hard Free Capacity:\s*//;
        if ( $line ne '' ) {
          $data_in{device}{'Total Hard Free Capacity'} = $line;
        }
      }
      if ( $line =~ m/^Total Soft Capacity\s*:/ ) {
        $line =~ s/^Total Soft Capacity\s*:\s*//;
        if ( $line ne '' ) {
          $data_in{device}{'Total Soft Capacity'} = $line;
        }
      }
      if ( $line =~ m/^Total Soft Free Capacity:/ ) {
        $line =~ s/^Total Soft Free Capacity:\s*//;
        if ( $line ne '' ) {
          $data_in{device}{'Total Soft Free Capacity'} = $line;
        }
      }
      # different conf file...
      if ( $line =~ m/^Total Capacity\s*:/ ) {
        $line =~ s/^Total Capacity\s*:\s*//;
        if ( $line ne '' ) {
          $data_in{device}{'Total Capacity'} = $line;
        }
      }
      if ( $line =~ m/^Total Free Capacity:/ ) {
        $line =~ s/^Total Free Capacity:\s*//;
        if ( $line ne '' ) {
          $data_in{device}{'Total Free Capacity'} = $line;
        }
      }
    }

    # Pool Level Configuration
    if ( $section eq "pool" ) {
      # parse header
      if ( $line =~ m/^name,id,status,mdisk_count,/ ) {
        $act_header = $line;
        next;
      }
      # parse data lines
      $data_in{$section}{$line_idx} = Xormon::parse_CSV_line_by_header(",",$act_header,$line);
    }

    # Volume Level Configuration
    if ( $section eq "volume" ) {
      # parse header
      if ( $line =~ m/^volume_id,id,name,/ ) {
        $act_header = $line;
        next;
      }
      # parse data lines
      $data_in{$section}{$line_idx} = Xormon::parse_CSV_line_by_header(",",$act_header,$line);
    }

    # Host Level Configuration
    if ( $section eq "host" ) {
      # parse header
      if ( $line =~ m/^id,host_id,name,/ ) {
        $act_header = $line;
        next;
      }
      # parse data lines
      $data_in{$section}{$line_idx} = Xormon::parse_CSV_line_by_header(",",$act_header,$line);
    }

  }
  #print Dumper \%data_in;

  return \%data_in;
}

sub remove_file {
  my $file = shift;

  # do not remove perf and conf files for debug purposes
  if ( exists $ENV{KEEP_OUT_FILES} && ( $ENV{KEEP_OUT_FILES} eq "1" || $ENV{KEEP_OUT_FILES} eq "$cfg->{'hostalias'}" ) ) {
    return 1;
  }

  if ( -f $file ) {
    unlink($file) || $logger->error( "Couldn't remove file $file $!" . __FILE__ . ":" . __LINE__ ) && return 0;
    $logger->info("remove file: $file");
  }

  return 1;
}

1;
